﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;

namespace Wcdf.Control
{
    /// <summary>
    /// A UniformGrid implementation for Silverlight that arranges children in a
    /// grid with all equal cell sizes.
    /// </summary>
    public class UniformGridEx : Panel
    {
        /// <summary>
        /// Gets or sets the computed row value.
        /// </summary>
        private int ComputedRows { get; set; }

        /// <summary>
        /// Gets or sets the computed column value.
        /// </summary>
        private int ComputedColumns { get; set; }

        /// <summary>
        /// Initializes a new instance of UniformGrid.
        /// </summary>
        public UniformGridEx()
        {
        }

        /// <summary>
        /// Gets or sets the number of first columns to leave blank.
        /// </summary>
        public int FirstColumn
        {
            get { return (int)GetValue(FirstColumnProperty); }
            set { SetValue(FirstColumnProperty, value); }
        }

        /// <summary>
        /// The FirstColumnProperty dependency property.
        /// </summary>
        public static readonly DependencyProperty FirstColumnProperty =
                DependencyProperty.Register(
                        "FirstColumn",
                        typeof(int),
                        typeof(UniformGridEx),
                        new PropertyMetadata(0, OnIntegerDependencyPropertyChanged));

        /// <summary>
        /// Gets or sets the number of columns in the grid. A value of zero 
        /// indicates that the count should be dynamically computed based on the
        /// number of rows and the number of non-collapsed children in the grid.
        /// </summary>
        public int Columns
        {
            get { return (int)GetValue(ColumnsProperty); }
            set { SetValue(ColumnsProperty, value); }
        }

        /// <summary>
        /// DependencyProperty for the Columns property.
        /// </summary>
        public static readonly DependencyProperty ColumnsProperty =
                DependencyProperty.Register(
                        "Columns",
                        typeof(int),
                        typeof(UniformGridEx),
                        new PropertyMetadata(0, OnIntegerDependencyPropertyChanged));

        /// <summary>
        /// Validate the new property value and silently revert if the new value
        /// is not appropriate. Used in place of WPF value coercian by the 
        /// dependency properties in UniformGrid.
        /// </summary>
        /// <param name="o">The dependency object.</param>
        /// <param name="e">The dependency property.</param>
        private static void OnIntegerDependencyPropertyChanged(DependencyObject o, DependencyPropertyChangedEventArgs e)
        {
            // Silently coerce the value back to >= 0 if negative.
            if (!(e.NewValue is int) || (int)e.NewValue < 0)
            {
                o.SetValue(e.Property, e.OldValue);
            }
        }

        public static readonly DependencyProperty CellWidthProperty =
       DependencyProperty.Register("CellWidth", typeof(double), typeof(UniformGridEx), new PropertyMetadata((double)0, OnDoubleDependencyPropertyChanged));
        /// <summary>
        /// 
        /// </summary>
        public double CellWidth
        {
            get { return (double)GetValue(CellWidthProperty); }
            set { SetValue(CellWidthProperty, value); }
        }

        public static readonly DependencyProperty CellHeightProperty =
        DependencyProperty.Register("CellHeight", typeof(double), typeof(UniformGridEx), new PropertyMetadata((double)0, OnDoubleDependencyPropertyChanged));
        /// <summary>
        /// 
        /// </summary>
        public double CellHeight
        {
            get { return (double)GetValue(CellHeightProperty); }
            set { SetValue(CellHeightProperty, value); }
        }

        public static readonly DependencyProperty IsFlowProperty =
        DependencyProperty.Register("IsFlow", typeof(bool), typeof(UniformGridEx), new PropertyMetadata(null));
        /// <summary>
        /// 是否为顺序排列
        /// </summary>
        public bool IsFlow
        {
            get { return (bool)GetValue(IsFlowProperty); }
            set { SetValue(IsFlowProperty, value); }
        }

        /// <summary>
        /// Validate the new property value and silently revert if the new value
        /// is not appropriate. Used in place of WPF value coercian by the 
        /// dependency properties in UniformGrid.
        /// </summary>
        /// <param name="o">The dependency object.</param>
        /// <param name="e">The dependency property.</param>
        private static void OnDoubleDependencyPropertyChanged(DependencyObject o, DependencyPropertyChangedEventArgs e)
        {
            // Silently coerce the value back to >= 0 if negative.
            if (!(e.NewValue is double) || (double)e.NewValue < 0)
            {
                o.SetValue(e.Property, e.OldValue);
            }
        }

        /// <summary>
        /// Gets or sets the number of rows in the grid. A value of zero 
        /// indicates that the row count should be dynamically computed based on
        /// the number of columns and the number of non-collapsed children in
        /// the grid.
        /// </summary>
        public int Rows
        {
            get { return (int)GetValue(RowsProperty); }
            set { SetValue(RowsProperty, value); }
        }

        /// <summary>
        /// The Rows DependencyProperty.
        /// </summary>
        public static readonly DependencyProperty RowsProperty =
                DependencyProperty.Register(
                        "Rows",
                        typeof(int),
                        typeof(UniformGridEx),
                        new PropertyMetadata(0, OnIntegerDependencyPropertyChanged));

        /// <summary>
        /// Compute the desired size of the UniformGrid by measuring all of the
        /// children with a constraint equal to a cell's portion of the given
        /// constraint. The maximum child width and maximum child height are 
        /// tracked, and then the desired size is computed by multiplying these
        /// maximums by the row and column count.
        /// </summary>
        /// <param name="constraint">控件提供的可用空间.</param>
        /// <returns>通过计算后确定所需的空间.</returns>
        protected override Size MeasureOverride(Size constraint)
        {
            UpdateComputedValues();
            if (this.IsFlow)
            {
                for (int i = 0, count = Children.Count; i < count; ++i)
                {
                    UIElement child = Children[i];
                    child.Measure(constraint);
                }
                return constraint;
            }
            double cellWidth = this.CellWidth > constraint.Width / ComputedColumns ? constraint.Width / ComputedColumns : this.CellWidth;
            double cellHeight = this.CellHeight == 0 ? constraint.Height / ComputedRows : this.CellHeight;
            Size childConstraint = new Size(cellWidth, cellHeight);
            double maxChildDesiredWidth = 0.0;
            double maxChildDesiredHeight = 0.0;

            //  Measure each child, keeping track of max desired width & height.
            for (int i = 0, count = Children.Count; i < count; ++i)
            {
                UIElement child = Children[i];
                child.Measure(childConstraint);
                Size childDesiredSize = child.DesiredSize;
                if (maxChildDesiredWidth < childDesiredSize.Width)
                {
                    maxChildDesiredWidth = childDesiredSize.Width;
                }
                if (maxChildDesiredHeight < childDesiredSize.Height)
                {
                    maxChildDesiredHeight = childDesiredSize.Height;
                }
            }
            return new Size((maxChildDesiredWidth * ComputedColumns), (maxChildDesiredHeight * ComputedRows));
        }

        /// <summary>
        /// Arrange the children of the UniformGrid by distributing space evenly
        /// among the children, making each child the size equal to a cell
        /// portion of the arrangeSize parameter.
        /// </summary>
        /// <param name="arrangeSize">所占的空间.</param>
        /// <returns>实际所用到的空间.</returns>
        protected override Size ArrangeOverride(Size arrangeSize)
        {
            if (this.IsFlow)
                return ArrangebyFlow(arrangeSize);
            double cellWidth = this.CellWidth < arrangeSize.Width / ComputedColumns ? arrangeSize.Width / ComputedColumns : this.CellWidth;
            double cellHeight = this.CellHeight == 0 ? arrangeSize.Height / ComputedRows : this.CellHeight;
            Rect childBounds = new Rect(0, 0, cellWidth, cellHeight);
            double xStep = childBounds.Width;
            double xBound = arrangeSize.Width - 1.0;
            childBounds.X += childBounds.Width * FirstColumn;

            // Arrange and Position each child to the same cell size
            foreach (UIElement child in Children)
            {
                child.Arrange(childBounds);
                if (child.Visibility != Visibility.Collapsed)
                {
                    childBounds.X += childBounds.Width;
                    if (childBounds.X >= xBound)
                    {
                        childBounds.Y += childBounds.Height;
                        childBounds.X = 0;
                    }
                }
            }

            return arrangeSize;
        }

        private Size ArrangebyFlow(Size arrangeSize)
        {
            double xBound = arrangeSize.Width - 1.0;
            Rect childBounds = new Rect(0, 0, this.CellWidth, this.CellHeight);
            Size childSize = new Size(this.CellWidth, this.CellHeight);
            // Arrange and Position each child to the same cell size
            int count = 0;
            foreach (UIElement child in Children)
            {
                //if (child.DesiredSize.Width > xStep)
                //{
                //    var width = Math.Ceiling(child.DesiredSize.Width / xStep);
                //    var height = 
                //    Size ns = new Size(width, 0);
                //}
                //Rect childBounds = new Rect(0, 0, child.Measure(.Width, child.DesiredSize.Height);
                //child.
                //child.Measure(childSize);
                //Rect childBounds = new Rect(0, 0, child.DesiredSize.Width, child.DesiredSize.Height);
                //MessageBox.Show(child.RenderSize.Height.ToString() + childBounds.Height.ToString());
                //if (child.DesiredSize.Width > childBounds.Width || count == 0)
                //{
                //    childBounds.Width = child.DesiredSize.Width;
                //}
                //else
                //{
                //    childBounds.Width = child.DesiredSize.Width;
                //}
                //子控件长度为原长度
                childBounds.Width = child.DesiredSize.Width;
                if (child.DesiredSize.Height > childBounds.Height || count == 0)
                {
                    childBounds.Height = child.DesiredSize.Height;
                }
                //else
                //    childBounds.Height = child.DesiredSize.Height;
                //child.Arrange(childBounds);
                if (childBounds.X + childBounds.Width > arrangeSize.Width)
                {
                    //换行
                    count = 0;
                    //childBounds.Width = child.DesiredSize.Width;                    
                    childBounds.Y += childBounds.Height;
                    childBounds.X = 0;
                    child.Arrange(childBounds);
                    childBounds.Height = child.DesiredSize.Height;
                    //childBounds.Width = this.CellWidth;
                    //childBounds.Height = this.CellHeight;
                }
                else
                {
                    child.Arrange(childBounds);
                }
                //child.tran
                if (child.Visibility != Visibility.Collapsed)
                {
                    count++;
                    //MessageBox.Show(childBounds.Width.ToString());
                    childBounds.X += childBounds.Width;
                    if (childBounds.X >= xBound || count >= this.Columns)
                    {
                        count = 0;
                        childBounds.Y += childBounds.Height;
                        //childBounds.Width = this.CellWidth;
                        //childBounds.Height = this.CellHeight;
                        childBounds.X = 0;
                    }
                }
            }

            return arrangeSize;
        }
        /// <summary>
        /// If the Rows or Columns values are set to 0, dynamically compute the
        /// values based on the actual number of non-collapsed children.
        /// </summary>
        /// <remarks>
        /// In the case when both Rows and Columns are set to 0, the Rows and 
        /// Columns will be equal, laying out a square grid.
        /// </remarks>
        private void UpdateComputedValues()
        {
            ComputedColumns = Columns;
            ComputedRows = Rows;

            // Reset the first column. This is the same logic performed by WPF.
            if (FirstColumn >= ComputedColumns)
            {
                FirstColumn = 0;
            }

            if ((ComputedRows == 0) || (ComputedColumns == 0))
            {
                int nonCollapsedCount = 0;
                for (int i = 0, count = Children.Count; i < count; ++i)
                {
                    UIElement child = Children[i];
                    if (child.Visibility != Visibility.Collapsed)
                    {
                        nonCollapsedCount++;
                    }
                }
                if (nonCollapsedCount == 0)
                {
                    nonCollapsedCount = 1;
                }
                if (ComputedRows == 0)
                {
                    if (ComputedColumns > 0)
                    {
                        ComputedRows = (nonCollapsedCount + FirstColumn + (ComputedColumns - 1)) / ComputedColumns;
                    }
                    else
                    {
                        ComputedRows = (int)Math.Sqrt(nonCollapsedCount);
                        if ((ComputedRows * ComputedRows) < nonCollapsedCount)
                        {
                            ComputedRows++;
                        }
                        ComputedColumns = ComputedRows;
                    }
                }
                else if (ComputedColumns == 0)
                {
                    ComputedColumns = (nonCollapsedCount + (ComputedRows - 1)) / ComputedRows;
                }
            }
        }
    }
}
