﻿using System;
using Wcdf.Core;
using System.Security.Principal;
using System.Configuration;
using System.Threading;

namespace Wcdf
{
    /// <summary>
    /// Provides consistent context information between the client
    /// and server DataPortal objects. 
    /// </summary>
    public static class ApplicationContext
    {
        #region Context Manager

        private static IContextManager _webContextManager;
        private static IContextManager _contextManager;
        private static Type _webManagerType;
        private static Type _xamlManagerType;

        static ApplicationContext()
        {
            _webManagerType = Type.GetType("Wcdf.Web.ApplicationContextManager, Wcdf.Web");
            _xamlManagerType = Type.GetType("Wcdf.Xaml.ApplicationContextManager, Wcdf.Xaml");

            if (_webManagerType != null)
                WebContextManager = (IContextManager)Activator.CreateInstance(_webManagerType);
            if (_xamlManagerType != null)
                _contextManager = (IContextManager)Activator.CreateInstance(_xamlManagerType);
            if (_contextManager == null)
                _contextManager = new ApplicationContextManager();
        }

        /// <summary>
        /// Gets or sets the web context manager.
        /// Will use default WebContextManager. 
        /// Only need to set for non-default WebContextManager.
        /// </summary>
        /// <value>
        /// The web context manager.
        /// </value>
        public static IContextManager WebContextManager
        {
            get { return _webContextManager; }
            set { _webContextManager = value; }
        }

        /// <summary>
        /// 读取和设置维护着用户及上下文信息.
        /// 根据运行环境来确定使用哪种上下文管理器        
        /// </summary>
        public static IContextManager ContextManager
        {
            get
            {
                if (WebContextManager != null && WebContextManager.IsValid)
                    return WebContextManager;
                return _contextManager;
            }
            set { _contextManager = value; }
        }

        #endregion

        #region User

        /// <summary>
        /// Get or set the current <see cref="IPrincipal" />
        /// object representing the user's identity.
        /// </summary>
        /// <remarks>
        /// This is discussed in Chapter 5. When running
        /// under IIS the HttpContext.Current.User value
        /// is used, otherwise the current Thread.CurrentPrincipal
        /// value is used.
        /// </remarks>
        public static IPrincipal User
        {
            get { return ContextManager.GetUser(); }
            set { ContextManager.SetUser(value); }
        }

        #endregion

        #region LocalContext

        /// <summary>
        /// Returns the application-specific context data that
        /// is local to the current AppDomain.
        /// </summary>
        /// <remarks>
        /// <para>
        /// The return value is a HybridDictionary. If one does
        /// not already exist, and empty one is created and returned.
        /// </para><para>
        /// Note that data in this context is NOT transferred to and from
        /// the client and server.
        /// </para>
        /// </remarks>
        public static ContextDictionary LocalContext
        {
            get
            {
                ContextDictionary ctx = ContextManager.GetLocalContext();
                if (ctx == null)
                {
                    ctx = new ContextDictionary();
                    ContextManager.SetLocalContext(ctx);
                }
                return ctx;
            }
        }

        #endregion

        #region Client/Global Context

        private static object _syncContext = new object();

        /// <summary>
        /// Returns the application-specific context data provided
        /// by the client.
        /// </summary>
        /// <remarks>
        /// <para>
        /// The return value is a HybridDictionary. If one does
        /// not already exist, and empty one is created and returned.
        /// </para><para>
        /// Note that data in this context is transferred from
        /// the client to the server. No data is transferred from
        /// the server to the client.
        /// </para><para>
        /// This property is thread safe in a Windows client
        /// setting and on an application server. It is not guaranteed
        /// to be thread safe within the context of an ASP.NET
        /// client setting (i.e. in your ASP.NET UI).
        /// </para>
        /// </remarks>
        public static ContextDictionary ClientContext
        {
            get
            {
                lock (_syncContext)
                {
                    ContextDictionary ctx = ContextManager.GetClientContext();
                    if (ctx == null)
                    {
                        ctx = new ContextDictionary();
                        ContextManager.SetClientContext(ctx);
                    }
                    return ctx;
                }
            }
        }

        /// <summary>
        /// Returns the application-specific context data shared
        /// on both client and server.
        /// </summary>
        /// <remarks>
        /// <para>
        /// The return value is a HybridDictionary. If one does
        /// not already exist, and empty one is created and returned.
        /// </para><para>
        /// Note that data in this context is transferred to and from
        /// the client and server. Any objects or data in this context
        /// will be transferred bi-directionally across the network.
        /// </para>
        /// </remarks>
        public static ContextDictionary GlobalContext
        {
            get
            {
                ContextDictionary ctx = ContextManager.GetGlobalContext();
                if (ctx == null)
                {
                    ctx = new ContextDictionary();
                    ContextManager.SetGlobalContext(ctx);
                }
                return ctx;
            }
        }

        internal static void SetContext(
          ContextDictionary clientContext,
          ContextDictionary globalContext)
        {
            lock (_syncContext)
                ContextManager.SetClientContext(clientContext);
            ContextManager.SetGlobalContext(globalContext);
        }

        /// <summary>
        /// Clears all context collections.
        /// </summary>
        public static void Clear()
        {
            SetContext(null, null);
            ContextManager.SetLocalContext(null);
        }

        #endregion

        #region Config Settings

        private static string _authenticationType;

        /// <summary>
        /// 验证方式
        /// </summary>
        /// <value></value>
        /// <returns></returns>
        /// <remarks>
        /// 该值从应用程序配置文件配置点"WcdfAuthentication"读取.
        /// 如果值为"Windows"表示应用Windows集成验证.任务值为应用自
        /// 定义的验证方式.        
        /// </remarks>
        public static string AuthenticationType
        {
            get
            {
                if (_authenticationType == null)
                {
                    _authenticationType = ConfigurationManager.AppSettings["WcdfAuthentication"];
                    _authenticationType = _authenticationType ?? "Wcdf";
                }
                return _authenticationType;
            }
        }

        private static string _validateUserNameType;
        /// <summary>
        /// 验证用户，描述如何对用户请求的验证方式
        /// 当为FirstLogin时，在第一次用户登陆验证后，以后的每次请求不再验证。
        /// 当为PerCall时，每次请求都进行验证。
        /// </summary>
        public static string ValidateUserNameType
        {
            get
            {
                if (_validateUserNameType == null)
                {
                    _validateUserNameType = ConfigurationManager.AppSettings["WcdfValidateUserNameType"];
                    _validateUserNameType = _validateUserNameType ?? "FirstLogin";
                }
                return _validateUserNameType;
            }
        }

        private static string _portalProxy;

        /// <summary>
        /// 门户代理类的完全限定类型名.
        /// </summary>
        /// <value>指定的处理类或者默认"LocalProxy"处理</value>
        /// <returns></returns>
        /// <remarks>
        /// <para>
        /// 如果该值为空,将会从应用程序配置文件中读取"WcdfDataPortalProxy".        
        /// </para><para>
        /// 代理类必须继承自Wcdf.Server.IDataPortalServer
        /// </para><para>
        /// 内部值"Local"是Localproxy类的简称.服务处理也运行在客户端.        
        /// </para><para>
        /// 框架中所包含的处理类:
        /// <list>
        /// <item>
        /// <term>Wcdf,Wcdf.DataPortalClient.RemotingProxy</term>
        /// <description>与采用 .NET Remoting做为应用服务端通讯.</description>
        /// </item>
        /// <item>
        /// <term>Wcdf,Wcdf.DataPortalClient.EnterpriseServicesProxy</term>
        /// <description>与采用DCOM(COM+)做为应用服务端.</description>
        /// </item>
        /// <item>
        /// <term>Wcdf,Csla.DataPortalClient.WcfProxy</term>
        /// <description>与采用WCF框架做为应用服务端通讯.</description>
        /// </item>
        /// </list>
        /// 每种类似的代理对应各自的服务构架.        
        /// </para>
        /// </remarks>
        public static string DataPortalProxy
        {
            get
            {
                if (string.IsNullOrEmpty(_portalProxy))
                {
                    _portalProxy = ConfigurationManager.AppSettings["WcdfDataPortalProxy"];
                    if (string.IsNullOrEmpty(_portalProxy))
                        _portalProxy = "Local";
                }
                return _portalProxy;
            }
            set
            {
                _portalProxy = value;
                //DataPortal.ResetProxyType();
            }
        }

        private static string _portalUrl = null;

        /// <summary>
        /// 数据门户地址的URL字段串形式.        
        /// </summary>
        /// <value>URL格式字符串,需要以"\"结尾.</value>
        public static string DataPortalUrlString
        {
            get
            {
                if (_portalUrl == null)
                {
                    _portalUrl = ConfigurationManager.AppSettings["WcdfDataPortalUrl"];
                }
                return _portalUrl;
            }
            set
            {
                _portalUrl = value;
            }
        }

        /// <summary>
        /// 数据门户地址的URL格式.
        /// </summary>
        /// <value></value>
        /// <returns></returns>
        /// <remarks>
        /// This value is read from the application configuration
        /// file with the key value "CslaDataPortalUrl". 
        /// </remarks>
        public static Uri DataPortalUrl
        {
            get { return new Uri(DataPortalUrlString); }
        }

        /// <summary>
        /// 是否在更新时发送原业务对象的副本.        
        /// </summary>
        public static bool AutoCloneOnUpdate
        {
            get
            {
                bool result = true;
                string setting = ConfigurationManager.AppSettings["WcdfAutoCloneOnUpdate"];
                if (!string.IsNullOrEmpty(setting))
                    result = bool.Parse(setting);
                return result;
            }
        }

        /// <summary>
        /// Gets the serialization formatter type used by CSLA .NET
        /// for all explicit object serialization (such as cloning,
        /// n-level undo, etc).
        /// </summary>
        /// <remarks>
        /// <para>
        /// If you use the DataContract and DataMember attributes
        /// to specify how your objects should be serialized then
        /// you <b>must</b> change this setting to use the
        /// <see cref="System.Runtime.Serialization.NetDataContractSerializer">
        /// NetDataContractSerializer</see> option. The default is to
        /// use the standard Microsoft .NET 
        /// <see cref="System.Runtime.Serialization.Formatters.Binary.BinaryFormatter"/>.
        /// </para>
        /// <para>
        /// This setting does <b>not affect</b> the serialization
        /// formatters used by the various data portal channels.
        /// </para>
        /// <para>
        /// If you are using the Remoting, Web Services or 
        /// Enterprise Services technologies, they will use the
        /// BinaryFormatter regardless of this setting, and will
        /// <b>fail to work</b> if you attempt to use the
        /// DataContract and DataMember attributes when building
        /// your business objects.
        /// </para>
        /// <para>
        /// If you want to use DataContract and DataMember, and
        /// you want a remote data portal server, you <b>must</b>
        /// use the WCF data portal channel, or create your own
        /// custom channel that uses the
        /// <see cref="System.Runtime.Serialization.NetDataContractSerializer">
        /// NetDataContractSerializer</see> provided as part of WCF.
        /// </para>
        /// </remarks>
        public static SerializationFormatters SerializationFormatter
        {
            get
            {
                string tmp = ConfigurationManager.AppSettings["CslaSerializationFormatter"];
                if (string.IsNullOrEmpty(tmp))
                    tmp = "BinaryFormatter";
                return (SerializationFormatters)
                  Enum.Parse(typeof(SerializationFormatters), tmp);
            }
        }

        private static PropertyChangedModes _propertyChangedMode;
        private static bool _propertyChangedModeSet;
        /// <summary>
        /// Gets or sets a value specifying how CSLA .NET should
        /// raise PropertyChanged events.
        /// </summary>
        public static PropertyChangedModes PropertyChangedMode
        {
            get
            {
                if (!_propertyChangedModeSet)
                {
                    string tmp = ConfigurationManager.AppSettings["CslaPropertyChangedMode"];
                    if (string.IsNullOrEmpty(tmp))
                        tmp = "Xaml";
                    _propertyChangedMode = (PropertyChangedModes)
                      Enum.Parse(typeof(PropertyChangedModes), tmp);
                    _propertyChangedModeSet = true;
                }
                return _propertyChangedMode;
            }
            set
            {
                _propertyChangedMode = value;
                _propertyChangedModeSet = true;
            }
        }

        /// <summary>
        /// Enum representing the serialization formatters
        /// supported by CSLA .NET.
        /// </summary>
        public enum SerializationFormatters
        {
            /// <summary>
            /// Use the standard Microsoft .NET
            /// <see cref="BinaryFormatter"/>.
            /// </summary>
            BinaryFormatter,
            /// <summary>
            /// Use the Microsoft .NET 3.0
            /// <see cref="System.Runtime.Serialization.NetDataContractSerializer">
            /// NetDataContractSerializer</see> provided as part of WCF.
            /// </summary>
            NetDataContractSerializer
        }

        /// <summary>
        /// Enum representing the locations code can execute.
        /// </summary>
        public enum ExecutionLocations
        {
            /// <summary>
            /// The code is executing on the client.
            /// </summary>
            Client,
            /// <summary>
            /// The code is executing on the application server.
            /// </summary>
            Server,
            /// <summary>
            /// The code is executing on the Silverlight client.
            /// </summary>
            Silverlight
        }

        /// <summary>
        /// Enum representing the way in which CSLA .NET
        /// should raise PropertyChanged events.
        /// </summary>
        public enum PropertyChangedModes
        {
            /// <summary>
            /// Raise PropertyChanged events as required
            /// by Windows Forms data binding.
            /// </summary>
            Windows,
            /// <summary>
            /// Raise PropertyChanged events as required
            /// by XAML data binding in WPF.
            /// </summary>
            Xaml
        }

        #endregion

        #region In-Memory Settings

        private static ExecutionLocations _executionLocation =
          ExecutionLocations.Client;

        /// <summary>
        /// 返回程序代码的执行位置.        
        /// </summary>
        public static ExecutionLocations ExecutionLocation
        {
            get { return _executionLocation; }
        }

        internal static void SetExecutionLocation(ExecutionLocations location)
        {
            _executionLocation = location;
        }

        /// <summary>
        /// The default RuleSet name 
        /// </summary>
        public const string DefaultRuleSet = "default";

        /// <summary>
        /// Gets or sets the RuleSet name to use for static HasPermission calls.
        /// </summary>
        /// <value>The rule set.</value>
        public static string RuleSet
        {
            get
            {
                var ruleSet = ApplicationContext.ClientContext["__ruleSet"] as string;
                return string.IsNullOrEmpty(ruleSet) ? ApplicationContext.DefaultRuleSet : ruleSet;
            }
            set
            {
                ApplicationContext.ClientContext["__ruleSet"] = value;
            }
        }
        #endregion

        #region Logical Execution Location

        /// <summary>
        /// 程序运行位置,根据程序部署的情况设置程序运行位置.
        /// 如果是数据操作,将会设置为Server.其他情况是Client.
        /// </summary>
        public enum LogicalExecutionLocations
        {
            /// <summary>
            /// The code is executing on the client.
            /// </summary>
            Client,
            /// <summary>
            /// The code is executing on the server.  This inlcudes
            /// Local mode execution
            /// </summary>
            Server
        }

        /// <summary>
        /// 取得逻辑层运行的位置.
        /// </summary>
        public static LogicalExecutionLocations LogicalExecutionLocation
        {
            get
            {
                if (LocalContext.Contains("__logicalExecutionLocation"))
                    return (LogicalExecutionLocations)LocalContext["__logicalExecutionLocation"];
                else
                    return LogicalExecutionLocations.Client;
            }
        }

        internal static void SetLogicalExecutionLocation(LogicalExecutionLocations location)
        {
            LocalContext["__logicalExecutionLocation"] = location;
        }

        #endregion

        #region Default context manager
        /// <summary>
        /// 默认的上下文信息管理,默认为Winform环境中
        /// </summary>
        private class ApplicationContextManager : IContextManager
        {
            private const string _localContextName = "Csla.LocalContext";
            private const string _clientContextName = "Csla.ClientContext";
            private const string _globalContextName = "Csla.GlobalContext";
            private static IPrincipal _principal;

            public bool IsValid
            {
                get { return true; }
            }

            public IPrincipal GetUser()
            {
                //IPrincipal current = Thread.CurrentPrincipal;
                //return current;
                IPrincipal current;
                if (System.Windows.Application.Current != null)
                {
                    if (_principal == null)
                    {
                        if (ApplicationContext.AuthenticationType != "Windows")
                            _principal = new Wcdf.Security.UnauthenticatedPrincipal();
                        else
                            _principal = new WindowsPrincipal(WindowsIdentity.GetCurrent());
                    }
                    current = _principal;
                }
                else
                    current = Thread.CurrentPrincipal;
                return current;
            }

            public void SetUser(IPrincipal principal)
            {
                //Thread.CurrentPrincipal = principal;
                if (System.Windows.Application.Current != null)
                    _principal = principal;
                Thread.CurrentPrincipal = principal;
            }

            public ContextDictionary GetLocalContext()
            {
                LocalDataStoreSlot slot = Thread.GetNamedDataSlot(_localContextName);
                return (ContextDictionary)Thread.GetData(slot);
            }

            public void SetLocalContext(ContextDictionary localContext)
            {
                LocalDataStoreSlot slot = Thread.GetNamedDataSlot(_localContextName);
                Thread.SetData(slot, localContext);
            }

            public ContextDictionary GetClientContext()
            {
                if (ApplicationContext.ExecutionLocation == ExecutionLocations.Client)
                {
                    return (ContextDictionary)AppDomain.CurrentDomain.GetData(_clientContextName);
                }
                else
                {
                    LocalDataStoreSlot slot = Thread.GetNamedDataSlot(_clientContextName);
                    return (ContextDictionary)Thread.GetData(slot);
                }
            }

            public void SetClientContext(ContextDictionary clientContext)
            {
                if (ApplicationContext.ExecutionLocation == ExecutionLocations.Client)
                {
                    AppDomain.CurrentDomain.SetData(_clientContextName, clientContext);
                }
                else
                {
                    LocalDataStoreSlot slot = Thread.GetNamedDataSlot(_clientContextName);
                    Thread.SetData(slot, clientContext);
                }
            }

            public ContextDictionary GetGlobalContext()
            {
                LocalDataStoreSlot slot = Thread.GetNamedDataSlot(_globalContextName);
                return (ContextDictionary)Thread.GetData(slot);
            }

            public void SetGlobalContext(ContextDictionary globalContext)
            {
                LocalDataStoreSlot slot = Thread.GetNamedDataSlot(_globalContextName);
                Thread.SetData(slot, globalContext);
            }
        }

        #endregion
    }
}
