﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Wcdf.Properties;
using Wcdf.Core;
using System.ComponentModel;
using System.Reflection;
using Wcdf.Reflection;
using System.Linq.Expressions;
using Wcdf.Data;

namespace Wcdf
{
    /// <summary>
    /// 这是大部分业务对象需要继承的类.    
    /// </summary>
    /// <remarks>
    /// <para>
    /// 该类是框架的核心类.建立业务对象,请继承该类.    
    /// </para>
    /// </remarks>
    /// <typeparam name="T">Type of the business object being defined.</typeparam>
    [Serializable()]
    public abstract class BusinessBase<T> :
      Core.BusinessBase, Core.ISavable where T : BusinessBase<T>
    {        

        #region System.Object Overrides

        /// <summary>
        /// 返回对象标识<see cref="GetIdValue"/>的字符串形式        
        /// </summary>
        public override string ToString()
        {
            object id = GetIdValue();
            if (id == null)
                return base.ToString();
            else
                return id.ToString();
        }

        #endregion

        #region Clone

        /// <summary>
        /// 对象复制
        /// </summary>
        /// <returns>
        /// A new object containing the exact data of the original object.
        /// </returns>
        public T Clone()
        {
            return (T)GetClone();
        }

        #endregion

        #region DataAccess
        [RunLocal]
        protected virtual void DataPortal_Create(object criteria)
        {
            //return default(T);
            BusinessRules.CheckRules();
        }

        protected virtual void DataPortal_Fetch(object criteria)
        {
            DataAccess dal = new DataAccess(typeof(T),this.GetDataLocation());
            dal.Fetch(this, criteria);
        }

        protected virtual void DataPortal_Update()
        {
            DataAccess dal = new DataAccess(typeof(T), this.GetDataLocation());
            dal.Update(this);
            this.FieldManager.UpdateChildren(null);
        }

        protected virtual void DataPortal_Delete(object criteria)
        {
            DataAccess dal = new DataAccess(typeof(T), this.GetDataLocation());
            dal.Delete<T>(criteria);
        }
        /// <summary>
        /// Saves the object to the database.
        /// </summary>
        /// <remarks>
        /// <para>
        /// Calling this method starts the save operation, causing the object
        /// to be inserted, updated or deleted within the database based on the
        /// object's current state.
        /// </para><para>
        /// If <see cref="Core.BusinessBase.IsDeleted" /> is <see langword="true"/>
        /// the object will be deleted. Otherwise, if <see cref="Core.BusinessBase.IsNew" /> 
        /// is <see langword="true"/> the object will be inserted. 
        /// Otherwise the object's data will be updated in the database.
        /// </para><para>
        /// All this is contingent on <see cref="Core.BusinessBase.IsDirty" />. If
        /// this value is <see langword="false"/>, no data operation occurs. 
        /// It is also contingent on <see cref="Core.BusinessBase.IsValid" />. 
        /// If this value is <see langword="false"/> an
        /// exception will be thrown to indicate that the UI attempted to save an
        /// invalid object.
        /// </para><para>
        /// It is important to note that this method returns a new version of the
        /// business object that contains any data updated during the save operation.
        /// You MUST update all object references to use this new version of the
        /// business object in order to have access to the correct object data.
        /// </para><para>
        /// You can override this method to add your own custom behaviors to the save
        /// operation. For instance, you may add some security checks to make sure
        /// the user can save the object. If all security checks pass, you would then
        /// invoke the base Save method via <c>base.Save()</c>.
        /// </para>
        /// </remarks>
        /// <returns>A new object containing the saved values.</returns>
        public virtual T Save()
        {          
            T result;
            OnBeforeSave();
            //if (this.IsChild)
            //    throw new NotSupportedException(Resources.NoSaveChildException);
            if (EditLevel > 0)
                throw new Rules.ValidationException(Resources.NoSaveEditingException);
            if (!IsValid && !IsDeleted)
                throw new Rules.ValidationException(Resources.NoSaveInvalidException);
            if (IsBusy)
                throw new Rules.ValidationException(Resources.BusyObjectsMayNotBeSaved);
            if (IsDirty)
                result = (T)DataPortal.Update(this);
            else
                result = (T)this;
            //保存成功后清除脏数据
            this.MarkOld();
            OnSaved(result, null, null);
            return result;
        }

        /// <summary>
        /// Saves the object to the database, forcing
        /// IsNew to <see langword="false"/> and IsDirty to True.
        /// </summary>
        /// <param name="forceUpdate">
        /// If <see langword="true"/>, triggers overriding IsNew and IsDirty. 
        /// If <see langword="false"/> then it is the same as calling Save().
        /// </param>
        /// <returns>A new object containing the saved values.</returns>
        /// <remarks>
        /// This overload is designed for use in web applications
        /// when implementing the Update method in your 
        /// data wrapper object.
        /// </remarks>
        public T Save(bool forceUpdate)
        {
            if (forceUpdate && IsNew)
            {
                // mark the object as old - which makes it
                // not dirty
                MarkOld();
                // now mark the object as dirty so it can save
                MarkDirty(true);
            }
            return this.Save();
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        public void BeginSave()
        {
            BeginSave(null);
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        /// <param name="userState">User state data.</param>
        public void BeginSave(object userState)
        {
            BeginSave(false, null, userState);
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        /// <param name="handler">
        /// Method called when the operation is complete.
        /// </param>
        public void BeginSave(EventHandler<SavedEventArgs> handler)
        {
            BeginSave(false, handler, null);
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        /// <param name="forceUpdate">
        /// If <see langword="true"/>, triggers overriding IsNew and IsDirty. 
        /// If <see langword="false"/> then it is the same as calling Save().
        /// </param>
        /// <param name="handler">
        /// Method called when the operation is complete.
        /// </param>
        /// <param name="userState">User state data.</param>
        public virtual void BeginSave(bool forceUpdate, EventHandler<SavedEventArgs> handler, object userState)
        {
            if (forceUpdate && IsNew)
            {
                // mark the object as old - which makes it
                // not dirty
                MarkOld();
                // now mark the object as dirty so it can save
                MarkDirty(true);
            }
            //if (this.IsChild)
            //{
            //    NotSupportedException error = new NotSupportedException(Resources.NoSaveChildException);
            //    OnSaved(null, error, userState);
            //    if (handler != null)
            //        handler(this, new SavedEventArgs(null, error, userState));
            //}
            //else
            if (EditLevel > 0)
            {
                Rules.ValidationException error = new Rules.ValidationException(Resources.NoSaveEditingException);
                OnSaved(null, error, userState);
                if (handler != null)
                    handler(this, new SavedEventArgs(null, error, userState));
            }
            else if (!IsValid && !IsDeleted)
            {
                Rules.ValidationException error = new Rules.ValidationException(Resources.NoSaveInvalidException);
                OnSaved(null, error, userState);
                if (handler != null)
                    handler(this, new SavedEventArgs(null, error, userState));
            }
            else if (IsBusy)
            {
                Rules.ValidationException error = new Rules.ValidationException(Resources.BusyObjectsMayNotBeSaved);
                OnSaved(null, error, userState);
                if (handler != null)
                    handler(this, new SavedEventArgs(null, error, userState));
            }
            else
            {
                if (IsDirty)
                {
                    MarkBusy();
                    if (userState == null)
                    {
                        DataPortal.BeginUpdate<T>(this, (o, e) =>
                        {
                            T result = e.Object;
                            OnSaved(result, e.Error, e.UserState);
                            if (handler != null)
                                handler(result, new SavedEventArgs(result, e.Error, e.UserState));
                            if (result != null)
                            {
                                //正确返回数据后,标记为OLD;
                                MarkOld();
                            }
                            MarkIdle();
                        });
                    }
                    else
                    {
                        DataPortal.BeginUpdate<T>(this, (o, e) =>
                        {
                            T result = e.Object;
                            OnSaved(result, e.Error, e.UserState);
                            if (handler != null)
                                handler(result, new SavedEventArgs(result, e.Error, e.UserState));
                            if (result != null)
                            {
                                //正确返回数据后,标记为OLD;
                                MarkOld();
                            }
                            MarkIdle();
                        }, userState);
                    }
                }
                else
                {
                    OnSaved((T)this, null, userState);
                    if (handler != null)
                        handler(this, new SavedEventArgs(this, null, userState));
                }
            }
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        /// <param name="forceUpdate">
        /// If <see langword="true"/>, triggers overriding IsNew and IsDirty. 
        /// If <see langword="false"/> then it is the same as calling Save().
        /// </param>
        /// <remarks>
        /// This overload is designed for use in web applications
        /// when implementing the Update method in your 
        /// data wrapper object.
        /// </remarks>
        public void BeginSave(bool forceUpdate)
        {
            this.BeginSave(false, null);
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        /// <param name="forceUpdate">
        /// If <see langword="true"/>, triggers overriding IsNew and IsDirty. 
        /// If <see langword="false"/> then it is the same as calling Save().
        /// </param>
        /// <param name="handler">
        /// Delegate reference to a callback handler that will
        /// be invoked when the async operation is complete.
        /// </param>
        /// <remarks>
        /// This overload is designed for use in web applications
        /// when implementing the Update method in your 
        /// data wrapper object.
        /// </remarks>
        public void BeginSave(bool forceUpdate, EventHandler<SavedEventArgs> handler)
        {
            this.BeginSave(forceUpdate, handler, null);
        }

        /// <summary>
        /// Saves the object to the database, forcing
        /// IsNew to <see langword="false"/> and IsDirty to True.
        /// </summary>
        /// <param name="handler">
        /// Delegate reference to a callback handler that will
        /// be invoked when the async operation is complete.
        /// </param>
        /// <param name="userState">User state data.</param>
        /// <remarks>
        /// This overload is designed for use in web applications
        /// when implementing the Update method in your 
        /// data wrapper object.
        /// </remarks>
        public void BeginSave(EventHandler<SavedEventArgs> handler, object userState)
        {
            this.BeginSave(false, handler, userState);
        }

        #endregion

        #region ISavable Members

        object Wcdf.Core.ISavable.Save()
        {
            return Save();
        }

        object Wcdf.Core.ISavable.Save(bool forceUpdate)
        {
            return Save(forceUpdate);
        }

        void Wcdf.Core.ISavable.SaveComplete(object newObject)
        {
            OnSaved((T)newObject, null, null);
        }

        [NonSerialized]
        [NotUndoable]
        private EventHandler<SavedEventArgs> _nonSerializableSavedHandlers;
        [NotUndoable]
        private EventHandler<SavedEventArgs> _serializableSavedHandlers;

        /// <summary>
        /// 保存方法结束后调用的事件.
        /// </summary>
        /// <remarks>
        /// 当事件方法如果为私有的方法.一般为客户端应用,则不进行序列化.
        /// </remarks>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design",
          "CA1062:ValidateArgumentsOfPublicMethods")]
        public event EventHandler<SavedEventArgs> Saved
        {
            add
            {
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableSavedHandlers = (EventHandler<SavedEventArgs>)
                      System.Delegate.Combine(_serializableSavedHandlers, value);
                else
                    _nonSerializableSavedHandlers = (EventHandler<SavedEventArgs>)
                      System.Delegate.Combine(_nonSerializableSavedHandlers, value);
            }
            remove
            {
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableSavedHandlers = (EventHandler<SavedEventArgs>)
                      System.Delegate.Remove(_serializableSavedHandlers, value);
                else
                    _nonSerializableSavedHandlers = (EventHandler<SavedEventArgs>)
                      System.Delegate.Remove(_nonSerializableSavedHandlers, value);
            }
        }

        /// <summary>
        /// 触发Saved事件,该事件在对象Save方法执行后执行.
        /// 并提供新对象的引用.       
        /// </summary>
        /// <param name="newObject">The new object instance.</param>
        /// <param name="e">Exception that occurred during operation.</param>
        /// <param name="userState">User state object.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void OnSaved(T newObject, Exception e, object userState)
        {
            MarkIdle();
            SavedEventArgs args = new SavedEventArgs(newObject, e, userState);
            if (_nonSerializableSavedHandlers != null)
                _nonSerializableSavedHandlers.Invoke(this, args);
            if (_serializableSavedHandlers != null)
                _serializableSavedHandlers.Invoke(this, args);
        }

        [NonSerialized]
        [NotUndoable]
        private EventHandler<BeforeSaveEventArgs> _nonSerializableBeforeSavedHandlers;
        [NotUndoable]
        private EventHandler<BeforeSaveEventArgs> _serializableBeforeSavedHandlers;

        /// <summary>
        /// 保存方法前调用的事件.
        /// </summary>
        /// <remarks>
        /// 当事件方法如果为私有的方法.一般为客户端应用,则不进行序列化.
        /// </remarks>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design",
          "CA1062:ValidateArgumentsOfPublicMethods")]
        public event EventHandler<BeforeSaveEventArgs> BeforeSave
        {
            add
            {
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableBeforeSavedHandlers = (EventHandler<BeforeSaveEventArgs>)
                      System.Delegate.Combine(_serializableBeforeSavedHandlers, value);
                else
                    _nonSerializableBeforeSavedHandlers = (EventHandler<BeforeSaveEventArgs>)
                      System.Delegate.Combine(_nonSerializableBeforeSavedHandlers, value);
            }
            remove
            {
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableBeforeSavedHandlers = (EventHandler<BeforeSaveEventArgs>)
                      System.Delegate.Remove(_serializableSavedHandlers, value);
                else
                    _nonSerializableBeforeSavedHandlers = (EventHandler<BeforeSaveEventArgs>)
                      System.Delegate.Remove(_nonSerializableSavedHandlers, value);
            }
        }

        /// <summary>
        /// 触发Saved事件,该事件在对象Save方法执行后执行.
        /// 并提供新对象的引用.       
        /// </summary>
        /// <param name="newObject">The new object instance.</param>
        /// <param name="e">Exception that occurred during operation.</param>
        /// <param name="userState">User state object.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void OnBeforeSave()
        {
            MarkIdle();
            BeforeSaveEventArgs args = new BeforeSaveEventArgs();
            if (_nonSerializableBeforeSavedHandlers != null)
                _nonSerializableBeforeSavedHandlers.Invoke(this, args);
            if (_serializableBeforeSavedHandlers != null)
                _serializableBeforeSavedHandlers.Invoke(this, args);
        }
        #endregion

        #region  Register Properties

        /// <summary>
        /// Indicates that the specified property belongs
        /// to the business object type.
        /// </summary>
        /// <typeparam name="P">
        /// Type of property.
        /// </typeparam>
        /// <param name="info">
        /// PropertyInfo object for the property.
        /// </param>
        /// <returns>
        /// The provided IPropertyInfo object.
        /// </returns>
        protected static PropertyInfo<P> RegisterProperty<P>(PropertyInfo<P> info)
        {
            return Core.FieldManager.PropertyInfoManager.RegisterProperty<P>(typeof(T), info);
        }

        /// <summary>
        /// Indicates that the specified property belongs
        /// to the business object type.
        /// </summary>
        /// <typeparam name="P">Type of property</typeparam>
        /// <param name="propertyLambdaExpression">Property Expression</param>
        /// <returns></returns>
        protected static PropertyInfo<P> RegisterProperty<P>(Expression<Func<T, object>> propertyLambdaExpression)
        {
            PropertyInfo reflectedPropertyInfo = Reflect<T>.GetProperty(propertyLambdaExpression);

            return RegisterProperty(Core.FieldManager.PropertyInfoFactory.Factory.Create<P>(typeof(T), reflectedPropertyInfo.Name));
        }

        /// <summary>
        /// Indicates that the specified property belongs
        /// to the business object type.
        /// </summary>
        /// <typeparam name="P">Type of property</typeparam>
        /// <param name="propertyLambdaExpression">Property Expression</param>
        /// <param name="relationship">Relationship with
        /// referenced object.</param>
        /// <returns></returns>
        protected static PropertyInfo<P> RegisterProperty<P>(Expression<Func<T, object>> propertyLambdaExpression, RelationshipTypes relationship)
        {
            PropertyInfo reflectedPropertyInfo = Reflect<T>.GetProperty(propertyLambdaExpression);

            return RegisterProperty(Core.FieldManager.PropertyInfoFactory.Factory.Create<P>(typeof(T), reflectedPropertyInfo.Name, reflectedPropertyInfo.Name, relationship));
        }

        /// <summary>
        /// Indicates that the specified property belongs
        /// to the business object type.
        /// </summary>
        /// <typeparam name="P">Type of property</typeparam>
        /// <param name="propertyLambdaExpression">Property Expression</param>
        /// <param name="friendlyName">Friendly description for a property to be used in databinding</param>
        /// <returns></returns>
        protected static PropertyInfo<P> RegisterProperty<P>(Expression<Func<T, object>> propertyLambdaExpression, string friendlyName)
        {
            PropertyInfo reflectedPropertyInfo = Reflect<T>.GetProperty(propertyLambdaExpression);

            return RegisterProperty(Core.FieldManager.PropertyInfoFactory.Factory.Create<P>(typeof(T), reflectedPropertyInfo.Name, friendlyName));
        }

        /// <summary>
        /// Indicates that the specified property belongs
        /// to the business object type.
        /// </summary>
        /// <typeparam name="P">Type of property</typeparam>
        /// <param name="propertyLambdaExpression">Property Expression</param>
        /// <param name="friendlyName">Friendly description for a property to be used in databinding</param>
        /// <param name="relationship">Relationship with
        /// referenced object.</param>
        /// <returns></returns>
        protected static PropertyInfo<P> RegisterProperty<P>(Expression<Func<T, object>> propertyLambdaExpression, string friendlyName, RelationshipTypes relationship)
        {
            PropertyInfo reflectedPropertyInfo = Reflect<T>.GetProperty(propertyLambdaExpression);

            return RegisterProperty(Core.FieldManager.PropertyInfoFactory.Factory.Create<P>(typeof(T), reflectedPropertyInfo.Name, friendlyName, relationship));
        }

        /// <summary>
        /// Indicates that the specified property belongs
        /// to the business object type.
        /// </summary>
        /// <typeparam name="P">Type of property</typeparam>
        /// <param name="propertyLambdaExpression">Property Expression</param>
        /// <param name="friendlyName">Friendly description for a property to be used in databinding</param>
        /// <param name="defaultValue">Default Value for the property</param>
        /// <returns></returns>
        protected static PropertyInfo<P> RegisterProperty<P>(Expression<Func<T, object>> propertyLambdaExpression, string friendlyName, P defaultValue)
        {
            PropertyInfo reflectedPropertyInfo = Reflect<T>.GetProperty(propertyLambdaExpression);

            return RegisterProperty(Core.FieldManager.PropertyInfoFactory.Factory.Create<P>(typeof(T), reflectedPropertyInfo.Name, friendlyName, defaultValue));
        }

        /// <summary>
        /// Indicates that the specified property belongs
        /// to the business object type.
        /// </summary>
        /// <typeparam name="P">Type of property</typeparam>
        /// <param name="propertyLambdaExpression">Property Expression</param>
        /// <param name="friendlyName">Friendly description for a property to be used in databinding</param>
        /// <param name="defaultValue">Default Value for the property</param>
        /// <param name="relationship">Relationship with
        /// referenced object.</param>
        /// <returns></returns>
        protected static PropertyInfo<P> RegisterProperty<P>(Expression<Func<T, object>> propertyLambdaExpression, string friendlyName, P defaultValue, RelationshipTypes relationship)
        {
            PropertyInfo reflectedPropertyInfo = Reflect<T>.GetProperty(propertyLambdaExpression);

            return RegisterProperty(Core.FieldManager.PropertyInfoFactory.Factory.Create<P>(typeof(T), reflectedPropertyInfo.Name, friendlyName, defaultValue, relationship));
        }

        #endregion

        #region Register Methods

        /// <summary>
        /// Registers a method for use in Authorization.
        /// </summary>
        /// <param name="methodLambdaExpression">The method lambda expression.</param>
        /// <returns></returns>
        protected static MethodInfo RegisterMethod(Expression<Action<T>> methodLambdaExpression)
        {
            System.Reflection.MethodInfo reflectedMethodInfo = Reflect<T>.GetMethod(methodLambdaExpression);

            return RegisterMethod(typeof(T), reflectedMethodInfo.Name);
        }

        #endregion

        #region Tool Method
        /// <summary>
        /// 判断字段是否发生变更
        /// </summary>
        /// <param name="propertyLambdaExpression"></param>
        /// <returns></returns>
        public bool IsFieldDirty(Expression<Func<T,object>> propertyLambdaExpression)
        {            
            PropertyInfo reflectedPropertyInfo = Reflect<T>.GetProperty(propertyLambdaExpression);
            var pro = this.FieldManager.GetRegisteredProperties().Find(o => o.Name == reflectedPropertyInfo.Name);
            return this.FieldManager.IsFieldDirty(pro);
        }
        #endregion
    }
}
