﻿using System;
using System.ComponentModel;
using Wcdf.Serialization;
using Wcdf.Core;
using Wcdf.Properties;
using System.Collections.Generic;
using System.Collections.Specialized;
using Wcdf.Serialization.Mobile;
using Wcdf.Server;
using System.Linq;
using Wcdf.Data;

namespace Wcdf
{
    /// <summary>
    /// 业务类集合类基类.
    /// </summary>
    /// <typeparam name="T">定义的集合类型.</typeparam>
    /// <typeparam name="C">集合类包含的子类类型.</typeparam>
    /// <remarks>集合不适合Winform方式的绑定,因为无法在SL继承IBindingList接口.
    /// 如果需要在Winform中获得绑定支持,请用BusinessBindingListBase</remarks>    
#if TESTING
  [System.Diagnostics.DebuggerStepThrough]
#endif
    [Serializable]
    public abstract class BusinessListBase<T, C> :
#if SILVERLIGHT
      ExtendedBindingList<C>,
#else
 ObservableBindingList<C>,
#endif
 IEditableCollection, Core.IUndoableObject, ICloneable,
        ISavable, Core.IParent, Server.IDataPortalTarget,
        INotifyBusy
        where T : BusinessListBase<T, C>
        where C : Core.IEditableBusinessObject
    {
        /// <summary>
        /// Creates an instance of the object.
        /// </summary>
        protected BusinessListBase()
        {
            AllowNew = true;
        }

        #region Data Location
        /// <summary>
        /// 对象的数据存储位置，如果是数据库，指定服务端连接字符配置
        /// </summary>
        /// <returns></returns>
        public virtual string GetDataLocation()
        {
            return null;
        }
        #endregion

        #region ICloneable

        object ICloneable.Clone()
        {
            return GetClone();
        }

        /// <summary>
        /// Creates a clone of the object.
        /// </summary>
        /// <returns>A new object containing the exact data of the original object.</returns>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual object GetClone()
        {
            return Core.ObjectCloner.Clone(this);
        }

        /// <summary>
        /// Creates a clone of the object.
        /// </summary>
        /// <returns>A new object containing the exact data of the original object.</returns>
        public T Clone()
        {
            return (T)GetClone();
        }

        #endregion

        #region Delete and Undelete child

        private MobileList<C> _deletedList;

        /// <summary>
        /// A collection containing all child objects marked
        /// for deletion.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage(
          "Microsoft.Design", "CA1002:DoNotExposeGenericLists")]
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected MobileList<C> DeletedList
        {
            get
            {
                if (_deletedList == null)
                    _deletedList = new MobileList<C>();
                return _deletedList;
            }
        }

        private void DeleteChild(C child)
        {
            // set child edit level
            Core.UndoableBase.ResetChildEditLevel(child, this.EditLevel, false);

            // TODO: Implement i4o
            // remove from the index
            //RemoveIndexItem(child);
            // remove from the position map
            //RemoveFromMap(child);

            // mark the object as deleted
            child.DeleteChild();
            // and add it to the deleted collection for storage
            DeletedList.Add(child);
        }

        private void UnDeleteChild(C child)
        {
            // since the object is no longer deleted, remove it from
            // the deleted collection
            DeletedList.Remove(child);

            // we are inserting an _existing_ object so
            // we need to preserve the object's editleveladded value
            // because it will be changed by the normal add process
            int saveLevel = child.EditLevelAdded;

            // TODO: implement i4o
            //InsertIndexItem(child);

            Add(child);
            child.EditLevelAdded = saveLevel;
        }

        /// <summary>
        /// Returns <see langword="true"/> if the internal deleted list
        /// contains the specified child object.
        /// </summary>
        /// <param name="item">Child object to check.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public bool ContainsDeleted(C item)
        {
            return DeletedList.Contains(item);
        }

        #endregion

        #region Begin/Cancel/ApplyEdit

        /// <summary>
        /// Starts a nested edit on the object.
        /// </summary>
        /// <remarks>
        /// <para>
        /// When this method is called the object takes a snapshot of
        /// its current state (the values of its variables). This snapshot
        /// can be restored by calling <see cref="CancelEdit" />
        /// or committed by calling <see cref="ApplyEdit" />.
        /// </para><para>
        /// This is a nested operation. Each call to BeginEdit adds a new
        /// snapshot of the object's state to a stack. You should ensure that 
        /// for each call to BeginEdit there is a corresponding call to either 
        /// CancelEdit or ApplyEdit to remove that snapshot from the stack.
        /// </para><para>
        /// See Chapters 2 and 3 for details on n-level undo and state stacking.
        /// </para><para>
        /// This method triggers the copying of all child object states.
        /// </para>
        /// </remarks>
        public void BeginEdit()
        {
            if (this.IsChild)
                throw new NotSupportedException(Resources.NoBeginEditChildException);

            CopyState(this.EditLevel + 1);
        }

        /// <summary>
        /// Cancels the current edit process, restoring the object's state to
        /// its previous values.
        /// </summary>
        /// <remarks>
        /// Calling this method causes the most recently taken snapshot of the 
        /// object's state to be restored. This resets the object's values
        /// to the point of the last <see cref="BeginEdit" />
        /// call.
        /// <para>
        /// This method triggers an undo in all child objects.
        /// </para>
        /// </remarks>
        public void CancelEdit()
        {
            if (this.IsChild)
                throw new NotSupportedException(Resources.NoCancelEditChildException);

            UndoChanges(this.EditLevel - 1);
        }

        /// <summary>
        /// Commits the current edit process.
        /// </summary>
        /// <remarks>
        /// Calling this method causes the most recently taken snapshot of the 
        /// object's state to be discarded, thus committing any changes made
        /// to the object's state since the last 
        /// <see cref="BeginEdit" /> call.
        /// <para>
        /// This method triggers an <see cref="Core.BusinessBase.ApplyEdit"/>
        ///  in all child objects.
        /// </para>
        /// </remarks>
        public void ApplyEdit()
        {
            if (this.IsChild)
                throw new NotSupportedException(Resources.NoApplyEditChildException);

            AcceptChanges(this.EditLevel - 1);
        }

        void Core.IParent.ApplyEditChild(Core.IEditableBusinessObject child)
        {
            EditChildComplete(child);
        }

        /// <summary>
        /// Override this method to be notified when a child object's
        /// <see cref="Core.BusinessBase.ApplyEdit" /> method has
        /// completed.
        /// </summary>
        /// <param name="child">The child object that was edited.</param>
        protected virtual void EditChildComplete(Core.IEditableBusinessObject child)
        {

            // do nothing, we don't really care
            // when a child has its edits applied
        }

        #endregion

        #region Insert, Remove, Clear

#if SILVERLIGHT
    /// <summary>
    /// Override this method to create a new object that is added
    /// to the collection. 
    /// </summary>
    protected override void  AddNewCore()
    {
      var item = DataPortal.CreateChild<C>();
      Add(item);
      OnAddedNew(item);
    }
#else
        /// <summary>
        /// Override this method to create a new object that is added
        /// to the collection. 
        /// </summary>
        protected override C AddNewCore()
        {
            //var item = DataPortal.CreateChild<C>();
            var item = DataPortal.Create<C>();
            (item as IDataPortalTarget).MarkAsChild();
            (item as IDataPortalTarget).MarkNew();
            //var item = default(C);
            Add(item);
            return item;
        }
#endif

        /// <summary>
        /// 将子对象加入到当前列表中.
        /// </summary>
        /// <param name="child"></param>
        /// <remarks>该操作只是纯碎的列表操作.被未将子对象进行业务校验.
        /// </remarks>
        void IEditableCollection.AddChild(IEditableBusinessObject child)
        {
            Add((C)child);
        }

        /// <summary>
        /// 从列表中移除子对象.        
        /// </summary>
        /// <param name="child">The child object to remove.</param>
        /// 
        void IEditableCollection.RemoveChild(IEditableBusinessObject child)
        {
            Remove((C)child);
        }

        object IEditableCollection.GetDeletedList()
        {
            return DeletedList;
        }

        /// <summary>
        /// This method is called by a child object when it
        /// wants to be removed from the collection.
        /// </summary>
        /// <param name="child">The child object to remove.</param>
        void Core.IParent.RemoveChild(Wcdf.Core.IEditableBusinessObject child)
        {
            Remove((C)child);
        }

        /// <summary>
        /// Sets the edit level of the child object as it is added.
        /// </summary>
        /// <param name="index">Index of the item to insert.</param>
        /// <param name="item">Item to insert.</param>
        protected override void InsertItem(int index, C item)
        {
            // set parent reference
            item.SetParent(this);
            // set child edit level
            Core.UndoableBase.ResetChildEditLevel(item, this.EditLevel, false);
            // when an object is inserted we assume it is
            // a new object and so the edit level when it was
            // added must be set
            item.EditLevelAdded = _editLevel;
            base.InsertItem(index, item);
        }

        /// <summary>
        /// Marks the child object for deletion and moves it to
        /// the collection of deleted objects.
        /// </summary>
        /// <param name="index">Index of the item to remove.</param>
        protected override void RemoveItem(int index)
        {
            // when an object is 'removed' it is really
            // being deleted, so do the deletion work
            C child = this[index];
            bool oldRaiseListChangedEvents = this.RaiseListChangedEvents;
            try
            {
                this.RaiseListChangedEvents = false;
                base.RemoveItem(index);
            }
            finally
            {
                this.RaiseListChangedEvents = oldRaiseListChangedEvents;
            }
            if (!_completelyRemoveChild)
            {
                // the child shouldn't be completely removed,
                // so copy it to the deleted list
                DeleteChild(child);
            }
            if (RaiseListChangedEvents)
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, child, index));
        }

        /// <summary>
        /// Replaces the item at the specified index with
        /// the specified item, first moving the original
        /// item to the deleted list.
        /// </summary>
        /// <param name="index">The zero-based index of the item to replace.</param>
        /// <param name="item">
        /// The new value for the item at the specified index. 
        /// The value can be null for reference types.
        /// </param>
        /// <remarks></remarks>
        protected override void SetItem(int index, C item)
        {
            C child = default(C);
            if (!(ReferenceEquals((C)(this[index]), item)))
                child = this[index];
            // replace the original object with this new
            // object
            bool oldRaiseListChangedEvents = this.RaiseListChangedEvents;
            try
            {
                this.RaiseListChangedEvents = false;
                // set parent reference
                item.SetParent(this);
                // set child edit level
                Core.UndoableBase.ResetChildEditLevel(item, this.EditLevel, false);
                // reset EditLevelAdded 
                item.EditLevelAdded = this.EditLevel;
                // add to list
                base.SetItem(index, item);
            }
            finally
            {
                this.RaiseListChangedEvents = oldRaiseListChangedEvents;
            }
            if (child != null)
                DeleteChild(child);
            if (RaiseListChangedEvents)
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, child, index));
        }

        /// <summary>
        /// Clears the collection, moving all active
        /// items to the deleted list.
        /// </summary>
        protected override void ClearItems()
        {
            while (base.Count > 0) RemoveItem(0);
            //DeferredLoadIndexIfNotLoaded();
            //_indexSet.ClearIndexes();
            //DeferredLoadPositionMapIfNotLoaded();
            //_positionMap.ClearMap();
            base.ClearItems();
        }

        #endregion

        #region Edit level tracking

        // keep track of how many edit levels we have
        private int _editLevel;

        /// <summary>
        /// Returns the current edit level of the object.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected int EditLevel
        {
            get { return _editLevel; }
        }

        int Core.IUndoableObject.EditLevel
        {
            get
            {
                return this.EditLevel;
            }
        }

        #endregion

        #region N-level undo

        void Core.IUndoableObject.CopyState(int parentEditLevel, bool parentBindingEdit)
        {
            if (!parentBindingEdit)
                CopyState(parentEditLevel);
        }

        void Core.IUndoableObject.UndoChanges(int parentEditLevel, bool parentBindingEdit)
        {
            if (!parentBindingEdit)
                UndoChanges(parentEditLevel);
        }

        void Core.IUndoableObject.AcceptChanges(int parentEditLevel, bool parentBindingEdit)
        {
            if (!parentBindingEdit)
                AcceptChanges(parentEditLevel);
        }

        private void CopyState(int parentEditLevel)
        {
            if (this.EditLevel + 1 > parentEditLevel)
                throw new Core.UndoException(string.Format(Resources.EditLevelMismatchException, "CopyState"));

            // we are going a level deeper in editing
            _editLevel += 1;

            // JMC 6/24/08
            // This used to be a foreach loop but there appears to be a bug
            // in the silverlight runtime (SL2 B2) since calling foreach here will result
            // in SEHException with the error code -2147467259, or Error Unkown.
            // Iterating on this collection outside of this call will result in
            // behavior as expected but for some reason doing it here results in
            // an unknown exception.

            // cascade the call to all child objects
            for (int x = 0; x < this.Count; x++)
            {
                C child = this[x];
                child.CopyState(_editLevel, false);
            }

            // cascade the call to all deleted child objects
            foreach (C child in DeletedList)
                child.CopyState(_editLevel, false);
        }

        private bool _completelyRemoveChild;

        private void UndoChanges(int parentEditLevel)
        {
            C child;

            if (this.EditLevel - 1 != parentEditLevel)
                throw new Core.UndoException(string.Format(Resources.EditLevelMismatchException, "UndoChanges"));

            // we are coming up one edit level
            _editLevel -= 1;
            if (_editLevel < 0) _editLevel = 0;

            bool oldRLCE = this.RaiseListChangedEvents;
            this.RaiseListChangedEvents = false;
            try
            {
                // Cancel edit on all current items
                for (int index = Count - 1; index >= 0; index--)
                {
                    child = this[index];

                    //ACE: Important, make sure to remove the item prior to
                    //     it going through undo, otherwise, it will
                    //     incur a more expensive RemoveByReference operation
                    //DeferredLoadIndexIfNotLoaded();
                    //_indexSet.RemoveItem(child);

                    child.UndoChanges(_editLevel, false);

                    //ACE: Now that we have undone the changes, we can add the item
                    //     back in the index.
                    //_indexSet.InsertItem(child);

                    // if item is below its point of addition, remove
                    if (child.EditLevelAdded > _editLevel)
                    {
                        bool oldAllowRemove = this.AllowRemove;
                        try
                        {
                            this.AllowRemove = true;
                            _completelyRemoveChild = true;
                            //RemoveIndexItem(child);
                            RemoveAt(index);
                        }
                        finally
                        {
                            _completelyRemoveChild = false;
                            this.AllowRemove = oldAllowRemove;
                        }
                    }
                }

                // cancel edit on all deleted items
                for (int index = DeletedList.Count - 1; index >= 0; index--)
                {
                    child = DeletedList[index];
                    child.UndoChanges(_editLevel, false);
                    if (child.EditLevelAdded > _editLevel)
                    {
                        // if item is below its point of addition, remove
                        DeletedList.RemoveAt(index);
                    }
                    else
                    {
                        // if item is no longer deleted move back to main list
                        if (!child.IsDeleted) UnDeleteChild(child);
                    }
                }
            }
            finally
            {
                this.RaiseListChangedEvents = oldRLCE;
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
            }
        }

        private void AcceptChanges(int parentEditLevel)
        {
            if (this.EditLevel - 1 != parentEditLevel)
                throw new Core.UndoException(string.Format(Resources.EditLevelMismatchException, "AcceptChanges"));

            // we are coming up one edit level
            _editLevel -= 1;
            if (_editLevel < 0) _editLevel = 0;

            // cascade the call to all child objects
            foreach (C child in this)
            {
                child.AcceptChanges(_editLevel, false);
                // if item is below its point of addition, lower point of addition
                if (child.EditLevelAdded > _editLevel) child.EditLevelAdded = _editLevel;
            }

            // cascade the call to all deleted child objects
            for (int index = DeletedList.Count - 1; index >= 0; index--)
            {
                C child = DeletedList[index];
                child.AcceptChanges(_editLevel, false);
                // if item is below its point of addition, remove
                if (child.EditLevelAdded > _editLevel)
                    DeletedList.RemoveAt(index);
            }
        }

        #endregion

        #region Mobile Object overrides

        /// <summary>
        /// Method called by MobileFormatter when an object
        /// should serialize its data. The data should be
        /// serialized into the SerializationInfo parameter.
        /// </summary>
        /// <param name="info">
        /// Object to contain the serialized data.
        /// </param>
        protected override void OnGetState(SerializationInfo info)
        {
            info.AddValue("Csla.BusinessListBase._isChild", _isChild);
            info.AddValue("Csla.BusinessListBase._editLevel", _editLevel);
            base.OnGetState(info);
        }

        /// <summary>
        /// Method called by MobileFormatter when an object
        /// should be deserialized. The data should be
        /// deserialized from the SerializationInfo parameter.
        /// </summary>
        /// <param name="info">
        /// Object containing the serialized data.
        /// </param>
        protected override void OnSetState(SerializationInfo info)
        {
            _isChild = info.GetValue<bool>("Csla.BusinessListBase._isChild");
            _editLevel = info.GetValue<int>("Csla.BusinessListBase._editLevel");
            base.OnSetState(info);
        }

        /// <summary>
        /// Method called by MobileFormatter when an object
        /// should serialize its child references. The data should be
        /// serialized into the SerializationInfo parameter.
        /// </summary>
        /// <param name="info">
        /// Object to contain the serialized data.
        /// </param>
        /// <param name="formatter">
        /// Reference to the formatter performing the serialization.
        /// </param>
        protected override void OnGetChildren(Wcdf.Serialization.Mobile.SerializationInfo info, Wcdf.Serialization.Mobile.MobileFormatter formatter)
        {
            base.OnGetChildren(info, formatter);
            if (_deletedList != null)
            {
                var fieldManagerInfo = formatter.SerializeObject(_deletedList);
                info.AddChild("_deletedList", fieldManagerInfo.ReferenceId);
            }
        }

        /// <summary>
        /// Method called by MobileFormatter when an object
        /// should deserialize its child references. The data should be
        /// deserialized from the SerializationInfo parameter.
        /// </summary>
        /// <param name="info">
        /// Object containing the serialized data.
        /// </param>
        /// <param name="formatter">
        /// Reference to the formatter performing the deserialization.
        /// </param>
        protected override void OnSetChildren(Wcdf.Serialization.Mobile.SerializationInfo info, Wcdf.Serialization.Mobile.MobileFormatter formatter)
        {
            if (info.Children.ContainsKey("_deletedList"))
            {
                var childData = info.Children["_deletedList"];
                _deletedList = (MobileList<C>)formatter.GetObject(childData.ReferenceId);
            }
            base.OnSetChildren(info, formatter);
        }

        #endregion

        #region IsChild

        [NotUndoable()]
        private bool _isChild = false;

        /// <summary>
        /// Indicates whether this collection object is a child object.
        /// </summary>
        /// <returns>True if this is a child object.</returns>
        protected bool IsChild
        {
            get { return _isChild; }
        }

        /// <summary>
        /// Marks the object as being a child object.
        /// </summary>
        /// <remarks>
        /// <para>
        /// By default all business objects are 'parent' objects. This means
        /// that they can be directly retrieved and updated into the database.
        /// </para><para>
        /// We often also need child objects. These are objects which are contained
        /// within other objects. For instance, a parent Invoice object will contain
        /// child LineItem objects.
        /// </para><para>
        /// To create a child object, the MarkAsChild method must be called as the
        /// object is created. Please see Chapter 7 for details on the use of the
        /// MarkAsChild method.
        /// </para>
        /// </remarks>
        protected void MarkAsChild()
        {
            _isChild = true;
        }

        #endregion

        #region IsDirty, IsValid, IsSavable

        /// <summary>
        /// Gets a value indicating whether this object's data has been changed.
        /// </summary>
        bool Core.ITrackStatus.IsSelfDirty
        {
            get { return IsDirty; }
        }

        /// <summary>
        /// Gets a value indicating whether this object's data has been changed.
        /// </summary>
        public bool IsDirty
        {
            get
            {
                // any non-new deletions make us dirty
                foreach (C item in DeletedList)
                    if (!item.IsNew)
                        return true;

                // run through all the child objects
                // and if any are dirty then then
                // collection is dirty
                foreach (C child in this)
                    if (child.IsDirty)
                        return true;
                return false;
            }
        }

        bool Core.ITrackStatus.IsSelfValid
        {
            get { return IsSelfValid; }
        }

        /// <summary>
        /// Gets a value indicating whether this object is currently in
        /// a valid state (has no broken validation rules).
        /// </summary>
        protected virtual bool IsSelfValid
        {
            get { return IsValid; }
        }

        /// <summary>
        /// Gets a value indicating whether this object is currently in
        /// a valid state (has no broken validation rules).
        /// </summary>
        public virtual bool IsValid
        {
            get
            {
                // run through all the child objects
                // and if any are invalid then the
                // collection is invalid
                foreach (C child in this)
                    if (!child.IsValid)
                        return false;
                return true;
            }
        }

        /// <summary>
        /// Returns <see langword="true" /> if this object is both dirty and valid.
        /// </summary>
        /// <returns>A value indicating if this object is both dirty and valid.</returns>
        public virtual bool IsSavable
        {
            get
            {
                //bool auth = Wcdf.Rules.BusinessRules.HasPermission(Rules.AuthorizationActions.EditObject, this);
                return (IsDirty && IsValid && /*auth &&*/ !IsBusy);
            }
        }

        /// <summary>
        /// Gets the busy status for this object and its child objects.
        /// </summary>
        public override bool IsBusy
        {
            get
            {
                // any non-new deletions make us dirty
                foreach (C item in DeletedList)
                    if (item.IsBusy)
                        return true;

                // run through all the child objects
                // and if any are dirty then then
                // collection is dirty
                foreach (C child in this)
                    if (child.IsBusy)
                        return true;

                return false;
            }
        }

        #endregion

        #region  ITrackStatus

        bool Core.ITrackStatus.IsNew
        {
            get
            {
                return false;
            }
        }

        bool Core.ITrackStatus.IsDeleted
        {
            get
            {
                return false;
            }
        }

        #endregion

        #region Serialization Notification

        /// <summary>
        /// Reset parent references on deserialization.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected override void OnDeserialized()
        {
            base.OnDeserialized();
            foreach (Core.IEditableBusinessObject child in this)
                child.SetParent(this);

            foreach (Core.IEditableBusinessObject child in DeletedList)
                child.SetParent(this);
        }

        #endregion

        #region  Child Data Access

        /// <summary>
        /// Initializes a new instance of the object
        /// with default values.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
#if SILVERLIGHT
    public virtual void Child_Create()
#else
        protected virtual void Child_Create()
#endif
        { /* do nothing - list self-initializes */ }

        /// <summary>
        /// 单独更新子对象
        /// as necessary.
        /// </summary>
        /// <param name="parameters">
        /// Optional parameters passed to child update
        /// methods.
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
#if SILVERLIGHT
    public virtual void Child_Update(params object[] parameters)
    {
      var oldRLCE = this.RaiseListChangedEvents;
      this.RaiseListChangedEvents = false;
      try
      {
        ChildDataPortal<C> dp = new ChildDataPortal<C>();
        foreach (var child in DeletedList)
          dp.Update(child, parameters);

        DeletedList.Clear();

        foreach (var child in this)
          if (child.IsDirty) dp.Update(child, parameters);
      }
      finally
      {
        this.RaiseListChangedEvents = oldRLCE;
      }
    }
#else
        protected virtual void Child_Update(params object[] parameters)
        {
            var oldRLCE = this.RaiseListChangedEvents;
            this.RaiseListChangedEvents = false;
            try
            {
                foreach (var child in DeletedList)
                    if (!child.IsNew)
                    {
                        (child as ISavable).Save();
                        //DataPortal.Update(child);
                    }


                foreach (var child in this)
                    if (child.IsDirty)
                    {
                        (child as ISavable).Save();
                        //(child as IDataPortalTarget).MarkOld();
                        //DataPortal.Update(child);
                    }
                DeletedList.Clear();
            }
            finally
            {
                this.RaiseListChangedEvents = oldRLCE;
            }
        }
#endif

        #endregion

        #region Data Access

#if !SILVERLIGHT
        /// <summary>
        /// Saves the object to the database.
        /// </summary>
        /// <remarks>
        /// <para>
        /// Calling this method starts the save operation, causing the all child
        /// objects to be inserted, updated or deleted within the database based on the
        /// each object's current state.
        /// </para><para>
        /// All this is contingent on <see cref="IsDirty" />. If
        /// this value is <see langword="false"/>, no data operation occurs. 
        /// It is also contingent on <see cref="IsValid" />. If this value is 
        /// <see langword="false"/> an exception will be thrown to 
        /// indicate that the UI attempted to save an invalid object.
        /// </para><para>
        /// It is important to note that this method returns a new version of the
        /// business collection that contains any data updated during the save operation.
        /// You MUST update all object references to use this new version of the
        /// business collection in order to have access to the correct object data.
        /// </para><para>
        /// You can override this method to add your own custom behaviors to the save
        /// operation. For instance, you may add some security checks to make sure
        /// the user can save the object. If all security checks pass, you would then
        /// invoke the base Save method via <c>MyBase.Save()</c>.
        /// </para>
        /// </remarks>
        /// <returns>A new object containing the saved values.</returns>
        public virtual T Save()
        {
            T result;
            if (this.IsChild)
                throw new NotSupportedException(Resources.NoSaveChildException);

            if (_editLevel > 0)
                throw new Rules.ValidationException(Resources.NoSaveEditingException);

            if (!IsValid)
                throw new Rules.ValidationException(Resources.NoSaveInvalidException);

            if (IsBusy)
                // TODO: Review this resource text
                throw new Rules.ValidationException(Resources.BusyObjectsMayNotBeSaved);
            if (IsDirty && this.Parent == null)
            {
                //纯列表数据.可单独进行更新.
                this.Child_Update(null);
                result = (T)this;
            }
            else if (IsDirty)
            {
                result = (T)DataPortal.Update(this);
            }
            else
                result = (T)this;
            OnSaved(result, null, null);
            return result;
        }
#endif

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        public void BeginSave()
        {
            BeginSave(null, null);
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        /// <param name="userState">User state object.</param>
        public void BeginSave(object userState)
        {
            BeginSave(null, userState);
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        /// <param name="handler">
        /// Method called when the operation is complete.
        /// </param>
        public void BeginSave(EventHandler<SavedEventArgs> handler)
        {
            BeginSave(handler, null);
        }

        /// <summary>
        /// Starts an async operation to save the object to the database.
        /// </summary>
        /// <param name="handler">
        /// Method called when the operation is complete.
        /// </param>
        /// <param name="userState">User state object.</param>
        public virtual void BeginSave(EventHandler<SavedEventArgs> handler, object userState)
        {
            if (this.IsChild)
            {
                NotSupportedException error = new NotSupportedException(Resources.NoSaveChildException);
                OnSaved(null, error, userState);
                if (handler != null)
                    handler(this, new SavedEventArgs(null, error, userState));
            }
            else if (EditLevel > 0)
            {
                Rules.ValidationException error = new Rules.ValidationException(Resources.NoSaveEditingException);
                OnSaved(null, error, userState);
                if (handler != null)
                    handler(this, new SavedEventArgs(null, error, userState));
            }
            else if (!IsValid)
            {
                Rules.ValidationException error = new Rules.ValidationException(Resources.NoSaveEditingException);
                OnSaved(null, error, userState);
                if (handler != null)
                    handler(this, new SavedEventArgs(null, error, userState));
            }
            else if (IsBusy)
            {
                Rules.ValidationException error = new Rules.ValidationException(Resources.BusyObjectsMayNotBeSaved);
                OnSaved(null, error, userState);
                if (handler != null)
                    handler(this, new SavedEventArgs(null, error, userState));
            }
            else
            {
                if (IsDirty)
                {
                    if (this.Parent == null)
                    {
                        //单独更新暂时不支持异步,在这样处理后,列表更新后OnSave事件将失效.
                        try
                        {
                            this.Child_Update(null);
                            OnSaved((T)this, null, null);
                        }
                        catch (Exception ex)
                        {
                            //T result = (T)this;
                            OnSaved(null, ex, null);
                        }
                    }
                    else if (userState == null)
                    {
                        //单独更新暂时不支持异步,在这样处理后,列表更新后OnSave事件将失效.
                        //foreach (var item in this)
                        //{
                        //    if (item.IsDirty)
                        //    {
                        //        EventHandler<SavedEventArgs> handle = new EventHandler<SavedEventArgs>(
                        //            (o, e) =>
                        //            {
                        //                if (e.Error != null)
                        //                {
                        //                    OnSaved(null, e.Error, null);
                        //                }
                        //            });
                        //        dynamic a = item;
                        //        a.BeginSave(handle);
                        //    }
                        //    //DataPortal.BeginUpdate<T>(this, (o, e) =>
                        //    //{
                        //    //    T result = e.Object;
                        //    //    OnSaved(result, e.Error, e.UserState);
                        //    //    if (handler != null)
                        //    //        handler(result, new SavedEventArgs(result, e.Error, null));
                        //    //});
                        //}
                        //foreach (var item in this.DeletedList)
                        //{
                        //    if (!item.IsNew)
                        //    {
                        //        EventHandler<SavedEventArgs> handle = new EventHandler<SavedEventArgs>(
                        //            (o, e) =>
                        //            {
                        //                if (e.Error != null)
                        //                {
                        //                    OnSaved(null, e.Error, null);
                        //                }
                        //            });
                        //        dynamic a = item;
                        //        a.BeginSave(handle);
                        //    }
                        //}
                        //this.DeletedList.Clear();                     

                        DataPortal.BeginUpdate<T>(this, (o, e) =>
                        {
                            T result = e.Object;
                            OnSaved(result, e.Error, e.UserState);
                            if (handler != null)
                                handler(result, new SavedEventArgs(result, e.Error, null));
                        });
                    }
                    else
                    {
                        DataPortal.BeginUpdate<T>(this, (o, e) =>
                        {
                            T result = e.Object;
                            OnSaved(result, e.Error, e.UserState);
                            if (handler != null)
                                handler(result, new SavedEventArgs(result, e.Error, e.UserState));
                        }, userState);
                    }
                }
                else
                {
                    OnSaved((T)this, null, userState);
                    if (handler != null)
                        handler(this, new SavedEventArgs(this, null, userState));
                }
            }
        }

        /// <summary>
        /// Override this method to load a new business object with default
        /// values from the database.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1707:IdentifiersShouldNotContainUnderscores", MessageId = "Member")]
#if SILVERLIGHT
    [EditorBrowsable(EditorBrowsableState.Advanced)]
    public virtual void DataPortal_Create(Csla.DataPortalClient.LocalProxy<T>.CompletedHandler handler)
#else
        protected virtual void DataPortal_Create()
#endif
        { }

#if !SILVERLIGHT
        /// <summary>
        /// 从数据库获取对象集合方法，从写该方法以自定义数据获取方式        
        /// </summary>
        /// <param name="criteria">过滤条件.</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1707:IdentifiersShouldNotContainUnderscores", MessageId = "Member")]
        protected virtual void DataPortal_Fetch(object criteria)
        {
            DataAccess dal = new DataAccess(this.GetDataLocation());
            dal.Fetch<BusinessListBase<T, C>, C>(this, criteria);
            //foreach (var item in this)
            //{
            //    (item as IDataPortalTarget).MarkOld();
            //    (item as IDataPortalTarget).MarkAsChild();
            //}
        }

        /// <summary>
        /// 更新业务对象.该方法在数据端执行        
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1707:IdentifiersShouldNotContainUnderscores", MessageId = "Member")]
        protected virtual void DataPortal_Update()
        {
            DataAccess dal = new DataAccess(typeof(C), this.GetDataLocation());
            foreach (var child in this)
            {
                if (child.IsDirty)
                {
                    dal.Update(child);
                    //对于列表对象IDataPortalTartget的接口并没有实现,从效率上考虑.在Update方法中实现
                    (child as IDataPortalTarget).MarkOld();
                }
            }//更新删除列表
            foreach (var child in this.DeletedList)
            {
                if (!child.IsNew)
                {
                    dal.Update(child);
                }
                this.DeletedList.Clear();
            }
        }

        /// <summary>
        /// Override this method to allow immediate deletion of a business object.
        /// </summary>
        /// <param name="criteria">An object containing criteria values to identify the object.</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1707:IdentifiersShouldNotContainUnderscores", MessageId = "Member")]
        protected virtual void DataPortal_Delete(object criteria)
        {
            throw new NotSupportedException(Resources.DeleteNotSupportedException);
        }
#endif
        #endregion

        #region ISavable Members
#if SILVERLIGHT

    void ISavable.BeginSave()
    {
      BeginSave();
    }

    void ISavable.SaveComplete(object newObject, Exception error, object userState)
    {
      OnSaved((T)newObject, error, userState);
    }

    /// <summary>
    /// Event raised when an object has been saved.
    /// </summary>
    public event EventHandler<Csla.Core.SavedEventArgs> Saved;

    /// <summary>
    /// Raises the <see cref="Saved"/> event, indicating that the
    /// object has been saved, and providing a reference
    /// to the new object instance.
    /// </summary>
    /// <param name="newObject">The new object instance.</param>
    /// <param name="error">Exception object.</param>
    /// <param name="userState">User state object.</param>
    protected void OnSaved(T newObject, Exception error, object userState)
    {
      if (Saved != null)
        Saved(this, new SavedEventArgs(newObject, error, userState));
    }

#else
        object Wcdf.Core.ISavable.Save()
        {
            return Save();
        }

        object Wcdf.Core.ISavable.Save(bool forceUpdate)
        {
            return Save();
        }

        void Wcdf.Core.ISavable.SaveComplete(object newObject)
        {
            OnSaved((T)newObject, null, null);
        }

        [NonSerialized()]
        [NotUndoable]
        private EventHandler<Wcdf.Core.SavedEventArgs> _savedEvent;

        /// <summary>
        /// Event raised when an object has been saved.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design",
          "CA1062:ValidateArgumentsOfPublicMethods")]
        public event EventHandler<Wcdf.Core.SavedEventArgs> Saved
        {
            add
            {
                _savedEvent = (EventHandler<Wcdf.Core.SavedEventArgs>)
                  System.Delegate.Combine(_savedEvent, value);
            }
            remove
            {
                _savedEvent = (EventHandler<Wcdf.Core.SavedEventArgs>)
                  System.Delegate.Remove(_savedEvent, value);
            }
        }

        [NonSerialized()]
        [NotUndoable]
        private EventHandler<Wcdf.Core.BeforeSaveEventArgs> _beforeSaveEvent;

        /// <summary>
        /// Event raised when an object has been saved.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design",
          "CA1062:ValidateArgumentsOfPublicMethods")]
        public event EventHandler<Wcdf.Core.BeforeSaveEventArgs> BeforeSave
        {
            add
            {
                _beforeSaveEvent = (EventHandler<Wcdf.Core.BeforeSaveEventArgs>)
                  System.Delegate.Combine(_beforeSaveEvent, value);
            }
            remove
            {
                _beforeSaveEvent = (EventHandler<Wcdf.Core.BeforeSaveEventArgs>)
                  System.Delegate.Remove(_beforeSaveEvent, value);
            }
        }

        /// <summary>
        /// Raises the <see cref="Saved"/> event, indicating that the
        /// object has been saved, and providing a reference
        /// to the new object instance.
        /// </summary>
        /// <param name="newObject">The new object instance.</param>
        /// <param name="e">Execption that occurred during the operation.</param>
        /// <param name="userState">User state object.</param>
        [System.ComponentModel.EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void OnSaved(T newObject, Exception e, object userState)
        {
            Wcdf.Core.SavedEventArgs args = new Wcdf.Core.SavedEventArgs(newObject, e, userState);
            if (_savedEvent != null)
                _savedEvent.Invoke(this, args);
        }
#endif
        #endregion

        #region  Parent/Child link

        [NotUndoable(), NonSerialized()]
        private Core.IParent _parent;

        /// <summary>
        /// Provide access to the parent reference for use
        /// in child object code.
        /// </summary>
        /// <remarks>
        /// This value will be Nothing for root objects.
        /// </remarks>
        [Browsable(false)]
        [System.ComponentModel.DataAnnotations.Display(AutoGenerateField = false)]
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public Core.IParent Parent
        {
            get
            {
                return _parent;
            }
        }

        /// <summary>
        /// Used by BusinessListBase as a child object is 
        /// created to tell the child object about its
        /// parent.
        /// </summary>
        /// <param name="parent">A reference to the parent collection object.</param>
        protected virtual void SetParent(Core.IParent parent)
        {
            _parent = parent;
        }

        /// <summary>
        /// Used by BusinessListBase as a child object is 
        /// created to tell the child object about its
        /// parent.
        /// </summary>
        /// <param name="parent">A reference to the parent collection object.</param>
        void Core.IEditableCollection.SetParent(Core.IParent parent)
        {
            this.SetParent(parent);
        }

        #endregion

        #region IDataPortalTarget Members

        void IDataPortalTarget.CheckRules()
        { }

        void IDataPortalTarget.MarkAsChild()
        {
            this.MarkAsChild();
        }

        void IDataPortalTarget.MarkNew()
        { }

        void IDataPortalTarget.MarkOld()
        {
        }

        #endregion

    }
}
