﻿using System;
using System.ComponentModel;

namespace Wcdf.Core
{
    /// <summary>
    /// 该类以序列化安全方式继承INotifyPropertyChanged和INotifyPropertyChanging接口    
    /// </summary>
    [Serializable()]
    public abstract class BindableBase : MobileObject,
      System.ComponentModel.INotifyPropertyChanged,
      INotifyPropertyChanging
    {
        /// <summary>
        /// 创建对象实例.
        /// </summary>
        protected BindableBase()
        {

        }

        #region INotifyPropertyChanged
#if !SILVERLIGHT
        [NonSerialized()]
        private PropertyChangedEventHandler _nonSerializableChangedHandlers;
#endif
        private PropertyChangedEventHandler _serializableChangedHandlers;

        /// <summary>
        /// 实现安全序列化的PropertyChanged事件
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design",
          "CA1062:ValidateArgumentsOfPublicMethods")]
        public event PropertyChangedEventHandler PropertyChanged
        {
            add
            {
#if !SILVERLIGHT
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableChangedHandlers = (PropertyChangedEventHandler)
                      System.Delegate.Combine(_serializableChangedHandlers, value);
                else
                    _nonSerializableChangedHandlers = (PropertyChangedEventHandler)
                      System.Delegate.Combine(_nonSerializableChangedHandlers, value);
#else
                _serializableChangedHandlers = (PropertyChangedEventHandler)
                      System.Delegate.Combine(_serializableChangedHandlers, value);
#endif
            }
            remove
            {
#if !SILVERLIGHT
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableChangedHandlers = (PropertyChangedEventHandler)
                      System.Delegate.Remove(_serializableChangedHandlers, value);
                else
                    _nonSerializableChangedHandlers = (PropertyChangedEventHandler)
                      System.Delegate.Remove(_nonSerializableChangedHandlers, value);
#else
                _serializableChangedHandlers = (PropertyChangedEventHandler)
                      System.Delegate.Remove(_serializableChangedHandlers, value);
#endif
            }
        }

        /// <summary>
        ///调用该方法以引发所以属性的PropertyChanged事情
        /// </summary>
        /// <remarks>
        /// 该方法自动由MarkDirty引发.它将自动发出空字符串的PropertyChanged
        /// 事件,告知数据绑定以刷新所有属性.        
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnUnknownPropertyChanged()
        {
            OnPropertyChanged(string.Empty);
        }

        /// <summary>
        /// 触发指定属性的PropertyChanged事件.        
        /// </summary>
        /// <param name="propertyName">发生变化的属性.</param>
        /// <remarks>
        /// 该方法可由业务类中的属性调用来指示具体的属性变化.        
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnPropertyChanged(string propertyName)
        {
#if !SILVERLIGHT
            if (_nonSerializableChangedHandlers != null)
                _nonSerializableChangedHandlers.Invoke(this,
                  new PropertyChangedEventArgs(propertyName));
#endif
            if (_serializableChangedHandlers != null)
                _serializableChangedHandlers.Invoke(this,
                  new PropertyChangedEventArgs(propertyName));
        }

        /// <summary>
        /// 触发指定属性的PropertyChanged事件.        
        /// </summary>
        /// <param name="propertyInfo">发生变化的属性容器对象.</param>
        /// <remarks>
        ///该方法可由业务类中的属性调用来指示具体的属性变化.  
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnPropertyChanged(IPropertyInfo propertyInfo)
        {
            OnPropertyChanged(propertyInfo.Name);
        }

        #endregion

        #region INotifyPropertyChanging
#if !SILVERLIGHT
        [NonSerialized()]
        private PropertyChangingEventHandler _nonSerializableChangingHandlers;
#endif
        private PropertyChangingEventHandler _serializableChangingHandlers;

        /// <summary>
        /// 实现序列化安全的PropertyChanging事件
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design",
          "CA1062:ValidateArgumentsOfPublicMethods")]
        public event PropertyChangingEventHandler PropertyChanging
        {
            add
            {
#if !SILVERLIGHT
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableChangingHandlers = (PropertyChangingEventHandler)
                      System.Delegate.Combine(_serializableChangingHandlers, value);
                else
                    _nonSerializableChangingHandlers = (PropertyChangingEventHandler)
                      System.Delegate.Combine(_nonSerializableChangingHandlers, value);
#else
                _serializableChangingHandlers = (PropertyChangingEventHandler)
                      System.Delegate.Combine(_serializableChangingHandlers, value);
#endif
            }
            remove
            {
#if !SILVERLIGHT
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableChangingHandlers = (PropertyChangingEventHandler)
                      System.Delegate.Remove(_serializableChangingHandlers, value);
                else
                    _nonSerializableChangingHandlers = (PropertyChangingEventHandler)
                      System.Delegate.Remove(_nonSerializableChangingHandlers, value);
#else
                _serializableChangingHandlers = (PropertyChangingEventHandler)
                      System.Delegate.Remove(_serializableChangingHandlers, value);
#endif
            }
        }

        /// <summary>
        ///调用该方法以引发所以属性的PropertyChanging事情
        /// </summary>
        /// <remarks>
        /// 该方法自动由MarkDirty引发.它将自动发出空字符串的PropertyChanging
        /// 事件,告知数据绑定以刷新所有属性.        
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnUnknownPropertyChanging()
        {
            OnPropertyChanging(string.Empty);
        }

        /// <summary>
        /// 触发指定属性的PropertyChanging事件.        
        /// </summary>
        /// <param name="propertyName">发生变化的属性名称.</param>
        /// <remarks>
        /// 该方法可由业务类中的属性调用来指示具体的属性变化.        
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnPropertyChanging(string propertyName)
        {
#if !SILVERLIGHT
            if (_nonSerializableChangingHandlers != null)
                _nonSerializableChangingHandlers.Invoke(this,
                  new PropertyChangingEventArgs(propertyName));
#endif
            if (_serializableChangingHandlers != null)
                _serializableChangingHandlers.Invoke(this,
                  new PropertyChangingEventArgs(propertyName));
        }

        /// <summary>
        /// 触发指定属性的PropertyChanging事件.        
        /// </summary>
        /// <param name="propertyName">发生变化的属性容器对象.</param>
        /// <remarks>
        /// 该方法可由业务类中的属性调用来指示具体的属性变化.        
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnPropertyChanging(IPropertyInfo propertyInfo)
        {
            OnPropertyChanging(propertyInfo.Name);
        }

        #endregion
    }
}
