﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using Wcdf.Security;
using Wcdf.Serialization.Mobile;
using System.ComponentModel.DataAnnotations;
using Wcdf.Core.FieldManager;
using System.Collections.Specialized;
using Wcdf.Rules;
using Wcdf.Properties;
using System.Runtime.Serialization;
using System.Security;
using Wcdf.Core.LoadManager;
using Wcdf.Server;
using Wcdf.Reflection;

namespace Wcdf.Core
{
    /// <summary>
    /// 非泛型基类,大部分业务对象继承    
    /// </summary>
    [Serializable]
    public abstract class BusinessBase : UndoableBase,
    IEditableBusinessObject,
    IEditableObject,
    ICloneable,
    IAuthorizeReadWrite,
    IParent,
    IDataPortalTarget,
    IManageProperties,
    Rules.IHostRules,
    ICheckRules,
    INotifyBusy,
    INotifyChildChanged,
    ISerializationNotification
#if SILVERLIGHT
    ,INotifyDataErrorInfo
#endif
#if !SILVERLIGHT
    , IDataErrorInfo
#endif
    {
        #region Constructors

        /// <summary>
        /// 构造器.
        /// </summary>
        protected BusinessBase()
        {
            Initialize();
            InitializeBusinessRules();
        }

        #endregion

        #region Initialize

        /// <summary>
        /// Override this method to set up event handlers so user
        /// code in a partial class can respond to events raised by
        /// generated code.
        /// </summary>
        protected virtual void Initialize()
        { /* allows subclass to initialize events before any other activity occurs */ }

        #endregion

        #region Object ID Value

        /// <summary>
        /// 重写该方法以返回对象的唯一标识
        /// </summary>
        public virtual object GetIdValue()
        {
            return null;
        }

        #endregion

        #region Data Location
        /// <summary>
        /// 对应的数据对象名
        /// </summary>
        public virtual string GetDataObjectName()
        {
            return null;
        }
        /// <summary>
        /// 对象的数据存储位置，如果是数据库，指定服务端连接字符配置
        /// </summary>
        /// <returns></returns>
        public virtual string GetDataLocation()
        {
            return null;
        }
        #endregion

        #region Parent/Child link

        [NotUndoable]
        [NonSerialized]
        private IParent _parent;

        /// <summary>
        /// 提供对父对象访问        
        /// </summary>
        /// <remarks>
        /// 顶级对象时为NULL.
        /// </remarks>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public Core.IParent Parent
        {
            get { return _parent; }
        }

        /// <summary>
        /// 用于BusinessListBase中的子对象设置父对象方法.        
        /// </summary>
        /// <param name="parent">A reference to the parent collection object.</param>
        protected virtual void SetParent(Core.IParent parent)
        {
            _parent = parent;
        }

        #endregion

        #region IsNew, IsDeleted, IsDirty, IsSavable

        //对象状态跟踪 .新 删除,修改.
        private bool _isNew = true;
        private bool _isDeleted;
        private bool _isDirty = true;

        /// <summary>
        /// 标识对象是否为新建立的
        /// </summary>
        /// <remarks>        
        /// </remarks>
        /// <returns>对象是否为新创建的.</returns>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        public bool IsNew
        {
            get { return _isNew; }
        }

        /// <summary>
        /// 对象被标记为删除时返回<see langword="true" />         
        /// </summary>
        /// <remarks>        
        /// </remarks>
        /// <returns>A value indicating if this object is marked for deletion.</returns>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        public bool IsDeleted
        {
            get { return _isDeleted; }
        }

        /// <summary>
        /// 当对象数据，字段及子对象，发生变更或者删除，则返回<see langword="true" />        
        /// </summary>
        /// <remarks>
        /// <para>
        /// 当数据保存后，该标识将会清除，返回<see langword="false" />
        /// </para>
        /// </remarks>
        /// <returns>指标数据是否发生变更.</returns>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        public virtual bool IsDirty
        {
            get { return IsSelfDirty || (_fieldManager != null && FieldManager.IsDirty()); }
        }

        /// <summary>
        /// 如果对象自身数据发更变更，则返回 <see langword="true" />        
        /// </summary>
        /// <remarks>
        /// <para>
        /// 当数据保存后，该标识将会清除，返回<see langword="false" />
        /// </para>
        /// </remarks>
        /// <returns>指标对象本身的数据是否发生变更.</returns>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        public virtual bool IsSelfDirty
        {
            get { return _isDirty; }
        }

        /// <summary>
        /// 将对象标记为新建对象，并标记为脏数据，并带取消删除标识。        
        /// </summary>
        /// <remarks>
        /// <para>
        /// 在新创建对象时将会调用该方法。
        /// </para><para>
        /// 如果重写本方法时，请确保调用重写方法最后调用基类方法        
        /// </para>
        /// </remarks>
        protected virtual void MarkNew()
        {
            _isNew = true;
            _isDeleted = false;
            MarkDirty();
        }

        /// <summary>
        /// 将对象标记为旧对象，同时清除脏数据标记。        
        /// </summary>
        /// <remarks>
        /// <para>
        /// 一般以从数据库取数后,或者更新完成后调用该方法。                
        /// </para><para>
        /// 如果重写该方法，请确保调用重写方法最后调用基类方法。        
        /// </para>
        /// </remarks>
        protected virtual void MarkOld()
        {
            _isNew = false;
            MarkClean();
        }

        /// <summary>
        /// Marks an object for deletion. This also marks the object
        /// as being dirty.
        /// </summary>
        /// <remarks>
        /// You should call this method in your business logic in the
        /// case that you want to have the object deleted when it is
        /// saved to the database.
        /// </remarks>
        protected void MarkDeleted()
        {
            _isDeleted = true;
            MarkDirty();
        }

        /// <summary>
        /// 标识对象为脏数据，表示发生过数据变化.
        /// </summary>
        /// <remarks>
        /// <para>
        /// 一般在对象内部数据发生变化后，调用该方法。
        /// </para><para>
        /// 该方法通过框架内部进行标识，或者通过.Net控件的数据绑定事件触发方法中
        /// 会调用本方法，以控制数据变化标记的准确性。       
        /// </para>
        /// </remarks>
        protected void MarkDirty()
        {
            MarkDirty(false);
        }

        /// <summary>
        /// 标识为脏数据
        /// </summary>
        /// <param name="suppressEvent">
        /// <see langword="true" /> 引发PropertyChanged事件
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void MarkDirty(bool suppressEvent)
        {
            _isDirty = true;
            if (!suppressEvent)
                OnUnknownPropertyChanged();
        }

        /// <summary>
        /// 执行属性变更处理        
        /// </summary>
        /// <param name="property">Property that
        /// has changed.</param>
        /// <remarks>
        /// 该方法调用CheckRules(propertyName)方法,MarkDirty,OnPropertyChanged(propertyName)        
        /// </remarks>
        protected virtual void PropertyHasChanged(Wcdf.Core.IPropertyInfo property)
        {
            MarkDirty(true);
            var propertyNames = BusinessRules.CheckRules(property);
            if (ApplicationContext.PropertyChangedMode == ApplicationContext.PropertyChangedModes.Windows)
                OnPropertyChanged(property);
            else
                foreach (var name in propertyNames)
                    OnPropertyChanged(name);
        }

        /// <summary>
        /// 检查对象的规则，并通知可能已更改的属性的UI。
        /// </summary>
        protected virtual void CheckObjectRules()
        {
            var propertyNames = BusinessRules.CheckObjectRules();
            if (ApplicationContext.PropertyChangedMode == ApplicationContext.PropertyChangedModes.Windows)
            {
                OnUnknownPropertyChanged();
            }
            else
                foreach (var name in propertyNames)
                    OnPropertyChanged(name);
        }
        
        private void PropertyHasChanged(string propertyName)
        {
            PropertyHasChanged(FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).First());
        }

        /// <summary>
        /// 清除脏对象标记
        /// </summary>
        /// <remarks>
        /// 该方法一般自动调用，最好不要手动调用        
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void MarkClean()
        {
            _isDirty = false;
            if (_fieldManager != null)
                FieldManager.MarkClean();
            OnUnknownPropertyChanged();
        }

        /// <summary>
        /// Returns <see langword="true" /> if this object is both dirty and valid.
        /// </summary>
        /// <remarks>
        /// An object is considered dirty (changed) if 
        /// <see cref="P:Csla.BusinessBase.IsDirty" /> returns <see langword="true" />. It is
        /// considered valid if IsValid
        /// returns <see langword="true" />. The IsSavable property is
        /// a combination of these two properties. 
        /// </remarks>
        /// <returns>A value indicating if this object is both dirty and valid.</returns>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        public virtual bool IsSavable
        {
            get
            {
                bool auth;
                if (IsDeleted)
                    auth = Wcdf.Rules.BusinessRules.HasPermission(Rules.AuthorizationActions.DeleteObject, this);
                else if (IsNew)
                    auth = Wcdf.Rules.BusinessRules.HasPermission(Rules.AuthorizationActions.CreateObject, this);
                else
                    auth = Wcdf.Rules.BusinessRules.HasPermission(Rules.AuthorizationActions.EditObject, this);
                return (auth && IsDirty && IsValid && !IsBusy); 
            }
        }

        #endregion

        #region Authorization

        [NotUndoable]
        [NonSerialized]
        private Dictionary<string, bool> _readResultCache;
        [NotUndoable]
        [NonSerialized]
        private Dictionary<string, bool> _writeResultCache;
        [NotUndoable]
        [NonSerialized]
        private Dictionary<string, bool> _executeResultCache;
        [NotUndoable]
        [NonSerialized]
        private System.Security.Principal.IPrincipal _lastPrincipal;

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to read the
        /// calling property.
        /// </summary>
        /// <param name="property">Property to check.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public virtual bool CanReadProperty(Wcdf.Core.IPropertyInfo property)
        {
            bool result = true;

            VerifyAuthorizationCache();

            if (!_readResultCache.TryGetValue(property.Name, out result))
            {
                result = BusinessRules.HasPermission(AuthorizationActions.ReadProperty, property);
                if (BusinessRules.CachePermissionResult(AuthorizationActions.ReadProperty, property))
                {
                    // store value in cache
                    _readResultCache[property.Name] = result;
                }
            }

            return result;
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to read the
        /// calling property.
        /// </summary>
        /// <returns><see langword="true" /> if read is allowed.</returns>
        /// <param name="property">Property to read.</param>
        /// <param name="throwOnFalse">Indicates whether a negative
        /// result should cause an exception.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public bool CanReadProperty(Wcdf.Core.IPropertyInfo property, bool throwOnFalse)
        {
            bool result = CanReadProperty(property);
            if (throwOnFalse && result == false)
            {
                System.Security.SecurityException ex = new System.Security.SecurityException(
                  String.Format("{0} ({1})",
                  Resources.PropertyGetNotAllowed, property.Name));
                throw ex;
            }
            return result;
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to read the
        /// specified property.
        /// </summary>
        /// <param name="propertyName">Name of the property to read.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public bool CanReadProperty(string propertyName)
        {
            var prop = FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).FirstOrDefault();
            if (prop == null)
                throw new ArgumentOutOfRangeException("propertyName");
            return CanReadProperty(prop);
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to read the
        /// specified property.
        /// </summary>
        /// <param name="propertyName">Name of the property to read.</param>
        /// <param name="throwOnFalse">Indicates whether a negative
        /// result should cause an exception.</param>
        private bool CanReadProperty(string propertyName, bool throwOnFalse)
        {
            var prop = FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).FirstOrDefault();
            if (prop == null)
                throw new ArgumentOutOfRangeException("propertyName");
            return CanReadProperty(prop, throwOnFalse);
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to write the
        /// specified property.
        /// </summary>
        /// <param name="property">Property to write.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public virtual bool CanWriteProperty(Wcdf.Core.IPropertyInfo property)
        {
            bool result = true;

            VerifyAuthorizationCache();

            if (!_writeResultCache.TryGetValue(property.Name, out result))
            {
                result = BusinessRules.HasPermission(AuthorizationActions.WriteProperty, property);
                if (BusinessRules.CachePermissionResult(AuthorizationActions.WriteProperty, property))
                {
                    // store value in cache
                    _writeResultCache[property.Name] = result;
                }
            }
            return result;
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to write the
        /// calling property.
        /// </summary>
        /// <returns><see langword="true" /> if write is allowed.</returns>
        /// <param name="property">Property to write.</param>
        /// <param name="throwOnFalse">Indicates whether a negative
        /// result should cause an exception.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public bool CanWriteProperty(Wcdf.Core.IPropertyInfo property, bool throwOnFalse)
        {
            bool result = CanWriteProperty(property);
            if (throwOnFalse && result == false)
            {
                System.Security.SecurityException ex = new System.Security.SecurityException(
                  String.Format("{0} ({1})", Resources.PropertySetNotAllowed, property.Name));
                throw ex;
            }
            return result;
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to write the
        /// specified property.
        /// </summary>
        /// <param name="propertyName">Name of the property to write.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public bool CanWriteProperty(string propertyName)
        {
            var propertyInfo = FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).FirstOrDefault();
            if (propertyInfo == null)
                throw new ArgumentOutOfRangeException("propertyName");
            return CanWriteProperty(propertyInfo);
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to write the
        /// specified property.
        /// </summary>
        /// <param name="propertyName">Name of the property to write.</param>
        /// <param name="throwOnFalse">Indicates whether a negative
        /// result should cause an exception.</param>
        private bool CanWriteProperty(string propertyName, bool throwOnFalse)
        {
            var propertyInfo = FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).FirstOrDefault();
            if (propertyInfo == null)
                throw new ArgumentOutOfRangeException("propertyName");
            return CanWriteProperty(propertyInfo, throwOnFalse);
        }

        private void VerifyAuthorizationCache()
        {
            if (_readResultCache == null)
                _readResultCache = new Dictionary<string, bool>();
            if (_writeResultCache == null)
                _writeResultCache = new Dictionary<string, bool>();
            if (_executeResultCache == null)
                _executeResultCache = new Dictionary<string, bool>();
            if (!ReferenceEquals(Wcdf.ApplicationContext.User, _lastPrincipal))
            {
                // the principal has changed - reset the cache
                _readResultCache.Clear();
                _writeResultCache.Clear();
                _executeResultCache.Clear();
                _lastPrincipal = Wcdf.ApplicationContext.User;
            }
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to execute
        /// the specified method.
        /// </summary>
        /// <param name="method">Method to execute.</param>
        /// <returns><see langword="true" /> if execute is allowed.</returns>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public virtual bool CanExecuteMethod(Wcdf.Core.IMemberInfo method)
        {
            bool result = true;

            VerifyAuthorizationCache();

            if (!_executeResultCache.TryGetValue(method.Name, out result))
            {
                result = BusinessRules.HasPermission(AuthorizationActions.ExecuteMethod, method);
                if (BusinessRules.CachePermissionResult(AuthorizationActions.ExecuteMethod, method))
                {
                    // store value in cache
                    _executeResultCache[method.Name] = result;
                }
            }
            return result;
        }

        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to execute
        /// the specified method.
        /// </summary>
        /// <returns><see langword="true" /> if execute is allowed.</returns>
        /// <param name="method">Method to execute.</param>
        /// <param name="throwOnFalse">Indicates whether a negative
        /// result should cause an exception.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public bool CanExecuteMethod(Wcdf.Core.IMemberInfo method, bool throwOnFalse)
        {

            bool result = CanExecuteMethod(method);
            if (throwOnFalse && result == false)
            {
                System.Security.SecurityException ex =
                  new System.Security.SecurityException(string.Format("{0} ({1})", Properties.Resources.MethodExecuteNotAllowed, method.Name));
                throw ex;
            }
            return result;

        }


        /// <summary>
        /// Returns <see langword="true" /> if the user is allowed to execute
        /// the specified method.
        /// </summary>
        /// <param name="methodName">Name of the method to execute.</param>
        /// <returns><see langword="true" /> if execute is allowed.</returns>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public virtual bool CanExecuteMethod(string methodName)
        {
            return CanExecuteMethod(methodName, false);
        }

        private bool CanExecuteMethod(string methodName, bool throwOnFalse)
        {

            bool result = CanExecuteMethod(new MethodInfo(methodName));
            if (throwOnFalse && result == false)
            {
                System.Security.SecurityException ex = new System.Security.SecurityException(string.Format("{0} ({1})", Properties.Resources.MethodExecuteNotAllowed, methodName));
                throw ex;
            }
            return result;
        }

        #endregion

        #region System.ComponentModel.IEditableObject

        private bool _neverCommitted = true;
        [NotUndoable]
        private bool _disableIEditableObject;

        /// <summary>
        /// 获取或设置是否禁用IEditableObject接口.        
        /// </summary>
        /// <value>
        /// 默认值为False,不禁用.
        /// </value>
        /// <remarks>
        /// 如果禁用,在WinForm数据绑定时将不会自动调用BeginEdit, CancelEdit or
        /// ApplyEdit方法.你只能自动调用这些方法手动获取N层Undo行为.  
        /// 由IEditableObject接口引发的只会触发单层Undo.
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected bool DisableIEditableObject
        {
            get
            {
                return _disableIEditableObject;
            }
            set
            {
                _disableIEditableObject = value;
            }
        }

        /// <summary>
        /// 允许事件绑定开始编辑对象
        /// </summary>
        /// <remarks>
        /// 数据绑定将会多次调用该方法.只有第一次调用才触发,所以我们用额外的代码来
        /// 控制.避免多次执行        
        /// </remarks>
        void System.ComponentModel.IEditableObject.BeginEdit()
        {
            if (!_disableIEditableObject && !BindingEdit)
            {
                BindingEdit = true;
                BeginEdit();
            }
        }

        /// <summary>
        /// 允许事件绑定放弃正在编辑的对象
        /// </summary>
        /// <remarks>
        /// 数据绑定将会多次调用该方法.只有第一次调用才触发,所以我们用额外的代码来
        /// 控制.避免多次执行        
        /// </remarks>
        void System.ComponentModel.IEditableObject.CancelEdit()
        {
            if (!_disableIEditableObject && BindingEdit)
            {
                CancelEdit();
                BindingEdit = false;
                if (IsNew && _neverCommitted && EditLevel <= EditLevelAdded)
                {
                    // we're new and no EndEdit or ApplyEdit has ever been
                    // called on us, and now we've been cancelled back to
                    // where we were added so we should have ourselves
                    // removed from the parent collection
                    if (Parent != null)
                        Parent.RemoveChild(this);
                }
            }
        }

        /// <summary>
        /// 允许事件绑定提交正在编辑的对象
        /// </summary>
        /// <remarks>
        /// 数据绑定将会多次调用该方法.只有第一次调用才触发,所以我们用额外的代码来
        /// 控制.避免多次执行      
        /// </remarks>
        void System.ComponentModel.IEditableObject.EndEdit()
        {
            if (!_disableIEditableObject && BindingEdit)
            {
                ApplyEdit();
                BindingEdit = false;
            }
        }

        #endregion

        #region Begin/Cancel/ApplyEdit

        /// <summary>
        /// 开始编辑对象
        /// </summary>
        /// <remarks>
        /// <para>
        /// 当调用该方法时会生成对象的快照放入堆栈中,当调用CancelEdit时,用快照
        /// 还能对象.ApplyEdit提交更改.        
        /// </para><para>
        /// 
        /// This is a nested operation. Each call to BeginEdit adds a new
        /// snapshot of the object's state to a stack. You should ensure that 
        /// for each call to BeginEdit there is a corresponding call to either 
        /// CancelEdit or ApplyEdit to remove that snapshot from the stack.
        /// </para><para>
        /// See Chapters 2 and 3 for details on n-level undo and state stacking.
        /// </para>
        /// </remarks>
        public void BeginEdit()
        {
            CopyState(this.EditLevel + 1);
        }

        /// <summary>
        /// 放弃正在编辑处理,将对象返回到上一个版本.        
        /// </summary>
        /// <remarks>
        /// Calling this method causes the most recently taken snapshot of the 
        /// object's state to be restored. This resets the object's values
        /// to the point of the last BeginEdit call.
        /// </remarks>
        public void CancelEdit()
        {
            UndoChanges(this.EditLevel - 1);
        }

        /// <summary>
        /// Called when an undo operation has completed.
        /// </summary>
        /// <remarks> 
        /// This method resets the object as a result of
        /// deserialization and raises PropertyChanged events
        /// to notify data binding that the object has changed.
        /// </remarks>
        protected override void UndoChangesComplete()
        {
            BusinessRules.SetTarget(this);
            InitializeBusinessRules();
            OnUnknownPropertyChanged();
            base.UndoChangesComplete();
        }

        /// <summary>
        /// Commits the current edit process.
        /// </summary>
        /// <remarks>
        /// Calling this method causes the most recently taken snapshot of the 
        /// object's state to be discarded, thus committing any changes made
        /// to the object's state since the last BeginEdit call.
        /// </remarks>
        public void ApplyEdit()
        {
            _neverCommitted = false;
            AcceptChanges(this.EditLevel - 1);
            //Next line moved to IEditableObject.ApplyEdit 
            //BindingEdit = false;
        }

        /// <summary>
        /// Notifies the parent object (if any) that this
        /// child object's edits have been accepted.
        /// </summary>
        protected override void AcceptChangesComplete()
        {
            //BindingEdit = false;
            base.AcceptChangesComplete();

            // !!!! Will trigger Save here when using DynamicListBase template
            if (Parent != null)
                Parent.ApplyEditChild(this);
        }

        #endregion

        #region IsChild

        [NotUndoable]
        private bool _isChild;

        /// <summary>
        /// 返回 <see langword="true" /> 如果对象是子对象.
        /// </summary>
        protected internal bool IsChild
        {
            get { return _isChild; }
        }

        /// <summary>
        /// 将对象标记为子对象.一般如果是BusinessList,该方法将会被DataPortal中创建对象时调用.以标记为子对象.        
        /// </summary>
        protected void MarkAsChild()
        {
            _isChild = true;
        }

        #endregion

        #region Delete

        /// <summary>
        /// 将对象标记为删除.在保存时进行删除操作        
        /// </summary>
        /// <remarks>
        /// <para>
        /// CSLA .NET supports both immediate and deferred deletion of objects. This
        /// method is part of the support for deferred deletion, where an object
        /// can be marked for deletion, but isn't actually deleted until the object
        /// is saved to the database. This method is called by the UI developer to
        /// mark the object for deletion.
        /// </para><para>
        /// To 'undelete' an object, use n-level undo as discussed in Chapters 2 and 3.
        /// </para>
        /// </remarks>
        public virtual void Delete()
        {
            if (this.IsChild)
                throw new NotSupportedException(Resources.ChildDeleteException);

            MarkDeleted();
        }

        /// <summary>
        /// 由父对象调用.将子对象标记为删除.        
        /// </summary>
        internal void DeleteChild()
        {
            if (!this.IsChild)
                throw new NotSupportedException(Resources.NoDeleteRootException);

            BindingEdit = false;
            MarkDeleted();
        }

        #endregion

        #region Edit Level Tracking (child only)

        //当我们增加层数时,需要跟踪.这样当用户放弃,在这层以下的快照
        //就可以删除.        
        [NotUndoable]
        private int _editLevelAdded;

        /// <summary>
        /// 取得或者设置当前对象的编辑层.        
        /// </summary>
        /// <remarks>
        /// 允许集合对象使用所需的层数.                
        /// </remarks>
        internal int EditLevelAdded
        {
            get { return _editLevelAdded; }
            set { _editLevelAdded = value; }
        }

        int IUndoableObject.EditLevel
        {
            get
            {
                return this.EditLevel;
            }
        }

        #endregion

        #region ICloneable
       
        object ICloneable.Clone()
        {
            return GetClone();
        }

        /// <summary>
        /// 该方法为Deep Clone
        /// </summary>
        /// <returns>
        /// A new object containing the exact data of the original object.
        /// </returns>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual object GetClone()
        {
            return ObjectCloner.Clone(this);
        }

        #endregion

        #region BusinessRules, IsValid

#if SILVERLIGHT
    /// <summary>
    /// Event raised when validation is complete.
    /// </summary>
    public event EventHandler ValidationComplete;

    /// <summary>
    /// Raises the ValidationComplete event.
    /// </summary>
    [EditorBrowsable(EditorBrowsableState.Never)]
    protected virtual void OnValidationComplete()
    {
      if (ValidationComplete != null)
        ValidationComplete(this, EventArgs.Empty);
    }
#else
        [NonSerialized]
        [NotUndoable]
        private EventHandler _validationCompleteHandlers;

        /// <summary>
        /// Event raised when validation is complete.
        /// </summary>
        public event EventHandler ValidationComplete
        {
            add
            {
                _validationCompleteHandlers = (EventHandler)
                  System.Delegate.Combine(_validationCompleteHandlers, value);
            }
            remove
            {
                _validationCompleteHandlers = (EventHandler)
                  System.Delegate.Remove(_validationCompleteHandlers, value);
            }
        }

        /// <summary>
        /// Raises the ValidationComplete event
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected virtual void OnValidationComplete()
        {
            if (_validationCompleteHandlers != null)
                _validationCompleteHandlers(this, EventArgs.Empty);
        }
#endif

        private void InitializeBusinessRules()
        {
            var rules = BusinessRuleManager.GetRulesForType(this.GetType());
            if (!rules.Initialized)
                lock (rules)
                    if (!rules.Initialized)
                    {
                        rules.Initialized = true;

                        try
                        {
                            AddBusinessRules();
                        }
                        catch (Exception)
                        {
                            BusinessRuleManager.CleanupRulesForType(this.GetType());
                            throw;  // and rethrow exception
                        }
                    }
        }

        private Wcdf.Rules.BusinessRules _businessRules;

        /// <summary>
        /// Provides access to the broken rules functionality.
        /// </summary>
        /// <remarks>
        /// This property is used within your business logic so you can
        /// easily call the AddRule() method to associate business
        /// rules with your object's properties.
        /// </remarks>
        protected BusinessRules BusinessRules
        {
            get
            {
                if (_businessRules == null)
                    _businessRules = new BusinessRules(this);
                else if (_businessRules.Target == null)
                    _businessRules.SetTarget(this);
                return _businessRules;
            }
        }

        void IHostRules.RuleStart(IPropertyInfo property)
        {
            OnBusyChanged(new BusyChangedEventArgs(property.Name, true));
        }

        void IHostRules.RuleComplete(IPropertyInfo property)
        {
            OnPropertyChanged(property);
            OnBusyChanged(new BusyChangedEventArgs(property.Name, false));
        }

        void IHostRules.RuleComplete(string property)
        {
            OnPropertyChanged(property);
        }

        void Rules.IHostRules.AllRulesComplete()
        {
            OnValidationComplete();
        }

        /// <summary>
        /// Override this method in your business class to
        /// be notified when you need to set up shared 
        /// business rules.
        /// </summary>
        /// <remarks>
        /// This method is automatically called by CSLA .NET
        /// when your object should associate per-type 
        /// validation rules with its properties.
        /// </remarks>
        protected virtual void AddBusinessRules()
        {
            BusinessRules.AddDataAnnotations();
        }

        /// <summary>
        /// Returns <see langword="true" /> if the object 
        /// and its child objects are currently valid, 
        /// <see langword="false" /> if the
        /// object or any of its child objects have broken 
        /// rules or are otherwise invalid.
        /// </summary>
        /// <remarks>
        /// <para>
        /// By default this property relies on the underling BusinessRules
        /// object to track whether any business rules are currently broken for this object.
        /// </para><para>
        /// You can override this property to provide more sophisticated
        /// implementations of the behavior. For instance, you should always override
        /// this method if your object has child objects, since the validity of this object
        /// is affected by the validity of all child objects.
        /// </para>
        /// </remarks>
        /// <returns>A value indicating if the object is currently valid.</returns>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        public virtual bool IsValid
        {
            get { return IsSelfValid && (_fieldManager == null || FieldManager.IsValid()); }
        }

        /// <summary>
        /// Returns <see langword="true" /> if the object is currently 
        /// valid, <see langword="false" /> if the
        /// object has broken rules or is otherwise invalid.
        /// </summary>
        /// <remarks>
        /// <para>
        /// By default this property relies on the underling BusinessRules
        /// object to track whether any business rules are currently broken for this object.
        /// </para><para>
        /// You can override this property to provide more sophisticated
        /// implementations of the behavior. 
        /// </para>
        /// </remarks>
        /// <returns>A value indicating if the object is currently valid.</returns>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        public virtual bool IsSelfValid
        {
            get { return BusinessRules.IsValid; }
        }

        /// <summary>
        /// Provides access to the readonly collection of broken business rules
        /// for this object.
        /// </summary>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        public virtual Rules.BrokenRulesCollection BrokenRulesCollection
        {
            get { return BusinessRules.GetBrokenRules(); }
        }

        #endregion        

#if !SILVERLIGHT
        #region IDataErrorInfo
        //UI错误信息通知接口实现

        string IDataErrorInfo.Error
        {
            get
            {
                if (!IsSelfValid)
                    return BusinessRules.GetBrokenRules().ToString(
                      Wcdf.Rules.RuleSeverity.Error);
                else
                    return String.Empty;
            }
        }

        string IDataErrorInfo.this[string columnName]
        {
            get
            {
                string result = string.Empty;
                if (!IsSelfValid)
                {
                    Rules.BrokenRule rule =
                      BusinessRules.GetBrokenRules().GetFirstBrokenRule(columnName);
                    if (rule != null)
                        result = rule.Description;
                }
                return result;
            }
        }

        #endregion
#endif
        #region Serialization Notification

        void ISerializationNotification.Deserialized()
        {
            OnDeserializedHandler(new StreamingContext());
        }

#if !IOS
        [OnDeserialized]
#endif
        private void OnDeserializedHandler(StreamingContext context)
        {
            BusinessRules.SetTarget(this);
            if (_fieldManager != null)
                FieldManager.SetPropertyList(this.GetType());
            InitializeBusinessRules();
            FieldDataDeserialized();

            OnDeserialized(context);
        }

        /// <summary>
        /// This method is called on a newly deserialized object
        /// after deserialization is complete.
        /// </summary>
        /// <param name="context">Serialization context object.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnDeserialized(StreamingContext context)
        { }

        #endregion

        #region Bubbling event Hooks

        /// <summary>
        /// For internal use.
        /// </summary>
        /// <param name="child">Child object.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected void AddEventHooks(IBusinessObject child)
        {
            OnAddEventHooks(child);
        }

        /// <summary>
        /// Hook child object events.
        /// </summary>
        /// <param name="child">Child object.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected virtual void OnAddEventHooks(IBusinessObject child)
        {
            INotifyBusy busy = child as INotifyBusy;
            if (busy != null)
                busy.BusyChanged += Child_BusyChanged;

            INotifyUnhandledAsyncException unhandled = child as INotifyUnhandledAsyncException;
            if (unhandled != null)
                unhandled.UnhandledAsyncException += Child_UnhandledAsyncException;

            INotifyPropertyChanged pc = child as INotifyPropertyChanged;
            if (pc != null)
                pc.PropertyChanged += Child_PropertyChanged;

#if !SILVERLIGHT
            IBindingList bl = child as IBindingList;
            if (bl != null)
                bl.ListChanged += Child_ListChanged;
#endif

            INotifyCollectionChanged ncc = child as INotifyCollectionChanged;
            if (ncc != null)
                ncc.CollectionChanged += Child_CollectionChanged;

            INotifyChildChanged cc = child as INotifyChildChanged;
            if (cc != null)
                cc.ChildChanged += Child_Changed;
        }

        /// <summary>
        /// For internal use only.
        /// </summary>
        /// <param name="child">Child object.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected void RemoveEventHooks(IBusinessObject child)
        {
            OnRemoveEventHooks(child);
        }

        /// <summary>
        /// Unhook child object events.
        /// </summary>
        /// <param name="child">Child object.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected virtual void OnRemoveEventHooks(IBusinessObject child)
        {
            INotifyBusy busy = child as INotifyBusy;
            if (busy != null)
                busy.BusyChanged -= Child_BusyChanged;

            INotifyUnhandledAsyncException unhandled = child as INotifyUnhandledAsyncException;
            if (unhandled != null)
                unhandled.UnhandledAsyncException -= Child_UnhandledAsyncException;

            INotifyPropertyChanged pc = child as INotifyPropertyChanged;
            if (pc != null)
                pc.PropertyChanged -= Child_PropertyChanged;

#if !SILVERLIGHT
            IBindingList bl = child as IBindingList;
            if (bl != null)
                bl.ListChanged -= Child_ListChanged;
#endif

            INotifyCollectionChanged ncc = child as INotifyCollectionChanged;
            if (ncc != null)
                ncc.CollectionChanged -= Child_CollectionChanged;

            INotifyChildChanged cc = child as INotifyChildChanged;
            if (cc != null)
                cc.ChildChanged -= Child_Changed;
        }

        #endregion

        #region Busy / Unhandled exception bubbling

        private void Child_UnhandledAsyncException(object sender, ErrorEventArgs e)
        {
            OnUnhandledAsyncException(e);
        }

        private void Child_BusyChanged(object sender, BusyChangedEventArgs e)
        {
            OnBusyChanged(e);
        }

        #endregion

        #region IEditableBusinessObject Members

        int IEditableBusinessObject.EditLevelAdded
        {
            get
            {
                return this.EditLevelAdded;
            }
            set
            {
                this.EditLevelAdded = value;
            }
        }

        void IEditableBusinessObject.DeleteChild()
        {
            this.DeleteChild();
        }

        void IEditableBusinessObject.SetParent(IParent parent)
        {
            this.SetParent(parent);
        }

        #endregion

        #region Register Methods

        /// <summary>
        /// Indicates that the specified method belongs
        /// to the type.
        /// </summary>
        /// <param name="objectType">
        /// Type of object to which the method belongs.
        /// </param>
        /// <param name="info">
        /// IMemberInfo object for the property.
        /// </param>
        /// <returns>
        /// The provided IMemberInfo object.
        /// </returns>
        protected static Core.IMemberInfo RegisterMethod(Type objectType, IMemberInfo info)
        {
            var reflected = objectType.GetMethod(info.Name);
            if (reflected == null)
                throw new ArgumentException(string.Format(Resources.NoSuchMethod, info.Name), "info");
            return info;
        }

        /// <summary>
        /// Indicates that the specified method belongs
        /// to the type.
        /// </summary>
        /// <param name="objectType">
        /// Type of object to which the method belongs.
        /// </param>
        /// <param name="methodName">
        /// Name of the method.
        /// </param>
        /// <returns>
        /// The provided IMemberInfo object.
        /// </returns>
        protected static MethodInfo RegisterMethod(Type objectType, string methodName)
        {
            var info = new MethodInfo(methodName);
            RegisterMethod(objectType, info);
            return info;
        }

        #endregion

        #region  Register Properties

        /// <summary>
        /// 将指定的属性注册到缓存中.
        /// Indicates that the specified property belongs
        /// to the type.
        /// </summary>
        /// <typeparam name="P">
        /// Type of property.
        /// </typeparam>
        /// <param name="objectType">
        /// Type of object to which the property belongs.
        /// </param>
        /// <param name="info">
        /// PropertyInfo object for the property.
        /// </param>
        /// <returns>
        /// The provided IPropertyInfo object.
        /// </returns>
        protected static PropertyInfo<P> RegisterProperty<P>(Type objectType, PropertyInfo<P> info)
        {
            return Core.FieldManager.PropertyInfoManager.RegisterProperty<P>(objectType, info);
        }

        #endregion

        #region  Get Properties

        /// <summary>
        /// Gets a property's value, first checking authorization.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="field">
        /// The backing field for the property.</param>
        /// <param name="propertyName">
        /// The name of the property.</param>
        /// <param name="defaultValue">
        /// Value to be returned if the user is not
        /// authorized to read the property.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetProperty<P>(string propertyName, P field, P defaultValue)
        {
            return GetProperty<P>(propertyName, field, defaultValue, Security.NoAccessBehavior.SuppressException);
        }

        /// <summary>
        /// Gets a property's value, first checking authorization.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="field">
        /// The backing field for the property.</param>
        /// <param name="propertyName">
        /// The name of the property.</param>
        /// <param name="defaultValue">
        /// Value to be returned if the user is not
        /// authorized to read the property.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to read this property.</param>
        protected P GetProperty<P>(string propertyName, P field, P defaultValue, Security.NoAccessBehavior noAccess)
        {
            #region Check to see if the property is marked with RelationshipTypes.PrivateField

            var propertyInfo = FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).FirstOrDefault();
            if (propertyInfo == null)
                throw new ArgumentOutOfRangeException("propertyName");

            if ((propertyInfo.RelationshipType & RelationshipTypes.PrivateField) != RelationshipTypes.PrivateField)
                throw new InvalidOperationException(Resources.PrivateFieldException);

            #endregion

            if (_bypassPropertyChecks || CanReadProperty(propertyInfo, noAccess == Security.NoAccessBehavior.ThrowException))
                return field;

            return defaultValue;
        }

        /// <summary>
        /// Gets a property's value, first checking authorization.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="field">
        /// The backing field for the property.</param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetProperty<P>(PropertyInfo<P> propertyInfo, P field)
        {
            return GetProperty<P>(propertyInfo.Name, field, propertyInfo.DefaultValue, Security.NoAccessBehavior.SuppressException);
        }

        /// <summary>
        /// Gets a property's value, first checking authorization.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="field">
        /// The backing field for the property.</param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="defaultValue">
        /// Value to be returned if the user is not
        /// authorized to read the property.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to read this property.</param>
        protected P GetProperty<P>(PropertyInfo<P> propertyInfo, P field, P defaultValue, Security.NoAccessBehavior noAccess)
        {
            return GetProperty<P>(propertyInfo.Name, field, defaultValue, noAccess);
        }

        /// <summary>
        /// Gets a property's value as 
        /// a specified type, first checking authorization.
        /// </summary>
        /// <typeparam name="F">
        /// Type of the field.
        /// </typeparam>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="field">
        /// The backing field for the property.</param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetPropertyConvert<F, P>(PropertyInfo<F> propertyInfo, F field)
        {
            return Utilities.CoerceValue<P>(typeof(F), null, GetProperty<F>(propertyInfo.Name, field, propertyInfo.DefaultValue, Security.NoAccessBehavior.SuppressException));
        }

        /// <summary>
        /// Gets a property's value as a specified type, 
        /// first checking authorization.
        /// </summary>
        /// <typeparam name="F">
        /// Type of the field.
        /// </typeparam>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="field">
        /// The backing field for the property.</param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to read this property.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetPropertyConvert<F, P>(PropertyInfo<F> propertyInfo, F field, Security.NoAccessBehavior noAccess)
        {
            return Utilities.CoerceValue<P>(typeof(F), null, GetProperty<F>(propertyInfo.Name, field, propertyInfo.DefaultValue, noAccess));
        }

        /// <summary>
        /// Gets a property's managed field value, 
        /// first checking authorization.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetProperty<P>(PropertyInfo<P> propertyInfo)
        {
            return GetProperty<P>(propertyInfo, Security.NoAccessBehavior.SuppressException);
        }

        /// <summary>
        /// Gets a property's value from the list of 
        /// managed field values, first checking authorization,
        /// and converting the value to an appropriate type.
        /// </summary>
        /// <typeparam name="F">
        /// Type of the field.
        /// </typeparam>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetPropertyConvert<F, P>(PropertyInfo<F> propertyInfo)
        {
            return Utilities.CoerceValue<P>(typeof(F), null, GetProperty<F>(propertyInfo, Security.NoAccessBehavior.SuppressException));
        }

        /// <summary>
        /// Gets a property's value from the list of 
        /// managed field values, first checking authorization,
        /// and converting the value to an appropriate type.
        /// </summary>
        /// <typeparam name="F">
        /// Type of the field.
        /// </typeparam>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to read this property.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetPropertyConvert<F, P>(PropertyInfo<F> propertyInfo, Security.NoAccessBehavior noAccess)
        {
            return Utilities.CoerceValue<P>(typeof(F), null, GetProperty<F>(propertyInfo, noAccess));
        }

        /// <summary>
        /// Gets a property's value as a specified type, 
        /// first checking authorization.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to read this property.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetProperty<P>(PropertyInfo<P> propertyInfo, Security.NoAccessBehavior noAccess)
        {
            P result = default(P);
            if (_bypassPropertyChecks || CanReadProperty(propertyInfo, noAccess == Wcdf.Security.NoAccessBehavior.ThrowException))
                result = ReadProperty<P>(propertyInfo);
            else
                result = propertyInfo.DefaultValue;
            return result;
        }

        /// <summary>
        /// Gets a property's value as a specified type.
        /// </summary>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected object GetProperty(IPropertyInfo propertyInfo)
        {
            object result = null;
            if (_bypassPropertyChecks || CanReadProperty(propertyInfo, false))
            {
                // call ReadProperty (may be overloaded in actual class)
                result = ReadProperty(propertyInfo);
            }
            else
            {
                result = propertyInfo.DefaultValue;
            }
            return result;
        }

        /// <summary>
        /// Gets a property's managed field value, 
        /// first checking authorization.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// If the user is not authorized to read the property
        /// value, the defaultValue value is returned as a
        /// result.
        /// </remarks>
        protected P GetProperty<P>(IPropertyInfo propertyInfo)
        {
            return (P)GetProperty(propertyInfo);
        }

        #endregion

        #region  Read Properties

        /// <summary>
        /// Gets a property's value from the list of 
        /// managed field values, converting the 
        /// value to an appropriate type.
        /// </summary>
        /// <typeparam name="F">
        /// Type of the field.
        /// </typeparam>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        protected P ReadPropertyConvert<F, P>(PropertyInfo<F> propertyInfo)
        {
            return Utilities.CoerceValue<P>(typeof(F), null, ReadProperty<F>(propertyInfo));
        }

        /// <summary>
        /// Gets a property's value as a specified type.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        protected P ReadProperty<P>(PropertyInfo<P> propertyInfo)
        {
            if (((propertyInfo.RelationshipType & RelationshipTypes.LazyLoad) == RelationshipTypes.LazyLoad) && !FieldManager.FieldExists(propertyInfo))
                throw new InvalidOperationException(Resources.PropertyGetNotAllowed);

            P result = default(P);
            FieldManager.IFieldData data = FieldManager.GetFieldData(propertyInfo);
            if (data != null)
            {
                FieldManager.IFieldData<P> fd = data as FieldManager.IFieldData<P>;
                if (fd != null)
                    result = fd.Value;
                else
                    result = (P)data.Value;
            }
            else
            {
                result = propertyInfo.DefaultValue;
                FieldManager.LoadFieldData<P>(propertyInfo, result);
            }
            return result;
        }

        /// <summary>
        /// Gets a property's value.
        /// </summary>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        protected virtual object ReadProperty(IPropertyInfo propertyInfo)
        {
            if ((propertyInfo.RelationshipType == RelationshipTypes.LazyLoad) && !FieldManager.FieldExists(propertyInfo))
                throw new InvalidOperationException(Resources.PropertyGetNotAllowed);

            if ((propertyInfo.RelationshipType & RelationshipTypes.PrivateField) == RelationshipTypes.PrivateField)
            {
                using (BypassPropertyChecks)
                {
                    return MethodCaller.CallPropertyGetter(this, propertyInfo.Name);
                }
            }

            object result = null;
            var info = FieldManager.GetFieldData(propertyInfo);
            if (info != null)
            {
                result = info.Value;
            }
            else
            {
                result = propertyInfo.DefaultValue;
                FieldManager.LoadFieldData(propertyInfo, result);
            }

            return result;
        }

        #endregion

        #region  Set Properties

        /// <summary>
        /// Sets a property's backing field with the supplied
        /// value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <param name="field">
        /// A reference to the backing field for the property.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// If the user is not authorized to change the property, this
        /// overload throws a SecurityException.
        /// </remarks>
        protected void SetProperty<P>(PropertyInfo<P> propertyInfo, ref P field, P newValue)
        {
            SetProperty<P>(propertyInfo.Name, ref field, newValue, Security.NoAccessBehavior.ThrowException);
        }

        /// <summary>
        /// Sets a property's backing field with the supplied
        /// value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <param name="field">
        /// A reference to the backing field for the property.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <param name="propertyName">
        /// The name of the property.</param>
        /// <remarks>
        /// If the user is not authorized to change the property, this
        /// overload throws a SecurityException.
        /// </remarks>
        protected void SetProperty<P>(string propertyName, ref P field, P newValue)
        {
            SetProperty<P>(propertyName, ref field, newValue, Security.NoAccessBehavior.ThrowException);
        }

        /// <summary>
        /// Sets a property's backing field with the 
        /// supplied value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the field being set.
        /// </typeparam>
        /// <typeparam name="V">
        /// Type of the value provided to the field.
        /// </typeparam>
        /// <param name="field">
        /// A reference to the backing field for the property.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// If the user is not authorized to change the property, this
        /// overload throws a SecurityException.
        /// </remarks>
        protected void SetPropertyConvert<P, V>(PropertyInfo<P> propertyInfo, ref P field, V newValue)
        {
            SetPropertyConvert<P, V>(propertyInfo, ref field, newValue, Security.NoAccessBehavior.ThrowException);
        }

        /// <summary>
        /// Sets a property's backing field with the 
        /// supplied value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the field being set.
        /// </typeparam>
        /// <typeparam name="V">
        /// Type of the value provided to the field.
        /// </typeparam>
        /// <param name="field">
        /// A reference to the backing field for the property.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to change this property.</param>
        /// <remarks>
        /// If the field value is of type string, any incoming
        /// null values are converted to string.Empty.
        /// </remarks>
        protected void SetPropertyConvert<P, V>(PropertyInfo<P> propertyInfo, ref P field, V newValue, Security.NoAccessBehavior noAccess)
        {
            SetPropertyConvert<P, V>(propertyInfo.Name, ref field, newValue, noAccess);
        }

        /// <summary>
        /// Sets a property's backing field with the supplied
        /// value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <param name="field">
        /// A reference to the backing field for the property.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <param name="propertyName">
        /// The name of the property.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to change this property.</param>
        protected void SetProperty<P>(string propertyName, ref P field, P newValue, Security.NoAccessBehavior noAccess)
        {
            try
            {
                #region Check to see if the property is marked with RelationshipTypes.PrivateField

                var propertyInfo = FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).FirstOrDefault();
                if (propertyInfo == null)
                    throw new ArgumentOutOfRangeException("propertyName");

                if ((propertyInfo.RelationshipType & RelationshipTypes.PrivateField) != RelationshipTypes.PrivateField)
                    throw new InvalidOperationException(Resources.PrivateFieldException);

                #endregion

                if (_bypassPropertyChecks || CanWriteProperty(propertyInfo, noAccess == Security.NoAccessBehavior.ThrowException))
                {
                    bool doChange = false;
                    if (field == null)
                    {
                        if (newValue != null)
                            doChange = true;
                    }
                    else
                    {
                        if (typeof(P) == typeof(string) && newValue == null)
                            newValue = Utilities.CoerceValue<P>(typeof(string), field, string.Empty);
                        if (!field.Equals(newValue))
                            doChange = true;
                    }
                    if (doChange)
                    {
                        if (!_bypassPropertyChecks) OnPropertyChanging(propertyName);
                        field = newValue;
                        if (!_bypassPropertyChecks) PropertyHasChanged(propertyName);
                    }
                }
            }
            catch (SecurityException)
            {
                throw;
            }
            catch (Exception ex)
            {
                throw new PropertyLoadException(
                  string.Format(Resources.PropertyLoadException, propertyName, ex.Message, ex.Message), ex);
            }
        }

        /// <summary>
        /// Sets a property's backing field with the 
        /// supplied value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the field being set.
        /// </typeparam>
        /// <typeparam name="V">
        /// Type of the value provided to the field.
        /// </typeparam>
        /// <param name="field">
        /// A reference to the backing field for the property.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <param name="propertyName">
        /// The name of the property.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to change this property.</param>
        /// <remarks>
        /// If the field value is of type string, any incoming
        /// null values are converted to string.Empty.
        /// </remarks>
        protected void SetPropertyConvert<P, V>(string propertyName, ref P field, V newValue, Security.NoAccessBehavior noAccess)
        {
            try
            {
                #region Check to see if the property is marked with RelationshipTypes.PrivateField

                var propertyInfo = FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).FirstOrDefault();
                if (propertyInfo == null)
                    throw new ArgumentOutOfRangeException("propertyName");

                if ((propertyInfo.RelationshipType & RelationshipTypes.PrivateField) != RelationshipTypes.PrivateField)
                    throw new InvalidOperationException(Resources.PrivateFieldException);

                #endregion

                if (_bypassPropertyChecks || CanWriteProperty(propertyInfo, noAccess == Security.NoAccessBehavior.ThrowException))
                {
                    bool doChange = false;
                    if (field == null)
                    {
                        if (newValue != null)
                            doChange = true;
                    }
                    else
                    {
                        if (typeof(V) == typeof(string) && newValue == null)
                            newValue = Utilities.CoerceValue<V>(typeof(string), null, string.Empty);
                        if (!field.Equals(newValue))
                            doChange = true;
                    }
                    if (doChange)
                    {
                        if (!_bypassPropertyChecks) OnPropertyChanging(propertyName);
                        field = Utilities.CoerceValue<P>(typeof(V), field, newValue);
                        if (!_bypassPropertyChecks) PropertyHasChanged(propertyName);
                    }
                }
            }
            catch (SecurityException)
            {
                throw;
            }
            catch (Exception ex)
            {
                throw new PropertyLoadException(
                  string.Format(Properties.Resources.PropertyLoadException, propertyName, ex.Message), ex);
            }
        }

        /// <summary>
        /// Sets a property's managed field with the 
        /// supplied value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <typeparam name="P">Property type.</typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <remarks>
        /// If the user is not authorized to change the property, this
        /// overload throws a SecurityException.
        /// </remarks>
        protected void SetProperty<P>(PropertyInfo<P> propertyInfo, P newValue)
        {
            SetProperty<P>(propertyInfo, newValue, Security.NoAccessBehavior.ThrowException);
        }

        /// <summary>
        /// Sets a property's managed field with the 
        /// supplied value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <remarks>
        /// If the user is not authorized to change the property, this
        /// overload throws a SecurityException.
        /// </remarks>
        protected void SetPropertyConvert<P, F>(PropertyInfo<P> propertyInfo, F newValue)
        {
            SetPropertyConvert<P, F>(propertyInfo, newValue, Security.NoAccessBehavior.ThrowException);
        }

        /// <summary>
        /// Sets a property's managed field with the 
        /// supplied value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to change this property.</param>
        protected void SetPropertyConvert<P, F>(PropertyInfo<P> propertyInfo, F newValue, Security.NoAccessBehavior noAccess)
        {
            try
            {
                if (_bypassPropertyChecks || CanWriteProperty(propertyInfo, noAccess == Security.NoAccessBehavior.ThrowException))
                {
                    P oldValue = default(P);
                    var fieldData = FieldManager.GetFieldData(propertyInfo);
                    if (fieldData == null)
                    {
                        oldValue = propertyInfo.DefaultValue;
                        fieldData = FieldManager.LoadFieldData<P>(propertyInfo, oldValue);
                    }
                    else
                    {
                        var fd = fieldData as FieldManager.IFieldData<P>;
                        if (fd != null)
                            oldValue = fd.Value;
                        else
                            oldValue = (P)fieldData.Value;
                    }
                    if (typeof(F) == typeof(string) && newValue == null)
                        newValue = Utilities.CoerceValue<F>(typeof(string), null, string.Empty);
                    LoadPropertyValue<P>(propertyInfo, oldValue, Utilities.CoerceValue<P>(typeof(F), oldValue, newValue), !_bypassPropertyChecks);
                }
            }
            catch (SecurityException)
            {
                throw;
            }
            catch (Exception ex)
            {
                throw new PropertyLoadException(
                  string.Format(Properties.Resources.PropertyLoadException, propertyInfo.Name, ex.Message), ex);
            }
        }

        /// <summary>
        /// Sets a property's managed field with the 
        /// supplied value, first checking authorization, and then
        /// calling PropertyHasChanged if the value does change.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <param name="noAccess">
        /// True if an exception should be thrown when the
        /// user is not authorized to change this property.</param>
        protected void SetProperty<P>(PropertyInfo<P> propertyInfo, P newValue, Security.NoAccessBehavior noAccess)
        {
            if (_bypassPropertyChecks || CanWriteProperty(propertyInfo, noAccess == Security.NoAccessBehavior.ThrowException))
            {
                try
                {
                    P oldValue = default(P);
                    var fieldData = FieldManager.GetFieldData(propertyInfo);
                    if (fieldData == null)
                    {
                        oldValue = propertyInfo.DefaultValue;
                        fieldData = FieldManager.LoadFieldData<P>(propertyInfo, oldValue);
                    }
                    else
                    {
                        var fd = fieldData as FieldManager.IFieldData<P>;
                        if (fd != null)
                            oldValue = fd.Value;
                        else
                            oldValue = (P)fieldData.Value;
                    }
                    if (typeof(P) == typeof(string) && newValue == null)
                        newValue = Utilities.CoerceValue<P>(typeof(string), null, string.Empty);
                    LoadPropertyValue<P>(propertyInfo, oldValue, newValue, !_bypassPropertyChecks);
                }
                catch (Exception ex)
                {
                    throw new PropertyLoadException(
                      string.Format(Properties.Resources.PropertyLoadException, propertyInfo.Name, ex.Message), ex);
                }
            }
        }

        /// <summary>
        /// Sets a property's managed field with the 
        /// supplied value, and then
        /// calls PropertyHasChanged if the value does change.
        /// </summary>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <remarks>
        /// If the user is not authorized to change the 
        /// property a SecurityException is thrown.
        /// </remarks>
        protected void SetProperty(IPropertyInfo propertyInfo, object newValue)
        {
            try
            {
                if (_bypassPropertyChecks || CanWriteProperty(propertyInfo, true))
                {
                    if (!_bypassPropertyChecks) OnPropertyChanging(propertyInfo);
                    FieldManager.SetFieldData(propertyInfo, newValue);
                    if (!_bypassPropertyChecks) PropertyHasChanged(propertyInfo);
                }
            }
            catch (SecurityException)
            {
                throw;
            }
            catch (Exception ex)
            {
                throw new PropertyLoadException(
                  string.Format(Resources.PropertyLoadException, propertyInfo.Name, ex.Message), ex);
            }
        }

        /// <summary>
        /// Sets a property's managed field with the 
        /// supplied value, and then
        /// calls PropertyHasChanged if the value does change.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <remarks>
        /// If the user is not authorized to change the 
        /// property a SecurityException is thrown.
        /// </remarks>
        protected void SetProperty<P>(IPropertyInfo propertyInfo, P newValue)
        {
            SetProperty(propertyInfo, (object)newValue);
        }

        #endregion

        #region  Load Properties

        /// <summary>
        /// Loads a property's managed field with the 
        /// supplied value.
        /// </summary>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <remarks>
        /// No authorization checks occur when this method is called,
        /// and no PropertyChanging or PropertyChanged events are raised.
        /// Loading values does not cause validation rules to be
        /// invoked.
        /// </remarks>
        protected void LoadPropertyConvert<P, F>(PropertyInfo<P> propertyInfo, F newValue)
        {
            try
            {
                P oldValue = default(P);
                var fieldData = FieldManager.GetFieldData(propertyInfo);
                if (fieldData == null)
                {
                    oldValue = propertyInfo.DefaultValue;
                    fieldData = FieldManager.LoadFieldData<P>(propertyInfo, oldValue);
                }
                else
                {
                    var fd = fieldData as FieldManager.IFieldData<P>;
                    if (fd != null)
                        oldValue = fd.Value;
                    else
                        oldValue = (P)fieldData.Value;
                }
                LoadPropertyValue<P>(propertyInfo, oldValue, Utilities.CoerceValue<P>(typeof(F), oldValue, newValue), false);
            }
            catch (Exception ex)
            {
                throw new PropertyLoadException(
                  string.Format(Properties.Resources.PropertyLoadException, propertyInfo.Name, ex.Message), ex);
            }
        }

        void Core.IManageProperties.LoadProperty<P>(PropertyInfo<P> propertyInfo, P newValue)
        {
            LoadProperty<P>(propertyInfo, newValue);
        }

        bool Core.IManageProperties.FieldExists(Core.IPropertyInfo property)
        {
            return FieldManager.FieldExists(property);
        }

        /// <summary>
        /// Loads a property's managed field with the 
        /// supplied value.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <remarks>
        /// No authorization checks occur when this method is called,
        /// and no PropertyChanging or PropertyChanged events are raised.
        /// Loading values does not cause validation rules to be
        /// invoked.
        /// </remarks>
        protected void LoadProperty<P>(PropertyInfo<P> propertyInfo, P newValue)
        {
            try
            {
                P oldValue = default(P);
                var fieldData = FieldManager.GetFieldData(propertyInfo);
                if (fieldData == null)
                {
                    oldValue = propertyInfo.DefaultValue;
                    fieldData = FieldManager.LoadFieldData<P>(propertyInfo, oldValue);
                }
                else
                {
                    var fd = fieldData as FieldManager.IFieldData<P>;
                    if (fd != null)
                        oldValue = fd.Value;
                    else
                        oldValue = (P)fieldData.Value;
                }
                LoadPropertyValue<P>(propertyInfo, oldValue, newValue, false);
            }
            catch (Exception ex)
            {
                throw new PropertyLoadException(
                  string.Format(Properties.Resources.PropertyLoadException, propertyInfo.Name, ex.Message), ex);
            }
        }

        private void LoadPropertyValue<P>(PropertyInfo<P> propertyInfo, P oldValue, P newValue, bool markDirty)
        {
            var valuesDiffer = false;
            if (oldValue == null)
                valuesDiffer = newValue != null;
            else
            {
                // use reference equals for objects that inherit from CSLA base class
                if (typeof(IBusinessObject).IsAssignableFrom(propertyInfo.Type))
                {
                    valuesDiffer = !(ReferenceEquals(oldValue, newValue));
                }
                else
                {
                    valuesDiffer = !(oldValue.Equals(newValue));
                }
            }

            if (valuesDiffer)
            {

                IBusinessObject old = oldValue as IBusinessObject;
                if (old != null)
                    RemoveEventHooks(old);
                IBusinessObject @new = newValue as IBusinessObject;
                if (@new != null)
                    AddEventHooks(@new);


                if (typeof(IEditableBusinessObject).IsAssignableFrom(propertyInfo.Type))
                {
                    if (markDirty)
                    {
                        OnPropertyChanging(propertyInfo);
                        FieldManager.SetFieldData<P>(propertyInfo, newValue);
                        PropertyHasChanged(propertyInfo);
                    }
                    else
                    {
                        FieldManager.LoadFieldData<P>(propertyInfo, newValue);
                    }
                    ResetChildEditLevel(newValue);
                }
                else if (typeof(IEditableCollection).IsAssignableFrom(propertyInfo.Type))
                {
                    if (markDirty)
                    {
                        OnPropertyChanging(propertyInfo);
                        FieldManager.SetFieldData<P>(propertyInfo, newValue);
                        PropertyHasChanged(propertyInfo);
                    }
                    else
                    {
                        FieldManager.LoadFieldData<P>(propertyInfo, newValue);
                    }
                    ResetChildEditLevel(newValue);
                }
                else
                {
                    if (markDirty)
                    {
                        OnPropertyChanging(propertyInfo);
                        FieldManager.SetFieldData<P>(propertyInfo, newValue);
                        PropertyHasChanged(propertyInfo);
                    }
                    else
                    {
                        FieldManager.LoadFieldData<P>(propertyInfo, newValue);
                    }
                }
            }
        }

        /// <summary>
        /// Loads a property's managed field with the 
        /// supplied value.
        /// </summary>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <remarks>
        /// No authorization checks occur when this method is called,
        /// and no PropertyChanging or PropertyChanged events are raised.
        /// Loading values does not cause validation rules to be
        /// invoked.
        /// </remarks>
        protected virtual void LoadProperty(IPropertyInfo propertyInfo, object newValue)
        {
            var t = this.GetType();
            var flags = System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Instance;
            var method = t.GetMethods(flags).Where(c => c.Name == "LoadProperty" && c.IsGenericMethod).FirstOrDefault();
            var gm = method.MakeGenericMethod(propertyInfo.Type);
            var p = new object[] { propertyInfo, newValue };
            gm.Invoke(this, p);
        }

        /// <summary>
        /// Makes sure that a child object is set up properly
        /// to be a child of this object.
        /// </summary>
        /// <param name="newValue">Potential child object</param>
        private void ResetChildEditLevel(object newValue)
        {
            IEditableBusinessObject child = newValue as IEditableBusinessObject;
            if (child != null)
            {
                child.SetParent(this);
                // set child edit level
                UndoableBase.ResetChildEditLevel(child, this.EditLevel, this.BindingEdit);
                // reset EditLevelAdded 
                child.EditLevelAdded = this.EditLevel;
            }
            else
            {
                IEditableCollection col = newValue as IEditableCollection;
                if (col != null)
                {
                    col.SetParent(this);
                    IUndoableObject undo = col as IUndoableObject;
                    if (undo != null)
                    {
                        // set child edit level
                        UndoableBase.ResetChildEditLevel(undo, this.EditLevel, this.BindingEdit);
                    }
                }
            }
        }

        //private AsyncLoadManager
        [NonSerialized]
        [NotUndoable]
        private AsyncLoadManager _loadManager;
        internal AsyncLoadManager LoadManager
        {
            get
            {
                if (_loadManager == null)
                {
                    _loadManager = new AsyncLoadManager();
                    _loadManager.BusyChanged += loadManager_BusyChanged;
                    _loadManager.UnhandledAsyncException += loadManager_UnhandledAsyncException;
                }
                return _loadManager;
            }
        }

        private void loadManager_UnhandledAsyncException(object sender, ErrorEventArgs e)
        {
            OnUnhandledAsyncException(e);
        }

        private void loadManager_BusyChanged(object sender, BusyChangedEventArgs e)
        {
            OnBusyChanged(e);
        }

        /// <summary>
        /// Loads a property value asynchronously.
        /// </summary>
        /// <typeparam name="R">Type of the property</typeparam>
        /// <typeparam name="P">Type of the parameter.</typeparam>
        /// <param name="property">Property to load.</param>
        /// <param name="factory">AsyncFactory delegate.</param>
        /// <param name="parameter">Parameter value.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected void LoadPropertyAsync<R, P>(PropertyInfo<R> property, AsyncFactoryDelegate<R, P> factory, P parameter)
        {
            AsyncLoader loader = new AsyncLoader(property, factory, LoadProperty, OnPropertyChanged, parameter);
            LoadManager.BeginLoad(loader, (EventHandler<DataPortalResult<R>>)loader.LoadComplete);
        }

        /// <summary>
        /// Loads a property value asynchronously.
        /// </summary>
        /// <typeparam name="R">Type of the property</typeparam>
        /// <typeparam name="P1">Type of the parameter.</typeparam>
        /// <typeparam name="P2">Type of the parameter.</typeparam>
        /// <param name="property">Property to load.</param>
        /// <param name="factory">AsyncFactory delegate.</param>
        /// <param name="p1">Parameter value.</param>
        /// <param name="p2">Parameter value.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected void LoadPropertyAsync<R, P1, P2>(PropertyInfo<R> property, AsyncFactoryDelegate<R, P1, P2> factory, P1 p1, P2 p2)
        {
            AsyncLoader loader = new AsyncLoader(property, factory, LoadProperty, OnPropertyChanged, p1, p2);
            LoadManager.BeginLoad(loader, (EventHandler<DataPortalResult<R>>)loader.LoadComplete);
        }

        /// <summary>
        /// Loads a property value asynchronously.
        /// </summary>
        /// <typeparam name="R">Type of the property</typeparam>
        /// <typeparam name="P1">Type of the parameter.</typeparam>
        /// <typeparam name="P2">Type of the parameter.</typeparam>
        /// <typeparam name="P3">Type of the parameter.</typeparam>
        /// <param name="property">Property to load.</param>
        /// <param name="factory">AsyncFactory delegate.</param>
        /// <param name="p1">Parameter value.</param>
        /// <param name="p2">Parameter value.</param>
        /// <param name="p3">Parameter value.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected void LoadPropertyAsync<R, P1, P2, P3>(PropertyInfo<R> property, AsyncFactoryDelegate<R, P1, P2, P3> factory, P1 p1, P2 p2, P3 p3)
        {
            AsyncLoader loader = new AsyncLoader(property, factory, LoadProperty, OnPropertyChanged, p1, p2, p3);
            LoadManager.BeginLoad(loader, (EventHandler<DataPortalResult<R>>)loader.LoadComplete);
        }

        /// <summary>
        /// Loads a property value asynchronously.
        /// </summary>
        /// <typeparam name="R">Type of the property</typeparam>
        /// <typeparam name="P1">Type of the parameter.</typeparam>
        /// <typeparam name="P2">Type of the parameter.</typeparam>
        /// <typeparam name="P3">Type of the parameter.</typeparam>
        /// <typeparam name="P4">Type of the parameter.</typeparam>
        /// <param name="property">Property to load.</param>
        /// <param name="factory">AsyncFactory delegate.</param>
        /// <param name="p1">Parameter value.</param>
        /// <param name="p2">Parameter value.</param>
        /// <param name="p3">Parameter value.</param>
        /// <param name="p4">Parameter value.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected void LoadPropertyAsync<R, P1, P2, P3, P4>(PropertyInfo<R> property, AsyncFactoryDelegate<R, P1, P2, P3, P4> factory, P1 p1, P2 p2, P3 p3, P4 p4)
        {
            AsyncLoader loader = new AsyncLoader(property, factory, LoadProperty, OnPropertyChanged, p1, p2, p3, p4);
            LoadManager.BeginLoad(loader, (EventHandler<DataPortalResult<R>>)loader.LoadComplete);
        }

        /// <summary>
        /// Loads a property value asynchronously.
        /// </summary>
        /// <typeparam name="R">Type of the property</typeparam>
        /// <typeparam name="P1">Type of the parameter.</typeparam>
        /// <typeparam name="P2">Type of the parameter.</typeparam>
        /// <typeparam name="P3">Type of the parameter.</typeparam>
        /// <typeparam name="P4">Type of the parameter.</typeparam>
        /// <typeparam name="P5">Type of the parameter.</typeparam>
        /// <param name="property">Property to load.</param>
        /// <param name="factory">AsyncFactory delegate.</param>
        /// <param name="p1">Parameter value.</param>
        /// <param name="p2">Parameter value.</param>
        /// <param name="p3">Parameter value.</param>
        /// <param name="p4">Parameter value.</param>
        /// <param name="p5">Parameter value.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected void LoadPropertyAsync<R, P1, P2, P3, P4, P5>(PropertyInfo<R> property, AsyncFactoryDelegate<R, P1, P2, P3, P4, P5> factory, P1 p1, P2 p2, P3 p3, P4 p4, P5 p5)
        {
            AsyncLoader loader = new AsyncLoader(property, factory, LoadProperty, OnPropertyChanged, p1, p2, p3, p4, p5);
            LoadManager.BeginLoad(loader, (EventHandler<DataPortalResult<R>>)loader.LoadComplete);
        }

        #endregion

        #region IsBusy / IsIdle

        [NonSerialized]
        [NotUndoable]
        private bool _isBusy;

        /// <summary>
        /// Mark the object as busy (it is
        /// running an async operation).
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void MarkBusy()
        {
            if (_isBusy)
                throw new InvalidOperationException(Resources.BusyObjectsMayNotBeMarkedBusy);

            _isBusy = true;
            OnBusyChanged(new BusyChangedEventArgs("", true));
        }

        /// <summary>
        /// 标记为非忙状态(该方法不要异步操作中调用)        
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void MarkIdle()
        {
            _isBusy = false;
            OnBusyChanged(new BusyChangedEventArgs("", false));
        }

        /// <summary>
        /// 指示对象或它的子对象是否处于忙状态.        
        /// </summary>
        [Browsable(false)]
        [Display(AutoGenerateField = false)]
        public bool IsBusy
        {
            get { return IsSelfBusy || (_fieldManager != null && FieldManager.IsBusy()); }
        }

        /// <summary>
        /// Gets a value indicating if this
        /// object is busy.
        /// </summary>
        [Browsable(false)]
        public bool IsSelfBusy
        {
            get { return _isBusy || BusinessRules.RunningAsyncRules || LoadManager.IsLoading; }
        }

        [NotUndoable]
        [NonSerialized]
        private BusyChangedEventHandler _busyChanged;

        /// <summary>
        /// Event indicating that the IsBusy property has changed.
        /// </summary>
        public event BusyChangedEventHandler BusyChanged
        {
            add { _busyChanged = (BusyChangedEventHandler)Delegate.Combine(_busyChanged, value); }
            remove { _busyChanged = (BusyChangedEventHandler)Delegate.Remove(_busyChanged, value); }
        }

        /// <summary>
        /// Raise the BusyChanged event.
        /// </summary>
        /// <param name="args">Event args.</param>
        protected virtual void OnBusyChanged(BusyChangedEventArgs args)
        {
            if (_busyChanged != null)
                _busyChanged(this, args);
        }

        /// <summary>
        /// Gets a value indicating whether a
        /// specific property is busy (has a
        /// currently executing async rule).
        /// </summary>
        /// <param name="property">
        /// Property to check.
        /// </param>
        public bool IsPropertyBusy(Wcdf.Core.IPropertyInfo property)
        {
            return BusinessRules.GetPropertyBusy(property);
        }

        /// <summary>
        /// Gets a value indicating whether a
        /// specific property is busy (has a
        /// currently executing async rule).
        /// </summary>
        /// <param name="propertyName">
        /// Name of the property.
        /// </param>
        public bool IsPropertyBusy(string propertyName)
        {
            return IsPropertyBusy(FieldManager.GetRegisteredProperties().Where(c => c.Name == propertyName).First());
        }

        #endregion

        #region INotifyUnhandledAsyncException Members

        [NotUndoable]
        [NonSerialized]
        private EventHandler<ErrorEventArgs> _unhandledAsyncException;

        /// <summary>
        /// Event indicating that an exception occurred during
        /// the processing of an async operation.
        /// </summary>
        public event EventHandler<ErrorEventArgs> UnhandledAsyncException
        {
            add { _unhandledAsyncException = (EventHandler<ErrorEventArgs>)Delegate.Combine(_unhandledAsyncException, value); }
            remove { _unhandledAsyncException = (EventHandler<ErrorEventArgs>)Delegate.Remove(_unhandledAsyncException, value); }
        }

        /// <summary>
        /// Raises the UnhandledAsyncException event.
        /// </summary>
        /// <param name="error">Args parameter.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnUnhandledAsyncException(ErrorEventArgs error)
        {
            if (_unhandledAsyncException != null)
                _unhandledAsyncException(this, error);
        }

        /// <summary>
        /// Raises the UnhandledAsyncException event.
        /// </summary>
        /// <param name="originalSender">Original sender of
        /// the event.</param>
        /// <param name="error">Exception object.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void OnUnhandledAsyncException(object originalSender, Exception error)
        {
            OnUnhandledAsyncException(new ErrorEventArgs(originalSender, error));
        }

        #endregion

        #region Child Change Notification

        [NonSerialized]
        [NotUndoable]
        private EventHandler<Wcdf.Core.ChildChangedEventArgs> _childChangedHandlers;

        /// <summary>
        /// 子对象发生更新事件
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design",
          "CA1062:ValidateArgumentsOfPublicMethods")]
        public event EventHandler<Wcdf.Core.ChildChangedEventArgs> ChildChanged
        {
            add
            {
                _childChangedHandlers = (EventHandler<Wcdf.Core.ChildChangedEventArgs>)
                  System.Delegate.Combine(_childChangedHandlers, value);
            }
            remove
            {
                _childChangedHandlers = (EventHandler<Wcdf.Core.ChildChangedEventArgs>)
                  System.Delegate.Remove(_childChangedHandlers, value);
            }
        }

        /// <summary>
        /// 引发子对象更改事件.        
        /// </summary>
        /// <param name="e">
        /// ChildChangedEventArgs object.
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnChildChanged(ChildChangedEventArgs e)
        {
            if (_childChangedHandlers != null)
                _childChangedHandlers.Invoke(this, e);
        }

        /// <summary>
        /// 引发子对象更改事件
        /// </summary>
        private void RaiseChildChanged(ChildChangedEventArgs e)
        {
            OnChildChanged(e);
        }

        /// <summary>
        /// 引发子对象更改事件.
        /// </summary>
        private void RaiseChildChanged(
          object childObject, PropertyChangedEventArgs propertyArgs)
        {
            ChildChangedEventArgs args = new ChildChangedEventArgs(childObject, propertyArgs);
            OnChildChanged(args);
        }

#if !SILVERLIGHT
        /// <summary>
        /// 引发子对象更改事件.
        /// </summary>
        private void RaiseChildChanged(
          object childObject, PropertyChangedEventArgs propertyArgs, ListChangedEventArgs listArgs)
        {
            ChildChangedEventArgs args = new ChildChangedEventArgs(childObject, propertyArgs, listArgs);
            OnChildChanged(args);
        }
#endif

        /// <summary>
        /// 引发子对象更改事件.
        /// </summary>
        private void RaiseChildChanged(
          object childObject, PropertyChangedEventArgs propertyArgs, NotifyCollectionChangedEventArgs listArgs)
        {
            ChildChangedEventArgs args = new ChildChangedEventArgs(childObject, propertyArgs, listArgs);
            OnChildChanged(args);
        }

        /// <summary>
        /// 子对象的属性变更事件,供内部方法调用        
        /// </summary>
        private void Child_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            RaiseChildChanged(sender, e);
        }

#if !SILVERLIGHT
        /// <summary>
        /// 子对象列表改变引发列表变更事件        
        /// </summary>
        private void Child_ListChanged(object sender, ListChangedEventArgs e)
        {
            if (e.ListChangedType != ListChangedType.ItemChanged)
                RaiseChildChanged(sender, null, e);
        }
#endif

        /// <summary>
        ///子对象列表改变引发列表变更事件
        /// </summary>
        private void Child_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            RaiseChildChanged(sender, null, e);
        }

        /// <summary>
        /// 子对象列表改变引发列表变更事件.
        /// </summary>
        private void Child_Changed(object sender, ChildChangedEventArgs e)
        {
            RaiseChildChanged(e);
        }

        #endregion

        #region  Field Manager

        private FieldManager.FieldDataManager _fieldManager;

        /// <summary>
        /// Gets the PropertyManager object for this
        /// business object.
        /// </summary>
        protected FieldDataManager FieldManager
        {
            get
            {
                if (_fieldManager == null)
                {
                    _fieldManager = new FieldDataManager(this.GetType());
                    UndoableBase.ResetChildEditLevel(_fieldManager, this.EditLevel, this.BindingEdit);
                }
                return _fieldManager;
            }
        }

        internal FieldDataManager GetFieldManager()
        {
            return this.FieldManager;
        }

        private void FieldDataDeserialized()
        {
            foreach (object item in FieldManager.GetChildren())
            {
                IBusinessObject business = item as IBusinessObject;
                if (business != null)
                    OnAddEventHooks(business);

                IEditableBusinessObject child = item as IEditableBusinessObject;
                if (child != null)
                {
                    child.SetParent(this);
                }
                IEditableCollection childCollection = item as IEditableCollection;
                if (childCollection != null)
                {
                    childCollection.SetParent(this);
                }
            }
        }

        #endregion
                                      
        #region  IParent

        /// <summary>
        /// 重写该方法以获得通知,在子对象<see cref="Wcdf.BusinessBase.ApplyEdit" />完成时.        
        /// </summary>
        /// <param name="child">The child object that was edited.</param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void EditChildComplete(Core.IEditableBusinessObject child)
        {
            // do nothing, we don't really care
            // when a child has its edits applied
        }

        void IParent.ApplyEditChild(Core.IEditableBusinessObject child)
        {
            this.EditChildComplete(child);
        }

        void IParent.RemoveChild(IEditableBusinessObject child)
        {
            var info = FieldManager.FindProperty(child);
            FieldManager.RemoveField(info);
        }

        #endregion

        #region IDataPortalTarget Members

        void Wcdf.Server.IDataPortalTarget.CheckRules()
        {
            BusinessRules.CheckRules();
        }

        void Wcdf.Server.IDataPortalTarget.MarkAsChild()
        {
            this.MarkAsChild();
        }

        void Wcdf.Server.IDataPortalTarget.MarkNew()
        {
            this.MarkNew();
        }

        void Wcdf.Server.IDataPortalTarget.MarkOld()
        {
            this.MarkOld();
        }
        
        #endregion

        #region IManageProperties Members

        bool IManageProperties.HasManagedProperties
        {
            get { return (_fieldManager != null && _fieldManager.HasFields); }
        }

        List<IPropertyInfo> IManageProperties.GetManagedProperties()
        {
            return FieldManager.GetRegisteredProperties();
        }

        object IManageProperties.GetProperty(IPropertyInfo propertyInfo)
        {
            return GetProperty(propertyInfo);
        }

        object IManageProperties.ReadProperty(IPropertyInfo propertyInfo)
        {
            return ReadProperty(propertyInfo);
        }

        P IManageProperties.ReadProperty<P>(PropertyInfo<P> propertyInfo)
        {
            return ReadProperty<P>(propertyInfo);
        }

        void IManageProperties.SetProperty(IPropertyInfo propertyInfo, object newValue)
        {
            SetProperty(propertyInfo, newValue);
        }

        void IManageProperties.LoadProperty(IPropertyInfo propertyInfo, object newValue)
        {
            LoadProperty(propertyInfo, newValue);
        }

        List<object> IManageProperties.GetChildren()
        {
            return FieldManager.GetChildren();
        }
        #endregion

        #region MobileFormatter

        /// <summary>
        /// Override this method to insert your field values
        /// into the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        /// <param name="mode">
        /// The StateMode indicating why this method was invoked.
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected override void OnGetState(Wcdf.Serialization.Mobile.SerializationInfo info, StateMode mode)
        {
            base.OnGetState(info, mode);
            info.AddValue("Csla.Core.BusinessBase._isNew", _isNew);
            info.AddValue("Csla.Core.BusinessBase._isDeleted", _isDeleted);
            info.AddValue("Csla.Core.BusinessBase._isDirty", _isDirty);
            info.AddValue("Csla.Core.BusinessBase._neverCommitted", _neverCommitted);
            info.AddValue("Csla.Core.BusinessBase._disableIEditableObject", _disableIEditableObject);
            info.AddValue("Csla.Core.BusinessBase._isChild", _isChild);
            info.AddValue("Csla.Core.BusinessBase._editLevelAdded", _editLevelAdded);
        }

        /// <summary>
        /// Override this method to retrieve your field values
        /// from the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        /// <param name="mode">
        /// The StateMode indicating why this method was invoked.
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected override void OnSetState(Wcdf.Serialization.Mobile.SerializationInfo info, StateMode mode)
        {
            base.OnSetState(info, mode);
            _isNew = info.GetValue<bool>("Csla.Core.BusinessBase._isNew");
            _isDeleted = info.GetValue<bool>("Csla.Core.BusinessBase._isDeleted");
            _isDirty = info.GetValue<bool>("Csla.Core.BusinessBase._isDirty");
            _neverCommitted = info.GetValue<bool>("Csla.Core.BusinessBase._neverCommitted");
            _disableIEditableObject = info.GetValue<bool>("Csla.Core.BusinessBase._disableIEditableObject");
            _isChild = info.GetValue<bool>("Csla.Core.BusinessBase._isChild");
            if (mode != StateMode.Undo)
                _editLevelAdded = info.GetValue<int>("Csla.Core.BusinessBase._editLevelAdded");
        }

        /// <summary>
        /// Override this method to insert your child object
        /// references into the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        /// <param name="formatter">
        /// Reference to MobileFormatter instance. Use this to
        /// convert child references to/from reference id values.
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected override void OnGetChildren(
          Wcdf.Serialization.Mobile.SerializationInfo info, Wcdf.Serialization.Mobile.MobileFormatter formatter)
        {
            base.OnGetChildren(info, formatter);

            if (_fieldManager != null)
            {
                var fieldManagerInfo = formatter.SerializeObject(_fieldManager);
                info.AddChild("_fieldManager", fieldManagerInfo.ReferenceId);
            }

            if (_businessRules != null)
            {
                var vrInfo = formatter.SerializeObject(_businessRules);
                info.AddChild("_businessRules", vrInfo.ReferenceId);
            }
        }

        /// <summary>
        /// Override this method to retrieve your child object
        /// references from the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        /// <param name="formatter">
        /// Reference to MobileFormatter instance. Use this to
        /// convert child references to/from reference id values.
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected override void OnSetChildren(Wcdf.Serialization.Mobile.SerializationInfo info, Wcdf.Serialization.Mobile.MobileFormatter formatter)
        {
            if (info.Children.ContainsKey("_fieldManager"))
            {
                var childData = info.Children["_fieldManager"];
                _fieldManager = (FieldManager.FieldDataManager)formatter.GetObject(childData.ReferenceId);
            }

            if (info.Children.ContainsKey("_businessRules"))
            {
                int refId = info.Children["_businessRules"].ReferenceId;
                _businessRules = (BusinessRules)formatter.GetObject(refId);
            }

            base.OnSetChildren(info, formatter);
        }

        #endregion

        #region Property Checks ByPass

        [NonSerialized]
        [NotUndoable]
        private bool _bypassPropertyChecks = false;

        [NonSerialized]
        [NotUndoable]
        private BypassPropertyChecksObject _bypassPropertyChecksObject = null;
        /// <summary>
        /// 该开关用来控制是否启用PropertyChange相关事件,默认为启用.
        /// </summary>
        /// <code>
        /// using(BypassPropertyChecks)
        /// {
        ///     //to do
        /// }
        /// </code>
        /// <remarks>
        /// 用法为当不需要控制时使用using语句        
        /// </remarks>
        protected internal BypassPropertyChecksObject BypassPropertyChecks
        {
            get
            {
                return BypassPropertyChecksObject.GetManager(this);
            }
        }

        /// <summary>
        /// Class that allows setting of property values on 
        /// current business object
        /// without raising PropertyChanged events
        /// and checking user rights.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected internal class BypassPropertyChecksObject : IDisposable
        {
            private BusinessBase _businessObject;
            private static object _lock = new object();

            internal BypassPropertyChecksObject(BusinessBase businessObject)
            {
                _businessObject = businessObject;
                _businessObject._bypassPropertyChecks = true;
            }

            #region IDisposable Members

            /// <summary>
            /// Disposes the object.
            /// </summary>
            public void Dispose()
            {
                Dispose(true);
                GC.SuppressFinalize(this);
            }

            /// <summary>
            /// Disposes the object.
            /// </summary>
            /// <param name="dispose">Dispose flag.</param>
            protected void Dispose(bool dispose)
            {
                DeRef();
            }

            /// <summary>
            /// Gets the BypassPropertyChecks object.
            /// </summary>
            /// <param name="businessObject">The business object.</param>
            /// <returns></returns>
            public static BypassPropertyChecksObject GetManager(BusinessBase businessObject)
            {
                lock (_lock)
                {
                    if (businessObject._bypassPropertyChecksObject == null)
                        businessObject._bypassPropertyChecksObject = new BypassPropertyChecksObject(businessObject);

                    businessObject._bypassPropertyChecksObject.AddRef();
                }
                return businessObject._bypassPropertyChecksObject;
            }

            #region  Reference counting

            private int _refCount;

            /// <summary>
            /// Gets the current reference count for this
            /// object.
            /// </summary>
            public int RefCount
            {
                get { return _refCount; }
            }

            private void AddRef()
            {
                _refCount += 1;
            }

            private void DeRef()
            {

                lock (_lock)
                {
                    _refCount -= 1;
                    if (_refCount == 0)
                    {
                        _businessObject._bypassPropertyChecks = false;
                        _businessObject._bypassPropertyChecksObject = null;
                        _businessObject = null;
                    }
                }
            }

            #endregion
            #endregion
        }

        #endregion

#if SILVERLIGHT
        #region UndoableBase overrides

    /// <summary>
    /// Copy object state.
    /// </summary>
    /// <param name="state">Serialization state.</param>
    [EditorBrowsable(EditorBrowsableState.Never)]
    protected override void OnCopyState(Csla.Serialization.Mobile.SerializationInfo state)
    {
      OnGetState(state, StateMode.Undo);
      ((IUndoableObject)FieldManager).CopyState(this.EditLevel + 1, false);
      ((IUndoableObject)BusinessRules).CopyState(this.EditLevel + 1, false);

      base.OnCopyState(state);
    }

    /// <summary>
    /// Undo object state.
    /// </summary>
    /// <param name="state">Serialization state.</param>
    [EditorBrowsable(EditorBrowsableState.Never)]
    protected override void OnUndoChanges(Csla.Serialization.Mobile.SerializationInfo state)
    {
      OnSetState(state, StateMode.Undo);
      ((IUndoableObject)FieldManager).UndoChanges(this.EditLevel - 1, false);
      ((IUndoableObject)BusinessRules).UndoChanges(this.EditLevel - 1, false);

      base.OnUndoChanges(state);
    }

    /// <summary>
    /// Accept object state.
    /// </summary>
    [EditorBrowsable(EditorBrowsableState.Never)]
    protected override void AcceptingChanges()
    {
      ((IUndoableObject)FieldManager).AcceptChanges(this.EditLevel - 1, false);
      ((IUndoableObject)BusinessRules).AcceptChanges(this.EditLevel - 1, false);

      base.AcceptingChanges();
    }

        #endregion

#if !WINRT
        #region INotifyDataErrorInfo

    /// <summary>
    /// Event raised when error information has changed.
    /// </summary>
    public event EventHandler<DataErrorsChangedEventArgs> ErrorsChanged;

    /// <summary>
    /// Raises the ErrorsChanged event.
    /// </summary>
    /// <param name="e">Event arguments.</param>
    protected virtual void OnErrorsChanged(DataErrorsChangedEventArgs e)
    {
      if (ErrorsChanged != null)
        ErrorsChanged(this, e);
    }

    System.Collections.IEnumerable INotifyDataErrorInfo.GetErrors(string propertyName)
    {
      return BusinessRules.GetBrokenRules()
        .Where(c => c.Property == propertyName && c.Severity == RuleSeverity.Error)
        .ToList();
    }

    bool INotifyDataErrorInfo.HasErrors
    {
      get { return !IsSelfValid; }
    }

    /// <summary>
    /// Call this method to raise the PropertyChanged event
    /// for a specific property.
    /// </summary>
    /// <param name="propertyName">Name of the property that
    /// is being changed.</param>
    protected override void OnPropertyChanged(string propertyName)
    {
      base.OnPropertyChanged(propertyName);
      OnErrorsChanged(new DataErrorsChangedEventArgs(propertyName));
    }

    /// <summary>
    /// Call this method to raise the PropertyChanged event
    /// for all object properties.
    /// </summary>
    protected override void OnUnknownPropertyChanged()
    {
      base.OnUnknownPropertyChanged();
      foreach (var p in FieldManager.GetRegisteredProperties())
        OnErrorsChanged(new DataErrorsChangedEventArgs(p.Name));
    }

        #endregion
#endif
#endif

        #region ISuppressRuleChecking Members

        /// <summary>
        /// Sets value indicating no rule methods will be invoked.
        /// </summary>
        void ICheckRules.SuppressRuleChecking()
        {
            BusinessRules.SuppressRuleChecking = true;
        }

        /// <summary>
        /// Resets value indicating all rule methods will be invoked.
        /// </summary>
        void ICheckRules.ResumeRuleChecking()
        {
            BusinessRules.SuppressRuleChecking = false;
        }

        /// <summary>
        /// Invokes all rules for the business type.
        /// </summary>
        void ICheckRules.CheckRules()
        {
            BusinessRules.CheckRules();
        }

        #endregion
    }
}
