﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Wcdf.Serialization.Mobile;
using System.ComponentModel;
using System.Runtime.Serialization;

namespace Wcdf.Core
{
    /// <summary>
    /// 泛型BindingList扩展.    
    /// </summary>
    /// <typeparam name="T">列表中项类型.</typeparam>
    [Serializable]
    public class ExtendedBindingList<T> : MobileBindingList<T>,
      IExtendedBindingList,
      IMobileObject,
      INotifyBusy,
      INotifyChildChanged,
      ISerializationNotification
    {
        #region RemovingItem event

        [NonSerialized()]
        private EventHandler<RemovingItemEventArgs> _nonSerializableHandlers;
        private EventHandler<RemovingItemEventArgs> _serializableHandlers;

        /// <summary>
        /// Implements a serialization-safe RemovingItem event.
        /// </summary>
        public event EventHandler<RemovingItemEventArgs> RemovingItem
        {
            add
            {
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableHandlers = (EventHandler<RemovingItemEventArgs>)
                      System.Delegate.Combine(_serializableHandlers, value);
                else
                    _nonSerializableHandlers = (EventHandler<RemovingItemEventArgs>)
                      System.Delegate.Combine(_nonSerializableHandlers, value);
            }
            remove
            {
                if (value.Method.IsPublic &&
                   (value.Method.DeclaringType.IsSerializable ||
                    value.Method.IsStatic))
                    _serializableHandlers = (EventHandler<RemovingItemEventArgs>)
                      System.Delegate.Remove(_serializableHandlers, value);
                else
                    _nonSerializableHandlers = (EventHandler<RemovingItemEventArgs>)
                      System.Delegate.Remove(_nonSerializableHandlers, value);
            }
        }

        /// <summary>
        /// Raise the RemovingItem event.
        /// </summary>
        /// <param name="removedItem">
        /// A reference to the item that 
        /// is being removed.
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected void OnRemovingItem(T removedItem)
        {
            if (_nonSerializableHandlers != null)
                _nonSerializableHandlers.Invoke(this,
                  new RemovingItemEventArgs(removedItem));
            if (_serializableHandlers != null)
                _serializableHandlers.Invoke(this,
                  new RemovingItemEventArgs(removedItem));
        }

        #endregion

        #region RemoveItem

        /// <summary>
        /// Remove the item at the
        /// specified index.
        /// </summary>
        /// <param name="index">
        /// The zero-based index of the item
        /// to remove.
        /// </param>
        protected override void RemoveItem(int index)
        {
            OnRemovingItem(this[index]);
            OnRemoveEventHooks(this[index]);
            base.RemoveItem(index);
        }

        #endregion

        #region AddRange

        /// <summary>
        /// Add a range of items to the list.
        /// </summary>
        /// <param name="range">List of items to add.</param>
        public void AddRange(System.Collections.Generic.IEnumerable<T> range)
        {
            foreach (var element in range)
                this.Add(element);
        }

        #endregion

        #region INotifyPropertyBusy Members

        [NotUndoable]
        [NonSerialized]
        private BusyChangedEventHandler _busyChanged = null;

        /// <summary>
        /// Event indicating that the busy status of the
        /// object has changed.
        /// </summary>
        public event BusyChangedEventHandler BusyChanged
        {
            add { _busyChanged = (BusyChangedEventHandler)Delegate.Combine(_busyChanged, value); }
            remove { _busyChanged = (BusyChangedEventHandler)Delegate.Remove(_busyChanged, value); }
        }

        /// <summary>
        /// Override this method to be notified when the
        /// IsBusy property has changed.
        /// </summary>
        /// <param name="args">Event arguments.</param>
        protected virtual void OnBusyChanged(BusyChangedEventArgs args)
        {
            if (_busyChanged != null)
                _busyChanged(this, args);
        }

        /// <summary>
        /// Raises the BusyChanged event for a specific property.
        /// </summary>
        /// <param name="propertyName">Name of the property.</param>
        /// <param name="busy">New busy value.</param>
        protected void OnBusyChanged(string propertyName, bool busy)
        {
            OnBusyChanged(new BusyChangedEventArgs(propertyName, busy));
        }

        /// <summary>
        /// Gets the busy status for this object and its child objects.
        /// </summary>
        [Browsable(false)]
        [System.ComponentModel.DataAnnotations.Display(AutoGenerateField = false)]
        public virtual bool IsBusy
        {
            get { throw new NotImplementedException(); }
        }

        /// <summary>
        /// Gets the busy status for this object.
        /// </summary>
        [Browsable(false)]
        [System.ComponentModel.DataAnnotations.Display(AutoGenerateField = false)]
        public virtual bool IsSelfBusy
        {
            get { return IsBusy; }
        }

        void busy_BusyChanged(object sender, BusyChangedEventArgs e)
        {
            OnBusyChanged(e);
        }

        #endregion

        #region INotifyUnhandledAsyncException Members

        [NotUndoable]
        [NonSerialized]
        private EventHandler<ErrorEventArgs> _unhandledAsyncException;

        /// <summary>
        /// Event indicating that an exception occurred during
        /// an async operation.
        /// </summary>
        public event EventHandler<ErrorEventArgs> UnhandledAsyncException
        {
            add { _unhandledAsyncException = (EventHandler<ErrorEventArgs>)Delegate.Combine(_unhandledAsyncException, value); }
            remove { _unhandledAsyncException = (EventHandler<ErrorEventArgs>)Delegate.Remove(_unhandledAsyncException, value); }
        }

        /// <summary>
        /// Method invoked when an unhandled async exception has
        /// occurred.
        /// </summary>
        /// <param name="error">Event arguments.</param>
        protected virtual void OnUnhandledAsyncException(ErrorEventArgs error)
        {
            if (_unhandledAsyncException != null)
                _unhandledAsyncException(this, error);
        }

        /// <summary>
        /// Raises the UnhandledAsyncException event.
        /// </summary>
        /// <param name="originalSender">Original sender of event.</param>
        /// <param name="error">Exception that occurred.</param>
        protected void OnUnhandledAsyncException(object originalSender, Exception error)
        {
            OnUnhandledAsyncException(new ErrorEventArgs(originalSender, error));
        }

        void unhandled_UnhandledAsyncException(object sender, ErrorEventArgs e)
        {
            OnUnhandledAsyncException(e);
        }

        #endregion

        #region AddChildHooks

        /// <summary>
        /// Invoked when an item is inserted into the list.
        /// </summary>
        /// <param name="index">Index of new item.</param>
        /// <param name="item">Reference to new item.</param>
        protected override void InsertItem(int index, T item)
        {
            base.InsertItem(index, item);
            OnAddEventHooks(item);
        }

        /// <summary>
        /// Method invoked when events are hooked for a child
        /// object.
        /// </summary>
        /// <param name="item">Reference to child object.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected virtual void OnAddEventHooks(T item)
        {
            INotifyBusy busy = item as INotifyBusy;
            if (busy != null)
                busy.BusyChanged += new BusyChangedEventHandler(busy_BusyChanged);

            INotifyUnhandledAsyncException unhandled = item as INotifyUnhandledAsyncException;
            if (unhandled != null)
                unhandled.UnhandledAsyncException += new EventHandler<ErrorEventArgs>(unhandled_UnhandledAsyncException);

            INotifyPropertyChanged c = item as INotifyPropertyChanged;
            if (c != null)
                c.PropertyChanged += Child_PropertyChanged;

            //IBindingList list = item as IBindingList;
            //if (list != null)
            //  list.ListChanged += new ListChangedEventHandler(Child_ListChanged);

            INotifyChildChanged child = item as INotifyChildChanged;
            if (child != null)
                child.ChildChanged += Child_Changed;
        }

        /// <summary>
        /// Method invoked when events are unhooked for a child
        /// object.
        /// </summary>
        /// <param name="item">Reference to child object.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected virtual void OnRemoveEventHooks(T item)
        {
            INotifyBusy busy = item as INotifyBusy;
            if (busy != null)
                busy.BusyChanged -= new BusyChangedEventHandler(busy_BusyChanged);

            INotifyUnhandledAsyncException unhandled = item as INotifyUnhandledAsyncException;
            if (unhandled != null)
                unhandled.UnhandledAsyncException -= new EventHandler<ErrorEventArgs>(unhandled_UnhandledAsyncException);

            INotifyPropertyChanged c = item as INotifyPropertyChanged;
            if (c != null)
                c.PropertyChanged -= new PropertyChangedEventHandler(Child_PropertyChanged);

            //IBindingList list = item as IBindingList;
            //if(list!=null)
            //  list.ListChanged -= new ListChangedEventHandler(Child_ListChanged);

            INotifyChildChanged child = item as INotifyChildChanged;
            if (child != null)
                child.ChildChanged -= new EventHandler<ChildChangedEventArgs>(Child_Changed);
        }

        #endregion

        #region ISerializationNotification Members

        /// <summary>
        /// This method is called on a newly deserialized object
        /// after deserialization is complete.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnDeserialized()
        {
            // do nothing - this is here so a subclass
            // could override if needed
        }

        [OnDeserialized]
        private void OnDeserializedHandler(StreamingContext context)
        {
            foreach (T item in this)
                OnAddEventHooks(item);

            OnDeserialized();
        }

        void ISerializationNotification.Deserialized()
        {
            // don't rehook events here, because the MobileFormatter has
            // created new objects and so the lists will auto-subscribe
            // the events
            OnDeserialized();
        }

        #endregion

        #region Child Change Notification

        [NonSerialized]
        [NotUndoable]
        private EventHandler<Wcdf.Core.ChildChangedEventArgs> _childChangedHandlers;

        /// <summary>
        /// Event raised when a child object has been changed.
        /// </summary>
        public event EventHandler<Wcdf.Core.ChildChangedEventArgs> ChildChanged
        {
            add
            {
                _childChangedHandlers = (EventHandler<Wcdf.Core.ChildChangedEventArgs>)
                  System.Delegate.Combine(_childChangedHandlers, value);
            }
            remove
            {
                _childChangedHandlers = (EventHandler<Wcdf.Core.ChildChangedEventArgs>)
                  System.Delegate.Remove(_childChangedHandlers, value);
            }
        }

        /// <summary>
        /// Raises the ChildChanged event, indicating that a child
        /// object has been changed.
        /// </summary>
        /// <param name="e">
        /// ChildChangedEventArgs object.
        /// </param>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void OnChildChanged(ChildChangedEventArgs e)
        {
            if (_childChangedHandlers != null)
                _childChangedHandlers.Invoke(this, e);
        }

        /// <summary>
        /// Creates a ChildChangedEventArgs and raises the event.
        /// </summary>
        private void RaiseChildChanged(
          object childObject, PropertyChangedEventArgs propertyArgs, ListChangedEventArgs listArgs)
        {
            ChildChangedEventArgs args = new ChildChangedEventArgs(childObject, propertyArgs, listArgs);
            OnChildChanged(args);
        }

        /// <summary>
        /// Handles any PropertyChanged event from 
        /// a child object and echoes it up as
        /// a ChildChanged event.
        /// </summary>
        /// <param name="sender">Object that raised the event.</param>
        /// <param name="e">Property changed args.</param>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected virtual void Child_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            RaiseChildChanged(sender, e, null);
        }

        /// <summary>
        /// Handles any ChildChanged event from
        /// a child object and echoes it up as
        /// a ChildChanged event.
        /// </summary>
        private void Child_Changed(object sender, ChildChangedEventArgs e)
        {
            RaiseChildChanged(e.ChildObject, e.PropertyChangedArgs, e.ListChangedArgs);
        }

        #endregion
    }
}
