﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel.DataAnnotations;
using System.ComponentModel;

namespace Wcdf.Core.FieldManager
{
    /// <summary>
    /// Creates PropertyInfo objects.
    /// </summary>
    internal class DefaultPropertyInfoFactory : Wcdf.Core.IPropertyInfoFactory
    {
        #region Csla.PropertyInfo<T>Factory Members

        /// <summary>
        /// Creates a new instance of PropertyInfo.
        /// </summary>
        /// <param name="containingType">
        /// Type of business class that contains the property
        /// declaration.
        /// </param>
        /// <param name="name">Name of the property.</param>
        public Wcdf.PropertyInfo<T> Create<T>(Type containingType, string name)
        {
            var friendlyName = GetFriendlyNameFromAttributes(containingType, name);
            return new Wcdf.PropertyInfo<T>(name, friendlyName);
        }

        /// <summary>
        /// Creates a new instance of PropertyInfo.
        /// </summary>
        /// <param name="containingType">
        /// Type of business class that contains the property
        /// declaration.
        /// </param>
        /// <param name="name">Name of the property.</param>
        /// <param name="friendlyName">
        /// Friendly display name for the property.
        /// </param>
        public Wcdf.PropertyInfo<T> Create<T>(Type containingType, string name, string friendlyName)
        {
            if (string.IsNullOrWhiteSpace(friendlyName))
                friendlyName = GetFriendlyNameFromAttributes(containingType, name);

            return new Wcdf.PropertyInfo<T>(name, friendlyName);
        }

        /// <summary>
        /// Creates a new instance of PropertyInfo.
        /// </summary>
        /// <param name="containingType">
        /// Type of business class that contains the property
        /// declaration.
        /// </param>
        /// <param name="name">Name of the property.</param>
        /// <param name="friendlyName">
        /// Friendly display name for the property.
        /// </param>
        /// <param name="relationship">Relationship with
        /// referenced object.</param>
        public Wcdf.PropertyInfo<T> Create<T>(Type containingType, string name, string friendlyName, RelationshipTypes relationship)
        {
            if (string.IsNullOrWhiteSpace(friendlyName))
                friendlyName = GetFriendlyNameFromAttributes(containingType, name);

            return new Wcdf.PropertyInfo<T>(name, friendlyName, relationship);
        }

        /// <summary>
        /// Creates a new instance of PropertyInfo.
        /// </summary>
        /// <param name="containingType">
        /// Type of business class that contains the property
        /// declaration.
        /// </param>
        /// <param name="name">Name of the property.</param>
        /// <param name="friendlyName">
        /// Friendly display name for the property.
        /// </param>
        /// <param name="defaultValue">
        /// Default value for the property.
        /// </param>
        public Wcdf.PropertyInfo<T> Create<T>(Type containingType, string name, string friendlyName, T defaultValue)
        {
            if (string.IsNullOrWhiteSpace(friendlyName))
                friendlyName = GetFriendlyNameFromAttributes(containingType, name);

            return new Wcdf.PropertyInfo<T>(name, friendlyName, defaultValue);
        }

        /// <summary>
        /// Creates a new instance of PropertyInfo.
        /// </summary>
        /// <param name="containingType">
        /// Type of business class that contains the property
        /// declaration.
        /// </param>
        /// <param name="name">Name of the property.</param>
        /// <param name="friendlyName">
        /// Friendly display name for the property.
        /// </param>
        /// <param name="defaultValue">
        /// Default value for the property.
        /// </param>
        /// <param name="relationship">Relationship with
        /// referenced object.</param>
        public Wcdf.PropertyInfo<T> Create<T>(Type containingType, string name, string friendlyName, T defaultValue, RelationshipTypes relationship)
        {
            if (string.IsNullOrWhiteSpace(friendlyName))
                friendlyName = GetFriendlyNameFromAttributes(containingType, name);

            return new Wcdf.PropertyInfo<T>(name, friendlyName, defaultValue, relationship);
        }

        #endregion

        #region Helpers

        /// <summary>
        /// 根据对象类型，属性名称，从属性定义的显示特性取得PeopertyInfo对象的FriendlyName名称。
        /// 解决了DataAnnotations 与 ComponentModel特性定义差异。        
        /// Note: This code should not throw any exceptions because the property must exist!
        /// </summary>
        /// <param name="type">类型.</param>
        /// <param name="name">属性名称.</param>
        /// <returns>属性的显示名（如果有定义显示特性）或者属性名称.</returns>
        private static string GetFriendlyNameFromAttributes(Type type, string name)
        {
            // If name is blank then check the DataAnnotations attribute and then the ComponentModel attribute.
            var propertyInfo = type.GetProperty(name);
            if (propertyInfo != null)
            {
                // DataAnnotations attribute.
                var display = propertyInfo.GetCustomAttributes(typeof(DisplayAttribute), true).OfType<DisplayAttribute>().FirstOrDefault();
                if (display != null)
                    name = display.GetName();
#if !SILVERLIGHT
                else
                {
                    // ComponentModel attribute.
                    var displayName = propertyInfo.GetCustomAttributes(typeof(DisplayNameAttribute), true).OfType<DisplayNameAttribute>().FirstOrDefault();
                    if (displayName != null)
                        name = displayName.DisplayName;
                }
#endif
            }
            return name;
        }

        #endregion
    }
}
