﻿using System;
using System.ComponentModel;

namespace Wcdf.Core.FieldManager
{
    /// <summary>
    /// 包含字段值和相关元数据
    /// </summary>
    /// <typeparam name="T">字段类型.</typeparam>
    [Serializable()]
    public class FieldData<T> : IFieldData<T>
    {
        private string _name;
        private T _data;
        private bool _isDirty;

        /// <summary>
        /// 创建新对象实例.
        /// </summary>
        /// <param name="name">
        /// 字段名称.
        /// </param>
        public FieldData(string name)
        {
            _name = name;
        }

        /// <summary>
        /// 取得字段名称.
        /// </summary>
        public string Name
        {
            get
            {
                return _name;
            }
        }

        /// <summary>
        /// 获取和设置字段值.
        /// </summary>
        public virtual T Value
        {
            get
            {
                return _data;
            }
            set
            {
                _data = value;
                _isDirty = true;
            }
        }

        object IFieldData.Value
        {
            get
            {
                return this.Value;
            }
            set
            {
                if (value == null)
                    this.Value = default(T);
                else
                    this.Value = (T)value;
            }
        }

        bool ITrackStatus.IsDeleted
        {
            get
            {
                ITrackStatus child = _data as ITrackStatus;
                if (child != null)
                {
                    return child.IsDeleted;

                }
                else
                {
                    return false;
                }
            }
        }

        bool ITrackStatus.IsSavable
        {
            get { return true; }
        }

        /// <summary>
        /// 获取字段是否发生更改
        /// Gets a value indicating whether the field
        /// has been changed.
        /// </summary>
        public virtual bool IsSelfDirty
        {
            get { return IsDirty; }
        }

        /// <summary>
        /// 获取字段是否发生改变        
        /// </summary>
        public virtual bool IsDirty
        {
            get
            {
                ITrackStatus child = _data as ITrackStatus;
                if (child != null)
                {
                    return child.IsDirty;

                }
                else
                {
                    return _isDirty;
                }
            }
        }

        /// <summary>
        /// 将标记字段为未改变.
        /// </summary>
        public virtual void MarkClean()
        {
            _isDirty = false;

        }

        bool ITrackStatus.IsNew
        {
            get
            {
                ITrackStatus child = _data as ITrackStatus;
                if (child != null)
                {
                    return child.IsNew;

                }
                else
                {
                    return false;
                }
            }
        }

        bool ITrackStatus.IsSelfValid
        {
            get { return IsValid; }
        }

        bool ITrackStatus.IsValid
        {
            get { return IsValid; }
        }

        /// <summary>
        /// 获取字段是否可用        
        /// </summary>
        protected virtual bool IsValid
        {
            get
            {
                ITrackStatus child = _data as ITrackStatus;
                if (child != null)
                {
                    return child.IsValid;

                }
                else
                {
                    return true;
                }
            }
        }


        #region INotifyBusy Members

        event BusyChangedEventHandler INotifyBusy.BusyChanged
        {
            add { throw new NotImplementedException(); }
            remove { throw new NotImplementedException(); }
        }

        /// <summary>
        /// Gets a value indicating whether this object or
        /// any of its child objects are busy.
        /// </summary>
        [Browsable(false)]
#if !SILVERLIGHT
        [System.ComponentModel.DataAnnotations.Display(AutoGenerateField = false)]
#endif
        public bool IsBusy
        {
            get
            {
                bool isBusy = false;
                ITrackStatus status = _data as ITrackStatus;
                if (status != null)
                    isBusy = status.IsBusy;

                return isBusy;
            }
        }

        bool INotifyBusy.IsSelfBusy
        {
            get { return IsBusy; }
        }

        #endregion

        #region INotifyUnhandledAsyncException Members

        [NotUndoable]
        [NonSerialized]
        private EventHandler<ErrorEventArgs> _unhandledAsyncException;

        /// <summary>
        /// Event indicating that an exception occurred on
        /// a background thread.
        /// </summary>
        public event EventHandler<ErrorEventArgs> UnhandledAsyncException
        {
            add { _unhandledAsyncException = (EventHandler<ErrorEventArgs>)Delegate.Combine(_unhandledAsyncException, value); }
            remove { _unhandledAsyncException = (EventHandler<ErrorEventArgs>)Delegate.Remove(_unhandledAsyncException, value); }
        }

        /// <summary>
        /// Raises the UnhandledAsyncException event.
        /// </summary>
        /// <param name="error">Exception that occurred on the background thread.</param>
        protected virtual void OnUnhandledAsyncException(ErrorEventArgs error)
        {
            if (_unhandledAsyncException != null)
                _unhandledAsyncException(this, error);
        }

        /// <summary>
        /// Raises the UnhandledAsyncException event.
        /// </summary>
        /// <param name="originalSender">Original source of the event.</param>
        /// <param name="error">Exception that occurred on the background thread.</param>
        protected void OnUnhandledAsyncException(object originalSender, Exception error)
        {
            OnUnhandledAsyncException(new ErrorEventArgs(originalSender, error));
        }

        #endregion
    }
}
