﻿using System;
using System.Collections.Generic;
using Wcdf.Serialization.Mobile;
using Wcdf.Properties;
using Wcdf.Serialization;
using System.IO;

namespace Wcdf.Core.FieldManager
{
    /// <summary>
    /// 为业务对象管理属性及属性数据,本对象应用于Core.BusinessBase中，
    /// 存储业务对象实例的字段数据。
    /// </summary>
    /// <remarks></remarks>
    [Serializable]
#if SILVERLIGHT
  public class FieldDataManager : MobileObject, IUndoableObject
#else
    public class FieldDataManager : IUndoableObject, IMobileObject
#endif
    {
        private string _businessObjectType;
        [NonSerialized]
        BusinessBase _parent;
        [NonSerialized()]
        private List<IPropertyInfo> _propertyList;
        private IFieldData[] _fieldData;

#if SILVERLIGHT
    /// <summary>
    /// Creates an instance of the object.
    /// </summary>
    public FieldDataManager() { }
#else
        private FieldDataManager()
        { /* exists to support MobileFormatter */ }
#endif

        internal FieldDataManager(Type businessObjectType)
        {
            SetPropertyList(businessObjectType);
            _fieldData = new IFieldData[_propertyList.Count];
        }

        /// <summary>
        /// Called when parent object is deserialized to
        /// restore property list.
        /// </summary>
        internal void SetPropertyList(Type businessObjectType)
        {
            _businessObjectType = businessObjectType.AssemblyQualifiedName;
            _propertyList = GetConsolidatedList(businessObjectType);
        }

        /// <summary>
        /// Called by parent to set the back-reference.
        /// </summary>
        internal void SetParent(BusinessBase parent)
        {
            _parent = parent;
        }

        /// <summary>
        /// 获取注册属性列表，如果返回null则该对象并未注册属性        
        /// </summary>
        public List<IPropertyInfo> GetRegisteredProperties()
        {
            return new List<IPropertyInfo>(_propertyList);
        }

        /// <summary>
        /// Gets a value indicating whether there
        /// are any managed fields available.
        /// </summary>
        public bool HasFields
        {
            get { return _propertyList.Count > 0; }
        }

        #region ConsolidatedPropertyList

        private static Dictionary<Type, List<IPropertyInfo>> _consolidatedLists = new Dictionary<Type, List<IPropertyInfo>>();
        /// <summary>
        /// 取得指定类型的注册属性列表
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        private static List<IPropertyInfo> GetConsolidatedList(Type type)
        {
            List<IPropertyInfo> result = null;
            if (!_consolidatedLists.TryGetValue(type, out result))
            {
                lock (_consolidatedLists)
                {
                    if (_consolidatedLists.ContainsKey(type))
                    {
                        result = _consolidatedLists[type];
                    }
                    else
                    {
                        result = CreateConsolidatedList(type);
                        _consolidatedLists.Add(type, result);
                    }
                }
            }
            return result;
        }
        
        private static List<IPropertyInfo> CreateConsolidatedList(Type type)
        {
            ForceStaticFieldInit(type);
            List<IPropertyInfo> result = new List<IPropertyInfo>();

            // get inheritance hierarchy
            Type current = type;
            List<Type> hierarchy = new List<Type>();
            do
            {
                hierarchy.Add(current);
                current = current.BaseType;
            } while (current != null && !current.Equals(typeof(BusinessBase)));

            // walk from top to bottom to build consolidated list
            for (int index = hierarchy.Count - 1; index >= 0; index--)
            {
                var source = PropertyInfoManager.GetPropertyListCache(hierarchy[index]);
                source.IsLocked = true;
                result.AddRange(source);
            }

            // set Index properties on all unindexed PropertyInfo objects
            int max = -1;
            foreach (var item in result)
            {
                if (item.Index == -1)
                {
                    max++;
                    item.Index = max;
                }
                else
                {
                    max = item.Index;
                }
            }
            // return consolidated list
            return result;
        }

        #endregion

        #region  Get/Set/Find fields

        /// <summary>
        /// 获取指定属性的关联<see cref="IFieldData" /> 对象
        /// for a specific field.
        /// </summary>
        /// <param name="propertyInfo">
        /// 指向字段的属性对象.
        /// </param>
        [System.ComponentModel.EditorBrowsable(System.ComponentModel.EditorBrowsableState.Advanced)]
        public IFieldData GetFieldData(IPropertyInfo propertyInfo)
        {
            if ((propertyInfo.RelationshipType & RelationshipTypes.PrivateField) == RelationshipTypes.PrivateField)
                throw new InvalidOperationException(Resources.PropertyIsPrivateField);

            try
            {
                return _fieldData[propertyInfo.Index];
            }
            catch (IndexOutOfRangeException ex)
            {
                throw new InvalidOperationException(Resources.PropertyNotRegistered, ex);
            }
        }

        private IFieldData GetOrCreateFieldData(IPropertyInfo prop)
        {
            try
            {
                var field = _fieldData[prop.Index];
                if (field == null)
                {
                    field = prop.NewFieldData(prop.Name);
                    _fieldData[prop.Index] = field;
                }
                return field;
            }
            catch (IndexOutOfRangeException ex)
            {
                throw new InvalidOperationException(Resources.PropertyNotRegistered, ex);
            }
        }

        internal IPropertyInfo FindProperty(object value)
        {
            var index = 0;
            foreach (var item in _fieldData)
            {
                if (item != null && item.Value != null && item.Value.Equals(value))
                    return _propertyList[index];
                index += 1;
            }
            return null;
        }

        internal IPropertyInfo FindProperty(string FriendName)
        {
            var index = 0;
            foreach (var item in _fieldData)
            {
                if (item.Name == FriendName)
                    return _propertyList[index];
                index += 1;
            }
            return null;
        }
        /// <summary>
        /// 设置指定字段的值.
        /// </summary>
        /// <param name="prop">
        /// 字段对应的属性。
        /// </param>
        /// <param name="value">
        /// 字段存储值。
        /// </param>
        internal void SetFieldData(IPropertyInfo prop, object value)
        {
            Type valueType;
            if (value != null)
                valueType = value.GetType();
            else
                valueType = prop.Type;
            value = Utilities.CoerceValue(prop.Type, valueType, null, value);
            var field = GetOrCreateFieldData(prop);
            field.Value = value;
        }

        /// <summary>
        /// 设置指定字段的值.
        /// </summary>
        /// <typeparam name="P">
        /// 字段值类型.
        /// </typeparam>
        /// <param name="prop">
        /// 字段对应的属性。
        /// </param>
        /// <param name="value">
        /// 字段存储值。
        /// </param>
        internal void SetFieldData<P>(IPropertyInfo prop, P value)
        {
            var field = GetOrCreateFieldData(prop);
            var fd = field as IFieldData<P>;
            if (fd != null)
                fd.Value = value;
            else
                field.Value = value;
        }

        /// <summary>
        /// 不标识脏标识的设置指定字段的值。        
        /// </summary>
        /// <param name="prop">
        /// 字段对应的属性。
        /// </param>
        /// <param name="value">
        /// 字段存储值.
        /// </param>
        internal IFieldData LoadFieldData(IPropertyInfo prop, object value)
        {
            Type valueType;
            if (value != null)
                valueType = value.GetType();
            else
                valueType = prop.Type;
            value = Utilities.CoerceValue(prop.Type, valueType, null, value);
            var field = GetOrCreateFieldData(prop);
            field.Value = value;
            field.MarkClean();
            return field;
        }

        /// <summary>
        /// 不标识脏标识的设置指定字段的值。
        /// </summary>
        /// <typeparam name="P">
        /// 字段值类型.
        /// </typeparam>
        /// <param name="prop">
        /// 字段对应的属性。
        /// </param>
        /// <param name="value">
        /// 字段存储值。
        /// </param>
        internal IFieldData LoadFieldData<P>(IPropertyInfo prop, P value)
        {
            var field = GetOrCreateFieldData(prop);
            var fd = field as IFieldData<P>;
            if (fd != null)
                fd.Value = value;
            else
                field.Value = value;
            field.MarkClean();
            return field;
        }

        /// <summary>
        /// 删除指定属性的值,保留<see cref="IFieldData" /> 对象        
        /// </summary>
        /// <param name="prop">
        /// 字段的对应属性.
        /// </param>
        internal void RemoveField(IPropertyInfo prop)
        {
            try
            {
                var field = _fieldData[prop.Index];
                if (field != null)
                    field.Value = null;
            }
            catch (IndexOutOfRangeException ex)
            {
                throw new InvalidOperationException(Resources.PropertyNotRegistered, ex);
            }
        }

        /// <summary>
        /// Returns a value indicating whether an
        /// <see cref="IFieldData" /> entry exists
        /// for the specified property.
        /// </summary>
        /// <param name="propertyInfo">
        /// The property corresponding to the field.
        /// </param>
        public bool FieldExists(IPropertyInfo propertyInfo)
        {
            try
            {
                return _fieldData[propertyInfo.Index] != null;
            }
            catch (IndexOutOfRangeException ex)
            {
                throw new InvalidOperationException(Resources.PropertyNotRegistered, ex);
            }
        }

        /// <summary>
        /// 指定字段值是否发生改变        
        /// </summary>
        /// <param name="propertyInfo">
        /// The property corresponding to the field.
        /// </param>
        /// <returns>True if the field has been changed.</returns>
        public bool IsFieldDirty(IPropertyInfo propertyInfo)
        {
            try
            {
                bool result = false;
                var field = _fieldData[propertyInfo.Index];
                if (field != null)
                    result = field.IsDirty;
                else
                    result = false;
                return result;

            }
            catch (IndexOutOfRangeException ex)
            {
                throw new InvalidOperationException(Properties.Resources.PropertyNotRegistered, ex);
            }
        }

        #endregion

        #region  IsValid/IsDirty/IsBusy

        /// <summary>
        /// 返回是否所有字段都可用
        /// </summary>
        public bool IsValid()
        {
            foreach (var item in _fieldData)
                if (item != null && !item.IsValid)
                    return false;
            return true;
        }

        /// <summary>
        /// 返回是否存在字段脏标识        
        /// </summary>
        public bool IsDirty()
        {
            foreach (var item in _fieldData)
                if (item != null && item.IsDirty)
                    return true;
            return false;
        }

        /// <summary>
        /// Marks all fields as clean
        /// (not dirty).
        /// </summary>
        internal void MarkClean()
        {
            foreach (var item in _fieldData)
                if (item != null && item.IsDirty)
                    item.MarkClean();
        }

        internal bool IsBusy()
        {
            foreach (var item in _fieldData)
                if (item != null && item.IsBusy)
                    return true;
            return false;
        }

        #endregion

        #region  IUndoableObject

#if SILVERLIGHT
    private Stack<SerializationInfo> _stateStack = new Stack<SerializationInfo>();
#else
        private Stack<byte[]> _stateStack = new Stack<byte[]>();
#endif

        /// <summary>
        /// 当前对象遥编辑层数        
        /// </summary>
        public int EditLevel
        {
            get { return _stateStack.Count; }
        }

        void Core.IUndoableObject.CopyState(int parentEditLevel, bool parentBindingEdit)
        {
            if (this.EditLevel + 1 > parentEditLevel)
                throw new UndoException(string.Format(Properties.Resources.EditLevelMismatchException, "CopyState"));

#if SILVERLIGHT
      SerializationInfo state = new SerializationInfo(0);
      OnGetState(state, StateMode.Undo);

      // This is used instead of a foreach because of some weird silverlight bug
      // throwing an unknown exception from a foreach here.
      for (var index = 0; index < _fieldData.Length; index++)
      {
        var item = _fieldData[index];
        if (item != null)
        {
          var child = item.Value as IUndoableObject;
          if (child != null)
          {
            // cascade call to child
            child.CopyState(parentEditLevel, parentBindingEdit);
          }
        }
      }

      _stateStack.Push(state);
#else
            IFieldData[] state = new IFieldData[_propertyList.Count];

            for (var index = 0; index < _fieldData.Length; index++)
            {
                var item = _fieldData[index];
                if (item != null)
                {
                    var child = item.Value as IUndoableObject;
                    if (child != null)
                    {
                        // cascade call to child
                        child.CopyState(parentEditLevel, parentBindingEdit);
                        // indicate that there was a value here
                        state[index] = new FieldData<bool>(item.Name);
                    }
                    else
                    {
                        // add the IFieldData object
                        state[index] = item;
                    }
                }
            }

            // serialize the state and stack it
            using (MemoryStream buffer = new MemoryStream())
            {
                var formatter = SerializationFormatterFactory.GetFormatter();
                formatter.Serialize(buffer, state);
                _stateStack.Push(buffer.ToArray());
            }
#endif
        }

        void Core.IUndoableObject.UndoChanges(int parentEditLevel, bool parentBindingEdit)
        {
            if (EditLevel > 0)
            {
                if (this.EditLevel - 1 != parentEditLevel)
                    throw new UndoException(string.Format(Properties.Resources.EditLevelMismatchException, "UndoChanges"));

#if SILVERLIGHT
        SerializationInfo state = _stateStack.Pop();
        OnSetState(state, StateMode.Undo);

        for (var index = 0; index < _fieldData.Length; index++)
        {
          var item = _fieldData[index];
          if (item != null)
          {
            // potential child object
            var child = item.Value as IUndoableObject;
            if (child != null)
            {
              child.UndoChanges(parentEditLevel, parentBindingEdit);
            }
          }
        }
#else
                IFieldData[] state = null;
                using (MemoryStream buffer = new MemoryStream(_stateStack.Pop()))
                {
                    buffer.Position = 0;
                    var formatter = SerializationFormatterFactory.GetFormatter();
                    state = (IFieldData[])(formatter.Deserialize(buffer));
                }

                for (var index = 0; index < _fieldData.Length; index++)
                {
                    var oldItem = state[index];
                    var item = _fieldData[index];
                    if (item != null)
                    {
                        var undoable = item.Value as IUndoableObject;
                        if (undoable != null)
                        {
                            // current value is undoable
                            if (oldItem != null)
                                undoable.UndoChanges(parentEditLevel, parentBindingEdit);
                            else
                                _fieldData[index] = null;
                            continue;
                        }
                    }
                    // restore IFieldData object into field collection
                    _fieldData[index] = oldItem;
                }
#endif
            }
        }

        void Core.IUndoableObject.AcceptChanges(int parentEditLevel, bool parentBindingEdit)
        {
            if (this.EditLevel - 1 != parentEditLevel)
                throw new UndoException(string.Format(Properties.Resources.EditLevelMismatchException, "AcceptChanges"));

            if (EditLevel > 0)
            {
                // discard latest recorded state
                _stateStack.Pop();

                foreach (var item in _fieldData)
                {
                    if (item != null)
                    {
                        var child = item.Value as IUndoableObject;
                        if (child != null)
                        {
                            // cascade call to child
                            child.AcceptChanges(parentEditLevel, parentBindingEdit);
                        }
                    }
                }
            }
        }

        #endregion

        #region  Child Objects

        /// <summary>
        /// 返回子对象集合        
        /// </summary>
        /// <remarks>
        /// 该方法返回子对象，而不是返回容器<see cref="IFieldData" />对象        
        /// </remarks>
        public List<object> GetChildren()
        {
            List<object> result = new List<object>();
            foreach (var item in _fieldData)
                if (item != null && (item.Value is IEditableBusinessObject || item.Value is IEditableCollection))
                    result.Add(item.Value);
            return result;
        }

        /// <summary>
        /// 调用数据门户提交所有子对象的更新。        
        /// </summary>
        public void UpdateChildren(params object[] parameters)
        {
            foreach (var item in _fieldData)
            {
                if (item != null)
                {
                    object obj = item.Value;
                    if (obj is IEditableBusinessObject || obj is IEditableCollection)
                        //Wcdf.DataPortal.UpdateChild(obj, parameters);
                        Wcdf.DataPortal.Update(obj);
                }
            }
        }

        #endregion

#if SILVERLIGHT
        #region IMobileObject Members

    /// <summary>
    /// Override this method to insert your field values
    /// into the MobileFormatter serialzation stream.
    /// </summary>
    /// <param name="info">
    /// Object containing the data to serialize.
    /// </param>
    /// <param name="mode">
    /// The StateMode indicating why this method was invoked.
    /// </param>
    protected override void OnGetState(SerializationInfo info, StateMode mode)
    {
      info.AddValue("_businessObjectType", _businessObjectType);

      if (mode == StateMode.Serialization && _stateStack.Count > 0)
      {
        MobileList<SerializationInfo> list = new MobileList<SerializationInfo>(_stateStack.ToArray());
        byte[] xml = MobileFormatter.Serialize(list);
        info.AddValue("_stateStack", xml);
      }

      foreach (IFieldData data in _fieldData)
      {
        if (data != null)
        {
          if (data.Value is IUndoableObject)
            info.AddValue("child_" + data.Name, true, false);
          else if (mode == StateMode.Undo && data.Value is IMobileObject) // is IMobileObject but isn't IUndoableObject (such as SmartDate)
            info.AddValue(data.Name, MobileFormatter.Serialize(data.Value), data.IsDirty);
          else if(!(data.Value is IMobileObject))
            info.AddValue(data.Name, data.Value, data.IsDirty);
            
        }
      }

      base.OnGetState(info, mode);
    }

    /// <summary>
    /// Serializes child objects.
    /// </summary>
    /// <param name="info">Serialization state</param>
    /// <param name="formatter">Serializer instance</param>
    protected override void OnGetChildren(SerializationInfo info, MobileFormatter formatter)
    {
      foreach (IFieldData data in _fieldData)
      {
        if (data != null)
        {
          IMobileObject mobile = data.Value as IMobileObject;
          if (mobile != null)
          {
            SerializationInfo childInfo = formatter.SerializeObject(mobile);
            info.AddChild(data.Name, childInfo.ReferenceId, data.IsDirty);
          }
        }
      }

      base.OnGetChildren(info, formatter);
    }

    /// <summary>
    /// Override this method to retrieve your field values
    /// from the MobileFormatter serialzation stream.
    /// </summary>
    /// <param name="info">
    /// Object containing the data to serialize.
    /// </param>
    /// <param name="mode">
    /// The StateMode indicating why this method was invoked.
    /// </param>
    protected override void OnSetState(SerializationInfo info, StateMode mode)
    {
      string type = (string)info.Values["_businessObjectType"].Value;
      Type businessObjecType = Csla.Reflection.MethodCaller.GetType(type);
      SetPropertyList(businessObjecType);

      if (mode == StateMode.Serialization)
      {
        _stateStack.Clear();
        if (info.Values.ContainsKey("_stateStack"))
        {
          //string xml = info.GetValue<string>("_stateStack");
          byte[] xml = info.GetValue<byte[]>("_stateStack");
          MobileList<SerializationInfo> list = (MobileList<SerializationInfo>)MobileFormatter.Deserialize(xml);
          SerializationInfo[] layers = list.ToArray();
          Array.Reverse(layers);
          foreach (SerializationInfo layer in layers)
            _stateStack.Push(layer);
        }
      }

      // Only clear this list on serialization, otherwise you'll lose
      // your children during an undo.
      if (mode == StateMode.Serialization)
        _fieldData = new IFieldData[_propertyList.Count];
      
      foreach (IPropertyInfo property in _propertyList)
      {
        if (info.Values.ContainsKey(property.Name))
        {
          SerializationInfo.FieldData value = info.Values[property.Name];

          IFieldData data = GetOrCreateFieldData(property);
          if (value.Value != null &&
            mode == StateMode.Undo &&
            typeof(IMobileObject).IsAssignableFrom(property.Type) &&
            !typeof(IUndoableObject).IsAssignableFrom(property.Type))
          {
            data.Value = MobileFormatter.Deserialize((byte[])value.Value);
          }
          else data.Value = value.Value;

          if (!value.IsDirty)
            data.MarkClean();
        }
        else if (mode == StateMode.Undo && !((property.RelationshipType & RelationshipTypes.PrivateField) == RelationshipTypes.PrivateField))
        {
          IFieldData data = GetFieldData(property);
          if (data != null)
          {
            if (!info.Values.ContainsKey("child_" + property.Name) || !info.GetValue<bool>("child_" + property.Name))
              _fieldData[property.Index] = null;
            
            // We don't want to reset children during an undo.
            else if (!typeof(IMobileObject).IsAssignableFrom(data.Value.GetType()))
              data.Value = property.DefaultValue;
          }
        }
      }

      base.OnSetState(info, mode);
    }

    /// <summary>
    /// Deserializes child objects.
    /// </summary>
    /// <param name="info">Serialization state</param>
    /// <param name="formatter">Serializer instance</param>
    protected override void OnSetChildren(SerializationInfo info, MobileFormatter formatter)
    {
      foreach (IPropertyInfo property in _propertyList)
      {
        if (info.Children.ContainsKey(property.Name))
        {
          SerializationInfo.ChildData childData = info.Children[property.Name];

          IFieldData data = GetOrCreateFieldData(property);
          data.Value = formatter.GetObject(childData.ReferenceId);
          if (!childData.IsDirty)
            data.MarkClean();
        }
      }

      base.OnSetChildren(info, formatter);
    }

        #endregion
#else
        #region IMobileObject Members

        void IMobileObject.GetState(SerializationInfo info)
        {
            info.AddValue("_businessObjectType", _businessObjectType);

            foreach (IFieldData data in _fieldData)
            {
                if (data != null)
                {
                    IMobileObject mobile = data.Value as IMobileObject;
                    if (mobile == null)
                        info.AddValue(data.Name, data.Value, data.IsDirty);
                }
            }

            OnGetState(info);
        }

        void IMobileObject.GetChildren(SerializationInfo info, MobileFormatter formatter)
        {
            foreach (IFieldData data in _fieldData)
            {
                if (data != null)
                {
                    IMobileObject mobile = data.Value as IMobileObject;
                    if (mobile != null)
                    {
                        SerializationInfo childInfo = formatter.SerializeObject(mobile);
                        info.AddChild(data.Name, childInfo.ReferenceId, data.IsDirty);
                    }
                }
            }

            OnGetChildren(info, formatter);
        }

        void IMobileObject.SetState(SerializationInfo info)
        {
            string type = (string)info.Values["_businessObjectType"].Value;
            Type businessObjecType = Type.GetType(type);
            SetPropertyList(businessObjecType);
            _fieldData = new IFieldData[_propertyList.Count];

            foreach (IPropertyInfo property in _propertyList)
            {
                if (info.Values.ContainsKey(property.Name))
                {
                    SerializationInfo.FieldData value = info.Values[property.Name];

                    IFieldData data = GetOrCreateFieldData(property);
                    data.Value = value.Value;
                    if (!value.IsDirty)
                        data.MarkClean();
                }
            }

            OnSetState(info);
        }

        void IMobileObject.SetChildren(SerializationInfo info, MobileFormatter formatter)
        {
            foreach (IPropertyInfo property in _propertyList)
            {
                if (info.Children.ContainsKey(property.Name))
                {
                    SerializationInfo.ChildData childData = info.Children[property.Name];

                    IFieldData data = GetOrCreateFieldData(property);
                    data.Value = formatter.GetObject(childData.ReferenceId);
                    if (!childData.IsDirty)
                        data.MarkClean();
                }
            }
            OnSetChildren(info, formatter);
        }

        /// <summary>
        /// Override this method to insert your field values
        /// into the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        protected virtual void OnGetState(SerializationInfo info) { }

        /// <summary>
        /// Override this method to retrieve your field values
        /// from the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        protected virtual void OnSetState(SerializationInfo info) { }

        /// <summary>
        /// Override this method to insert your child object
        /// references into the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        /// <param name="formatter">
        /// Reference to MobileFormatter instance. Use this to
        /// convert child references to/from reference id values.
        /// </param>
        protected virtual void OnGetChildren(SerializationInfo info, MobileFormatter formatter) { }

        /// <summary>
        /// Override this method to retrieve your child object
        /// references from the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        /// <param name="formatter">
        /// Reference to MobileFormatter instance. Use this to
        /// convert child references to/from reference id values.
        /// </param>
        protected virtual void OnSetChildren(SerializationInfo info, MobileFormatter formatter) { }

        #endregion
#endif

        #region Force Static Field Init

        /// <summary>
        /// 强制用对象类型中声明为静态类型的字段的类型或者基类进行初始化
        /// </summary>
        /// <param name="type">Type of object to initialize.</param>
        public static void ForceStaticFieldInit(Type type)
        {
#if SILVERLIGHT
      var attr =
        System.Reflection.BindingFlags.Static |
        System.Reflection.BindingFlags.Public |
        System.Reflection.BindingFlags.DeclaredOnly;
#else
            var attr =
              System.Reflection.BindingFlags.Static |
              System.Reflection.BindingFlags.Public |
              System.Reflection.BindingFlags.DeclaredOnly |
              System.Reflection.BindingFlags.NonPublic;
#endif
            var t = type;
            while (t != null)
            {
                var fields = t.GetFields(attr);
                if (fields.Length > 0)
                    fields[0].GetValue(null);
                t = t.BaseType;
            }
        }

        #endregion
    }
}
