﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Wcdf.Core
{
    /// <summary>
    /// 定义了对象自身状态跟踪有关的属性.    
    /// </summary>
    public interface ITrackStatus : INotifyBusy
    {
        /// <summary>
        /// 如果对象与子对象当前都为可用,则返回<see langword="true" />,
        /// 如果对象或子对象违反规则约束则不可用,返回<see langword="false" />.        
        /// </summary>
        /// <remarks>
        /// <para>
        /// 默认情况下,这个属性回应业务对象中的ValidationRules对象,用于跟踪对象是否违反业务规则        
        /// </para><para>
        /// 你可以重写该属性以提供更复杂的操作实现.例如,你应该总是重写该属性,如果你的对象具有子对象且对象的可性受
        /// 所以子对象的可用性影响.    
        /// </para>
        /// </remarks>
        /// <returns>指示对象是否可用.</returns>
        bool IsValid { get; }
        /// <summary>
        /// 如果对象当前都为可用,则返回<see langword="true" />,
        /// 如果对象违反规则约束则不可用,返回<see langword="false" />.                
        /// </summary>
        /// <remarks>
        /// <para>
        /// 默认情况下,这个属性回应业务对象中的ValidationRules对象,用于跟踪对象是否违反业务规则   
        /// </para><para>
        ///  你可以重写该属性以提供更复杂的操作实现.
        /// </para>
        /// </remarks>
        /// <returns>指示对象是否可用.</returns>
        bool IsSelfValid { get; }
        /// <summary>
        /// 如果对象数据或者字段,子对象数据发生改变,刚返回<see langword="true" />   
        /// </summary>
        /// <remarks>
        /// <para>
        /// 当对象数据发生改变,框架对改变做记录,认为该对象为'脏数据'.该值用于优化数据更新,
        /// 因为未发生改变的对象不需求更新到数据库中.所以新建对象都被认为是'脏数据'.
        /// 所有标记为删除的对象也被认为是'脏数据'.        
        /// </para><para>
        /// 一但对象数据提交到数据库中,'脏'标记就被清除,对象被认为未发生改变的.
        /// 新近从数据库读取的对象也是被认为未发生改变的        
        /// </para>
        /// </remarks>
        /// <returns>指示对象是否发生改变.</returns>
        bool IsDirty { get; }
        /// <summary>
        /// 如果对象数据或者字段,子对象数据发生改变,刚返回<see langword="true" />   
        /// </summary>
        /// <remarks>
        /// <para>
        /// 当对象数据发生改变,框架对改变做记录,认为该对象为'脏数据'.该值用于优化数据更新,
        /// 因为未发生改变的对象不需求更新到数据库中.所以新建对象都被认为是'脏数据'.
        /// 所有标记为删除的对象也被认为是'脏数据'.        
        /// </para><para>
        /// 一但对象数据提交到数据库中,'脏'标记就被清除,对象被认为未发生改变的.
        /// 新近从数据库读取的对象也是被认为未发生改变的        
        /// </para>
        /// </remarks>
        /// <returns>指示对象是否发生改变.</returns>
        bool IsSelfDirty { get; }
        /// <summary>
        /// 如果对象被标记为删除刚返回<see langword="true" />         
        /// </summary>
        /// <remarks>
        /// 框架支持立即删除及延迟删除.该属性是支持延迟删除的一部分,当对象被删除为删除,
        /// 但并不是真正的删除直到该对象被保存到数据库中.本属性指示当前对象是否已经被
        /// 标记为删除.如果是,则返回<see langword="true" />,在对象保存到数据库后将会删除,
        /// 否则它将会被指定进行其他数据库操作(删除,更新).        
        /// </remarks>
        /// <returns>指示对象是否标记为删除.</returns>
        bool IsDeleted { get; }
        /// <summary>
        /// 如果新对象返回<see langword="true" />
        /// 如果是已经存在过的对象,刚返回<see langword="false" />        
        /// </summary>
        /// <remarks>
        /// 如果对象的主键值不存在于数据库中,刚认为它是一新对象.换句话说,如果数据还未保存到数据库中
        /// 就是为新.同时,如果对象数据已经从数据库删除,同该对象也被认为是新的.        
        /// </remarks>
        /// <returns>指示对象是否为新对象.</returns>
        bool IsNew { get; }
        /// <summary>
        /// 如果对象是脏标记和可用标记,刚返回<see langword="true" />        
        /// </summary>
        /// <remarks>
        /// 对象为脏如果<see cref="P:Wcdf.BusinessBase.IsDirty" />返回<see langword="true" />.
        /// 它被认为可用如果IsValid返回真.IsSavable属性由这两个属性组合成的        
        /// </remarks>
        /// <returns>指示对象是否可保存.</returns>
        bool IsSavable { get; }
    }
}
