﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Collections.Specialized;
using System.IO;
using Wcdf.Serialization.Mobile;
using Wcdf.Properties;
using Wcdf.Reflection;
using Wcdf.Serialization;
using System.Reflection;

namespace Wcdf.Core
{
    /// <summary>
    /// 具有N层撤消能力的类.    
    /// </summary>
    [Serializable()]
    public abstract class UndoableBase : Wcdf.Core.BindableBase,
      Wcdf.Core.IUndoableObject
    {
        // keep a stack of object state values.
        [NotUndoable()]
        private Stack<byte[]> _stateStack = new Stack<byte[]>();
        [NotUndoable]
        private bool _bindingEdit;

        /// <summary>
        /// Creates an instance of the object.
        /// </summary>
        protected UndoableBase()
        {

        }

        /// <summary>
        /// 获取和设置一个值标记是否是通过IEditableObject接口调用N层撤消通过IEditableObject处理方式有区别,
        /// 在BusinessBase中.会继承IEditableObject接口实现对数据绑定的调用支持.
        /// 
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected bool BindingEdit
        {
            get
            {
                return _bindingEdit;
            }
            set
            {
                _bindingEdit = value;
            }
        }

        int IUndoableObject.EditLevel
        {
            get { return EditLevel; }
        }

        /// <summary>
        /// 返回对象的编辑层数.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected int EditLevel
        {
            get { return _stateStack.Count; }
        }

        #region IUndoableObject接口显式实现
        void IUndoableObject.CopyState(int parentEditLevel, bool parentBindingEdit)
        {
            if (!parentBindingEdit)
                CopyState(parentEditLevel);
        }

        void IUndoableObject.UndoChanges(int parentEditLevel, bool parentBindingEdit)
        {
            if (!parentBindingEdit)
                UndoChanges(parentEditLevel);
        }

        void IUndoableObject.AcceptChanges(int parentEditLevel, bool parentBindingEdit)
        {
            if (!parentBindingEdit)
                AcceptChanges(parentEditLevel);
        }
        #endregion

        #region CopyState 相关方法
        /// <summary>
        /// 可重写的在CopyState操作开始前被调用方法
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void CopyingState()
        {
        }

        /// <summary>
        /// 可重写的在CopyState操作结束后被调用方法.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void CopyStateComplete()
        {
        }

        /// <summary>
        /// 将目前的对象快照置于堆栈顶        
        /// </summary>
        /// <remarks>
        /// 该方法是具体的实现方法,IUndoableObject中的接口方法调用本方法.
        /// </remarks>     
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected internal void CopyState(int parentEditLevel)
        {
            CopyingState();

            Type currentType = this.GetType();
            HybridDictionary state = new HybridDictionary();

            if (this.EditLevel + 1 > parentEditLevel)
                throw new UndoException(string.Format(Resources.EditLevelMismatchException, "CopyState"));

            do
            {
                var currentTypeName = currentType.FullName;
                // get the list of fields in this type
                List<DynamicMemberHandle> handlers =
                  UndoableHandler.GetCachedFieldHandlers(currentType);
                foreach (var h in handlers)
                {
                    var value = h.DynamicMemberGet(this);
                    if (typeof(Wcdf.Core.IUndoableObject).IsAssignableFrom(h.MemberType))
                    {
                        // make sure the variable has a value
                        if (value == null)
                        {
                            // variable has no value - store that fact
                            state.Add(GetFieldName(currentTypeName, h.MemberName), null);
                        }
                        else
                        {
                            // this is a child object, cascade the call
                            ((Core.IUndoableObject)value).CopyState(this.EditLevel + 1, BindingEdit);
                        }
                    }
                    else
                    {
                        // this is a normal field, simply trap the value
                        state.Add(GetFieldName(currentTypeName, h.MemberName), value);
                    }
                }

                currentType = currentType.BaseType;
            } while (currentType != typeof(UndoableBase));

            // serialize the state and stack it
            using (MemoryStream buffer = new MemoryStream())
            {
                ISerializationFormatter formatter =
                  SerializationFormatterFactory.GetFormatter();
                formatter.Serialize(buffer, state);
                _stateStack.Push(buffer.ToArray());
            }
            CopyStateComplete();
        }

        #endregion

        #region UndoChanges相关
        /// <summary>
        /// 可重写的在UndoChanges操作完成时被调用方法.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void UndoChangesComplete()
        {
        }

        /// <summary>
        /// 可重写的在UndoChanges操作开始时被调用方法.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void UndoingChanges()
        {
        }

        /// <summary>
        /// 用堆栈顶的对象快照还原对象.        
        /// </summary>
        /// <remarks>
        /// 对象返回上个版本的字段值.堆栈顶端值被取出.        
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected internal void UndoChanges(int parentEditLevel)
        {
            UndoingChanges();

            // if we are a child object we might be asked to
            // undo below the level of stacked states,
            // so just do nothing in that case
            if (EditLevel > 0)
            {
                if (this.EditLevel - 1 != parentEditLevel)
                    throw new UndoException(string.Format(Resources.EditLevelMismatchException, "UndoChanges"));

                HybridDictionary state;
                using (MemoryStream buffer = new MemoryStream(_stateStack.Pop()))
                {
                    buffer.Position = 0;
                    ISerializationFormatter formatter =
                      SerializationFormatterFactory.GetFormatter();
                    state = (HybridDictionary)formatter.Deserialize(buffer);
                }

                Type currentType = this.GetType();

                do
                {
                    var currentTypeName = currentType.FullName;

                    // get the list of fields in this type
                    List<DynamicMemberHandle> handlers = UndoableHandler.GetCachedFieldHandlers(currentType);
                    foreach (var h in handlers)
                    {
                        // the field is undoable, so restore its value
                        var value = h.DynamicMemberGet(this);

                        if (typeof(Wcdf.Core.IUndoableObject).IsAssignableFrom(h.MemberType))
                        {
                            // this is a child object
                            // see if the previous value was empty
                            //if (state.Contains(h.MemberName))
                            if (state.Contains(GetFieldName(currentTypeName, h.MemberName)))
                            {
                                // previous value was empty - restore to empty
                                h.DynamicMemberSet(this, null);
                            }
                            else
                            {
                                // make sure the variable has a value
                                if (value != null)
                                {
                                    // this is a child object, cascade the call.
                                    ((Core.IUndoableObject)value).UndoChanges(this.EditLevel,
                                                                               BindingEdit);
                                }
                            }
                        }
                        else
                        {
                            // this is a regular field, restore its value
                            h.DynamicMemberSet(this, state[GetFieldName(currentTypeName, h.MemberName)]);
                            //h.DynamicMemberSet(this, state[h.MemberName]);
                        }
                    }

                    currentType = currentType.BaseType;
                } while (currentType != typeof(UndoableBase));
            }
            UndoChangesComplete();
        }

        #endregion

        #region AcceptingChanges相关方法
        /// <summary>
        /// This method is invoked before the AcceptChanges
        /// operation begins.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void AcceptingChanges()
        {
        }

        /// <summary>
        /// This method is invoked after the AcceptChanges
        /// operation is complete.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected virtual void AcceptChangesComplete()
        {
        }

        /// <summary>
        /// Accepts any changes made to the object since the last
        /// state copy was made.
        /// </summary>
        /// <remarks>
        /// The most recent state copy is removed from the state
        /// stack and discarded, thus committing any changes made
        /// to the object's state.
        /// </remarks>
        [EditorBrowsable(EditorBrowsableState.Never)]
        protected internal void AcceptChanges(int parentEditLevel)
        {
            AcceptingChanges();

            if (EditLevel > 0)
            {
                if (this.EditLevel - 1 != parentEditLevel)
                    throw new UndoException(string.Format(Resources.EditLevelMismatchException, "AcceptChanges"));

                _stateStack.Pop();
                Type currentType = this.GetType();

                do
                {
                    // get the list of fields in this type
                    List<DynamicMemberHandle> handlers = UndoableHandler.GetCachedFieldHandlers(currentType);
                    foreach (var h in handlers)
                    {
                        // the field is undoable so see if it is a child object
                        if (typeof(Wcdf.Core.IUndoableObject).IsAssignableFrom(h.MemberType))
                        {
                            object value = h.DynamicMemberGet(this);
                            // make sure the variable has a value
                            if (value != null)
                            {
                                // it is a child object so cascade the call
                                ((Core.IUndoableObject)value).AcceptChanges(this.EditLevel, BindingEdit);
                            }
                        }
                    }

                    currentType = currentType.BaseType;
                } while (currentType != typeof(UndoableBase));
            }
            AcceptChangesComplete();
        }
        #endregion

        #region Helper Functions

        private static bool NotUndoableField(FieldInfo field)
        {
            return Attribute.IsDefined(field, typeof(NotUndoableAttribute));
        }

        /// <summary>
        /// 取得字段的完整限定名."类名.字段名"        
        /// </summary>
        /// <param name="typeName">类名.</param>
        /// <param name="memberName">成员字段名.</param>
        private static string GetFieldName(string typeName, string memberName)
        {
            return typeName + "." + memberName;
        }

        private static string GetFieldName(FieldInfo field)
        {
            return field.DeclaringType.FullName + "!" + field.Name;
        }

        #endregion

        #region  Reset child edit level

        internal static void ResetChildEditLevel(IUndoableObject child, int parentEditLevel, bool bindingEdit)
        {
            int targetLevel = parentEditLevel;
            if (bindingEdit && targetLevel > 0 && !(child is FieldManager.FieldDataManager))
                targetLevel--;
            // if item's edit level is too high,
            // reduce it to match list
            while (child.EditLevel > targetLevel)
                child.AcceptChanges(targetLevel, false);
            // if item's edit level is too low,
            // increase it to match list
            while (child.EditLevel < targetLevel)
                child.CopyState(targetLevel, false);
        }

        #endregion

        #region MobileObject overrides

        /// <summary>
        /// Override this method to insert your field values
        /// into the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        /// <param name="mode">
        /// The StateMode indicating why this method was invoked.
        /// </param>
        protected override void OnGetState(SerializationInfo info, StateMode mode)
        {
            info.AddValue("_bindingEdit", _bindingEdit);
            base.OnGetState(info, mode);
        }

        /// <summary>
        /// Override this method to retrieve your field values
        /// from the MobileFormatter serialzation stream.
        /// </summary>
        /// <param name="info">
        /// Object containing the data to serialize.
        /// </param>
        /// <param name="mode">
        /// The StateMode indicating why this method was invoked.
        /// </param>
        protected override void OnSetState(SerializationInfo info, StateMode mode)
        {
            _stateStack.Clear();

            _bindingEdit = info.GetValue<bool>("_bindingEdit");
            base.OnSetState(info, mode);
        }

        #endregion
    }
}
