﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;

namespace Wcdf.Data
{
    /// <summary>
    /// SQL中的过滤类,在SQL中的过滤条件都需要分解成二元或者一元表达式.
    /// </summary>
    /// <remarks>
    /// 当同时设置Field属性和字段
    /// </remarks>
    [Serializable]
    public class Filter
    {
        /// <summary>
        /// 条件间的关系 And 或者 Or
        /// </summary>
        public FilterRelation Relation { get; set; }

        /// <summary>
        /// 条件语句.
        /// </summary>
        public string Condition { get; set; }

        public List<Filter> ChildFilter = new List<Filter>();

        #region Construction
        /// <summary>
        /// 默认的构造,条件之间为And关系
        /// </summary>
        public Filter()
            : this(FilterRelation.And)
        {
        }
        /// <summary>
        /// 单Filter过滤设置,该类不在客户端直接初始化.应该调用Criteria方法.
        /// </summary>
        /// <param name="op"></param>
        /// <param name="field"></param>
        /// <param name="value"></param>        
        public Filter(FilterRelation relation)
        {
            this.Relation = relation;
        }

        /// <summary>
        /// 直接添加sql条件语句做为过滤对象.
        /// 采用本构造方法.表示为子条件.
        /// </summary>
        /// <param name="sqlStatement"></param>
        public Filter(string condition)
        {
            Condition = condition;
        }

        #endregion

        public bool IsEmpty()
        {
            return this.ChildFilter.Count == 0 && string.IsNullOrEmpty(this.Condition);
        }

        public void Clear()
        {
            this.Condition = null;
            this.ChildFilter.Clear();
        }

        public void SetCondition(string condition)
        {
            this.Condition = condition;
        }

        public void SetCondition<TBusiness>(Expression<Func<TBusiness, bool>> predicate)
        {
            ConditionBuilder conditionBuilder = new ConditionBuilder();
            conditionBuilder.Build(predicate.Body);
            this.Condition = conditionBuilder.Condition;
        }

        #region 字符过滤条件生成方法
        /// <summary>
        /// 根据映射信息,将filter转换为查询SQL过滤条件语言
        /// </summary>
        /// <param name="filter"></param>
        /// <returns></returns>
        public string GetCondition()
        {
            return GetCondition(this);
        }

        private string GetCondition(Filter filter)
        {
            //none child and field info
            if (filter.IsEmpty())
                return string.Empty;
            //如果直接是过滤语句,则返回该语句.
            if (!string.IsNullOrEmpty(filter.Condition))
            {
                return filter.Condition;
            }
            StringBuilder sb = new StringBuilder();
            sb.Append("(");
            string rel = " " + filter.Relation.ToString() + " ";
            string cf = string.Empty;
            bool isChildSql = false;
            this.ChildFilter.ForEach(delegate(Filter child)
            {
                cf = child.GetCondition();
                if (!string.IsNullOrEmpty(cf))
                {
                    sb.Append(cf);
                    sb.Append(rel);
                    isChildSql = true;
                }
            }
            );
            if (isChildSql)
                //移除最后一个关系运算符    
                sb.Remove(sb.Length - rel.Length, rel.Length);
            sb.Append(")");
            if (sb.ToString() == "()")
            {
                return string.Empty;
            }
            return sb.ToString();
        }
        #endregion
    }
}
