﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;
using System.Collections.ObjectModel;
using Wcdf.Core.FieldManager;
using System.Data.Common;
using Wcdf.Data.SqlCommandBuilder;
using System.Data;
using Wcdf.Core;

namespace Wcdf.Data
{
    /// <summary>
    /// 数据访问通过操作类.
    /// </summary>
    public class DataAccess
    {
        /// <summary>
        /// 默认的连接字符串配置字节名
        /// </summary>
        internal const string DefaultDataBase = "DataBase";

        /// <summary>
        /// 数据提供者
        /// </summary>
        private string ProviderName { get; set; }

        /// <summary>
        /// 连接字符串
        /// </summary>
        private string ConnectionString { get; set; }

        /// <summary>
        /// 本DataAccess指向的连接配置.
        /// </summary>
        public string ConfigName { get; set; }

        private Collection<ISqlCommand> SqlPool;
        /// <summary>
        /// 字段信息
        /// </summary>
        internal PropertyInfoList PropertyInfos { get; set; }

        #region constructor
        /// <summary>
        /// 默认构造方法,为配置的第一个连接串.
        /// </summary>
        public DataAccess()
            : this(null, DefaultDataBase)
        {
        }

        /// <summary>
        /// 
        /// </summary>
        public DataAccess(string database)
            : this(null, database)
        {
        }

        public DataAccess(Type type)
            : this(type, DefaultDataBase)
        {
            Init(type);
        }
        /// <summary>
        /// 通过连接配置初始化
        /// </summary>
        /// <param name="database"></param>
        public DataAccess(Type type, string database)
        {
            //ApplicationContext.LocalContext[database] = ConfigurationManager;
            Init(type);
            this.ConfigName = string.IsNullOrEmpty(database) ? DefaultDataBase : database;
            this.ProviderName = ConfigurationManager.ConnectionStrings[ConfigName].ProviderName;
            this.ConnectionString = ConfigurationManager.ConnectionStrings[ConfigName].ConnectionString;
            SqlPool = new Collection<ISqlCommand>();
        }
        #endregion
        /// <summary>
        /// 初始化字段信息
        /// </summary>
        /// <param name="type"></param>
        public void Init(Type type)
        {
            if (type != null)
                PropertyInfos = Wcdf.Core.FieldManager.PropertyInfoManager.GetRegisteredProperties(type);
        }
        /// <summary>
        /// 获取指定条件的数据库记录,并转化为业务对象.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="biz"></param>
        /// <param name="criteria"></param>
        /// <returns></returns>
        public T Fetch<T>(T biz, object criteria) where T : Wcdf.Core.BusinessBase
        {
            string tableName = biz.GetDataObjectName();
            //非数据表.或者为BusinessBaseList.
            if (string.IsNullOrEmpty(tableName))
                return biz;

            Criteria cri = criteria as Criteria;
            if (criteria == null)
            {
                cri = new Criteria();
            }
            using (TransactionManager trm = TransactionManager.GetManager(this.ConfigName, false))
            {
                DbCommand cmd = trm.Connection.CreateCommand();
                QueryCommand comm = new QueryCommand(tableName, cri.Columns, PropertyInfos, cri);
                comm.BuildCommand(cmd);
                using (DbDataReader reader = cmd.ExecuteReader(CommandBehavior.CloseConnection))
                {
                    while (reader.Read())
                    {
                        IPropertyInfo pInfo;
                        for (int i = 0; i < reader.FieldCount; i++)
                        {
                            pInfo = this.PropertyInfos.FirstOrDefault(op => op.FriendlyName == reader.GetName(i));
                            if (pInfo != null)
                            {
                                if (reader[i] is DBNull)
                                    (biz as IManageProperties).LoadProperty(pInfo, null);
                                else
                                    (biz as IManageProperties).LoadProperty(pInfo, reader[i]);
                            }
                            //reader.g
                        }
                        //只取第一条记录.
                        break;
                    }
                }
                trm.Commit();
            }
            return biz;
        }

        /// <summary>
        /// 获取集合类型的业务对象列表.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="C"></typeparam>
        /// <param name="obj"></param>
        /// <param name="list"></param>
        /// <param name="criteria"></param>
        /// <returns></returns>
        public T Fetch<T, C>(T list, object criteria)
            where C : Wcdf.Core.IEditableBusinessObject
            where T : Wcdf.Core.IEditableCollection
        {
            Criteria cri = criteria as Criteria;
            if (criteria == null)
            {
                cri = new Criteria();
            }
            Type ctype = typeof(C);
            var item = Wcdf.Reflection.MethodCaller.CreateInstance(ctype);

            this.PropertyInfos = new PropertyInfoList((item as BusinessBase).GetFieldManager().GetRegisteredProperties());
            string tableName = (item as BusinessBase).GetDataObjectName();
            using (TransactionManager trm = TransactionManager.GetManager(this.ConfigName, false))
            {
                DbCommand cmd = trm.Connection.CreateCommand();
                QueryCommand comm = new QueryCommand(tableName, cri.Columns, PropertyInfos, cri);
                comm.BuildCommand(cmd);
                using (DbDataReader reader = cmd.ExecuteReader(CommandBehavior.CloseConnection))
                {
                    while (reader.Read())
                    {
                        var biz = Wcdf.Reflection.MethodCaller.CreateInstance(ctype) as IEditableBusinessObject;
                        var target = biz as IManageProperties;
                        IPropertyInfo pInfo;
                        for (int i = 0; i < reader.FieldCount; i++)
                        {
                            pInfo = this.PropertyInfos.FirstOrDefault(op => op.FriendlyName == reader.GetName(i));
                            if (reader[i] is DBNull)
                                (biz as IManageProperties).LoadProperty(pInfo, null);
                            else
                                (biz as IManageProperties).LoadProperty(pInfo, reader[i]);
                        }
                        //在列表前将子对象标识，考虑效率，并不在业务列表对象中Fetch方法中实现。
                        (biz as Wcdf.Server.IDataPortalTarget).MarkOld();
                        (biz as Wcdf.Server.IDataPortalTarget).MarkAsChild();
                        list.AddChild(biz);
                    }
                }
                trm.Commit();
            }
            return list;
        }

        #region Update Method
        /// <summary>
        /// 将业务对象更新到数据库
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="obj"></param>
        /// <returns></returns>
        public T Update<T>(T obj)
        {
            if (obj is Core.BusinessBase)
            {
                var busObj = obj as Core.BusinessBase;
                UpdatePrepare(busObj);
                this.Flush();
            }
            else
            {
                //集合对象

            }
            return obj;

        }

        /// <summary>
        /// 业务对象更新
        /// </summary>
        /// <param name="busObj"></param>
        private void UpdatePrepare(Core.BusinessBase busObj)
        {
            if (busObj.IsDeleted)
            {
                if (!busObj.IsNew)
                {
                    DeletePrepare(busObj);
                }
            }
            else
            {
                if (busObj.IsNew)
                {
                    InsertPrepare(busObj);
                }
                else
                {
                    //更新
                    UpdateCommand updcomm = new UpdateCommand(busObj.GetDataObjectName(), this.PropertyInfos, busObj.GetFieldManager());
                    this.SqlPool.Add(updcomm);
                }
            }
        }
        #endregion

        #region Delete Method
        /// <summary>
        /// 从数据库删除指定的业务对象数据.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="criteria"></param>
        public void Delete<T>(object criteria)
        {

        }

        /// <summary>
        /// 删除指定的业务对象的数据.
        /// </summary>
        /// <param name="tableName"></param>
        /// <param name="criteria"></param>
        private void Delete(string tableName, object criteria)
        {
            var cri = criteria as Criteria;
            DeleteCommand comm = new DeleteCommand(tableName, cri);
            this.SqlPool.Add(comm);
        }

        /// <summary>
        /// 从数据库中删除对应业务对象的数据.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="obj"></param>
        private void DeletePrepare(Core.BusinessBase obj)
        {
            var busObj = obj as Core.BusinessBase;
            //删除
            Criteria cri = new Criteria();
            var mangeField = obj as IManageProperties;
            var keysProperty = this.PropertyInfos.FindAll(o => o.RelationshipType == RelationshipTypes.PrimaryKey);
            keysProperty.ForEach((O) =>
            {
                object value = mangeField.ReadProperty(O);
                cri.AddFilter(string.Format("{0}={1}", O.FriendlyName, TypeResolver.ResolveToSqlString(value)));
            }
            );
            this.Delete(busObj.GetDataObjectName(), cri);
        }
        #endregion

        /// <summary>
        /// 将业务对象数据插入到数据库中.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="obj"></param>
        private void InsertPrepare(Core.BusinessBase obj)
        {
            var busObj = obj as Core.BusinessBase;
            //插入
            var colProperty = this.PropertyInfos.FindAll(o => !string.IsNullOrEmpty(o.FriendlyName));
            FieldDataManager fdm = busObj.GetFieldManager();
            InsertCommand comm = new InsertCommand(busObj.GetDataObjectName(), this.PropertyInfos, fdm);
            this.SqlPool.Add(comm);
        }

        private int Flush()
        {
            int rec = 0;
            using (TransactionManager trm = TransactionManager.GetManager(this.ConfigName, false))
            {
                foreach (ISqlCommand sql in this.SqlPool)
                {
                    DbCommand comm = trm.Connection.CreateCommand();
                    comm.Transaction = trm.Transaction;
                    sql.BuildCommand(comm);
                    rec += comm.ExecuteNonQuery();
                }
            }
            this.SqlPool.Clear();
            return rec;
        }

        /// <summary>
        /// 根据sql语句招待,返回DataSet结果,无权限控制,需要在SQL,留出权限语句位置,但字段权限会有问题.
        /// </summary>
        /// <param name="sql"></param>
        /// <returns></returns>
        public DataSet ExecuteDataSet(string sql)
        {
            using (TransactionManager trm = TransactionManager.GetManager(this.ConfigName, false))
            {
                DbCommand cmd = trm.Connection.CreateCommand();
                DbProviderFactory dbfactory = DbProviderFactories.GetFactory(this.ProviderName);
                DbDataAdapter dbDataAdapter = dbfactory.CreateDataAdapter();
                dbDataAdapter.SelectCommand = cmd;
                cmd.CommandText = sql;
                DataSet ds = new DataSet();
                dbDataAdapter.Fill(ds);
                trm.Commit();
                return ds;
            }
        }

        /// <summary>
        /// 执行查询，并返回查询所返回的结果集中第一行的第一列。忽略其他列或行。
        /// </summary>
        /// <param name="sql">需要执行的SQL语句</param>
        /// <returns></returns>
        public object ExecuteScalar(string sql)
        {
            using (TransactionManager trm = TransactionManager.GetManager(this.ConfigName, false))
            {
                DbCommand cmd = trm.Connection.CreateCommand();
                cmd.CommandText = sql;
                object result = cmd.ExecuteScalar();
                trm.Commit();
                return result;
            }
        }

        /// <summary>
        /// 执行SQL语句.
        /// </summary>
        /// <param name="sql"></param>
        /// <returns></returns>
        public int ExecuteNonQuery(string sql)
        {
            using (TransactionManager trm = TransactionManager.GetManager(this.ConfigName, false))
            {
                DbCommand cmd = trm.Connection.CreateCommand();
                cmd.CommandText = sql;
                int result = cmd.ExecuteNonQuery();
                trm.Commit();
                return result;
            }
        }
        /// <summary>
        /// 根据执行语句，返回对象集合
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="C"></typeparam>
        /// <param name="list"></param>
        /// <param name="sql"></param>
        /// <returns></returns>
        public T ExecuteList<T, C>(T list, string sql)
            where C : Wcdf.Core.IEditableBusinessObject
            where T : Wcdf.Core.IEditableCollection
        {
            Type ctype = typeof(C);
            var item = Wcdf.Reflection.MethodCaller.CreateInstance(ctype);

            this.PropertyInfos = new PropertyInfoList((item as BusinessBase).GetFieldManager().GetRegisteredProperties());
            string tableName = (item as BusinessBase).GetDataObjectName();
            using (TransactionManager trm = TransactionManager.GetManager(this.ConfigName, false))
            {
                DbCommand cmd = trm.Connection.CreateCommand();
                cmd.CommandType = CommandType.Text;
                cmd.CommandText = sql;
                using (DbDataReader reader = cmd.ExecuteReader(CommandBehavior.CloseConnection))
                {
                    while (reader.Read())
                    {
                        var biz = Wcdf.Reflection.MethodCaller.CreateInstance(ctype) as IEditableBusinessObject;
                        var target = biz as IManageProperties;
                        IPropertyInfo pInfo;
                        for (int i = 0; i < reader.FieldCount; i++)
                        {
                            pInfo = this.PropertyInfos.FirstOrDefault(op => op.FriendlyName == reader.GetName(i));
                            if (pInfo != null)
                                if (reader[i] is DBNull)
                                    (biz as IManageProperties).LoadProperty(pInfo, null);
                                else
                                    (biz as IManageProperties).LoadProperty(pInfo, reader[i]);
                        }
                        //在列表前将子对象标识，考虑效率，并不在业务列表对象中Fetch方法中实现。
                        (biz as Wcdf.Server.IDataPortalTarget).MarkOld();
                        (biz as Wcdf.Server.IDataPortalTarget).MarkAsChild();
                        list.AddChild(biz);
                    }
                }
                trm.Commit();
            }
            return list;
        }

    }
}
