﻿using System;
using Wcdf.Reflection;
using System.Configuration;

namespace Wcdf.Data
{
    public static class DataMapper
    {
        #region SetValue

        /// <summary>
        /// Sets an object's property with the specified value,
        /// coercing that value to the appropriate type if possible.
        /// </summary>
        /// <param name="target">Object containing the property to set.</param>
        /// <param name="propertyName">Name of the property to set.</param>
        /// <param name="value">Value to set into the property.</param>
        public static void SetPropertyValue(
          object target, string propertyName, object value)
        {
            DynamicMemberHandle handle = MethodCaller.GetCachedProperty(target.GetType(), propertyName);
            SetValueWithCoercion(target, handle, value);
        }

        private static void SetValueWithCoercion(object target, DynamicMemberHandle handle, object value)
        {
            var oldValue = handle.DynamicMemberGet(target);

            Type pType = handle.MemberType;

            if (!pType.IsGenericType
                || (pType.IsGenericType && pType.GetGenericTypeDefinition() != typeof(Nullable<>)))
            {
                if (pType.IsValueType && (pType.IsPrimitive || pType == typeof(decimal)) && value == null)
                {
                    value = 0;
                }
            }

            if (value != null)
            {
                Type vType = Utilities.GetPropertyType(value.GetType());
                value = Utilities.CoerceValue(pType, vType, oldValue, value);
            }
            handle.DynamicMemberSet(target, value);
        }

        /// <summary>
        /// Sets an object's field with the specified value,
        /// coercing that value to the appropriate type if possible.
        /// </summary>
        /// <param name="target">Object containing the field to set.</param>
        /// <param name="fieldName">Name of the field (public or non-public) to set.</param>
        /// <param name="value">Value to set into the field.</param>
        public static void SetFieldValue(
          object target, string fieldName, object value)
        {
            DynamicMemberHandle handle = MethodCaller.GetCachedField(target.GetType(), fieldName);
            SetValueWithCoercion(target, handle, value);
        }

        /// <summary>
        /// Gets an object's field value.
        /// </summary>
        /// <param name="target">Object whose field value to get.</param>
        /// <param name="fieldName">The name of the field.</param>
        /// <returns>The value of the field.</returns>
        public static object GetFieldValue(
          object target, string fieldName)
        {
            DynamicMemberHandle handle = MethodCaller.GetCachedField(target.GetType(), fieldName);
            return handle.DynamicMemberGet.Invoke(target);
        }

        #endregion

        #region 占位符生成方法
        /// <summary>
        /// 占位符生成方法
        /// </summary>
        /// <param name="configName"></param>
        /// <param name="colName"></param>
        /// <returns></returns>
        public static string ParameterCreator(string provideName, string colName)
        {
            //string provideName = ConfigurationManager.ConnectionStrings[configName].ProviderName;
            switch (provideName)
            {
                case "System.Data.SqlClient":
                    return "@" + colName;
                case "System.Data.OracleClient":
                case "Oracle.DataAccess.Client":
                    return ":" + colName;
                case "System.Data.OleDb":
                case "Microsoft.Data.Odbc":
                    return "?";
                case "MySql.Data.MySqlClient":
                    return "?" + colName;
                default:
                    throw new NotSupportedException("不支持的数据访问组件");
            }

        }
        #endregion
    }
}
