﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;
using System.Data.Common;

namespace Wcdf.Data
{
    /// <summary>
    /// 数据库事务及连接管理
    /// </summary>
    /// <typeparam name="C"></typeparam>
    /// <typeparam name="T"></typeparam>
    public class TransactionManager : IDisposable
    {
        private const string noTrans = "noTrans";
        private static object _lock = new object();
        DbConnection _connection;
        DbTransaction _transaction;
        private string _connectionString;
        private string _label;

        /// <summary>
        /// 根据连接字符串及事务标识构造管理类
        /// </summary>     
        /// <param name="providerName">数据库提供程序名称</param>
        /// <param name="connectionString">数据库接连字符串</param>
        /// <param name="label">事务标识,当不需要启动事务时标识为"noTrans"</param>
        private TransactionManager(string providerName, string connectionString, string label)
        {
            DbProviderFactory dbfactory = DbProviderFactories.GetFactory(providerName);
            _connection = dbfactory.CreateConnection();
            //create conn            
            // create and open connection
            _label = label;
            _connectionString = connectionString;
            _connection.ConnectionString = connectionString;
            _connection.Open();
            //启动事务,如果不为noTrans设置
            if (label != noTrans)
                _transaction = _connection.BeginTransaction();
        }

        /// <summary>
        /// 连接对象
        /// </summary>
        public DbConnection Connection
        {
            get { return this._connection; }
        }

        /// <summary>
        /// 事务对象
        /// </summary>
        public DbTransaction Transaction
        {
            get { return this._transaction; }
        }

        #region 事务引用管理
        private int _refCount;
        public int RefCount
        {
            get { return _refCount; }
        }

        /// <summary>
        /// 在本次连接上,增加一次事务
        /// </summary>        
        private void AddRef()
        {
            _refCount += 1;
        }

        /// <summary>
        /// 
        /// </summary>
        private void DeRef()
        {
            lock (_lock)
            {
                _refCount -= 1;
                if (_refCount == 0)
                {
                    //if not need trans
                    if (!_label.Equals(noTrans))
                    {
                        if (_commit)
                            _transaction.Commit();
                        else
                            _transaction.Rollback();

                        _transaction.Dispose();
                    }
                    _connection.Dispose();
                    ApplicationContext.LocalContext.Remove(GetContextName(_connectionString, _label));
                }
                //else if (_label.Equals(noTrans))
                //{
                //    //no need trans直接关闭
                //    _connection.Dispose();
                //    ApplicationContext.Current.LocalContext.Remove(GetContextName(_connectionString, _label));
                //}
            }

        }
        #endregion

        #region 事务提交
        private bool _commit = false;
        public void Commit()
        {
            if (RefCount == 1)
                _commit = true;
        }
        #endregion

        /// <summary>
        /// 拼接上下文标识,以线程ID做为标识
        /// </summary>
        /// <param name="connectionString"></param>
        /// <param name="label"></param>
        /// <returns></returns>
        private static string GetContextName(string connectionString, string label)
        {
            return "__transaction:" + connectionString + label + "-" + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString();
        }

        /// <summary>
        /// 根据默认的接字符串及标识取得事务管理.事务管理器会保存在本地缓存中
        /// </summary>
        /// <param name="configName">数据库配置节点</param>
        /// <returns></returns>
        public static TransactionManager GetManager(string configName)
        {
            return GetManager(ConfigurationManager.ConnectionStrings[configName].ProviderName, ConfigurationManager.ConnectionStrings[configName].ConnectionString, "default");
        }

        /// <summary>
        /// 根据默认的接字符串及标识取得事务管理,该方法不执行事务
        /// </summary>
        /// <param name="configName">数据库配置节点</param>
        /// <param name="isTransaction"></param>
        /// <returns></returns>
        public static TransactionManager GetManager(string configName, bool isTransaction)
        {
            if (isTransaction)
                return GetManager(configName);
            else
                return GetManager(ConfigurationManager.ConnectionStrings[configName].ProviderName, ConfigurationManager.ConnectionStrings[configName].ConnectionString, noTrans);
        }

        /// <summary>
        /// 根据连接字符串及标识取得事务管理.事务管理器会保存在本地缓存中
        /// </summary>
        /// <param name="providerName">数据库类型提供者</param>
        /// <param name="connectionString">连接字符串</param>
        /// <param name="label">事务标识,如果多次起事务,以标识进行区分</param>
        /// <returns></returns>
        public static TransactionManager GetManager(string providerName, string connectionString, string label)
        {
            lock (_lock)
            {
                var contextLabel = GetContextName(connectionString, label);
                TransactionManager mgr = null;
                if (ApplicationContext.LocalContext.Contains(contextLabel))
                {
                    mgr = (TransactionManager)(ApplicationContext.LocalContext[contextLabel]);
                    if (mgr.Connection.State == System.Data.ConnectionState.Closed)
                    {
                        mgr.Connection.Open();
                    }
                }
                else
                {
                    mgr = new TransactionManager(providerName, connectionString, label);
                    ApplicationContext.LocalContext[contextLabel] = mgr;
                }
                mgr.AddRef();
                return mgr;
            }
        }

        #region IDisposable 成员

        public void Dispose()
        {
            DeRef();
        }

        #endregion
    }
}
