﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Wcdf.Serialization.Mobile;
using Wcdf.Reflection;
using System.ComponentModel;
using Wcdf.Server;
using Wcdf.Properties;
using Wcdf.DataPortalClient;

namespace Wcdf
{
    /// <summary>
    /// 客户端数据门户接入口.
    /// </summary>
    public static class DataPortal
    {
        #region DataPortal events

        /// <summary>
        /// Raised by DataPortal before it starts
        /// setting up to call a server-side
        /// DataPortal method.
        /// </summary>
        public static event Action<System.Object> DataPortalInitInvoke;

        /// <summary>
        /// Raised by DataPortal prior to calling the 
        /// requested server-side DataPortal method.
        /// </summary>
        public static event Action<DataPortalEventArgs> DataPortalInvoke;

        /// <summary>
        /// Raised by DataPortal after the requested 
        /// server-side DataPortal method call is complete.
        /// </summary>
        public static event Action<DataPortalEventArgs> DataPortalInvokeComplete;

        private static void OnDataPortalInitInvoke(object e)
        {
            Action<System.Object> action = DataPortalInitInvoke;
            if (action != null)
                action(e);
        }

        private static void OnDataPortalInvoke(DataPortalEventArgs e)
        {
            Action<DataPortalEventArgs> action = DataPortalInvoke;
            if (action != null)
                action(e);
        }

        private static void OnDataPortalInvokeComplete(DataPortalEventArgs e)
        {
            Action<DataPortalEventArgs> action = DataPortalInvokeComplete;
            if (action != null)
                action(e);
        }

        #endregion

        #region Data Access methods

        //private static readonly EmptyCriteria EmptyCriteria = new EmptyCriteria();

        /// <summary>
        /// 调用方法
        /// </summary>
        /// <typeparam name="TChannel">WCF形式声明的接口类型</typeparam>
        /// <param name="action">接口方法委托</param>
        /// <param name="moduleName">配置模块名</param>
        public static void Invoke<TChannel>(System.Action<TChannel> action)
        {
            Invoke(action, null);
        }

        public static void Invoke<TChannel>(System.Action<TChannel> action, string moduleName)
        {
            DataPortalClient.IDataPortalProxy proxy;
            proxy = GetDataPortalProxy(null, false);

            //var dpContext =
            //   new Server.DataPortalContext(GetPrincipal(),
            //   proxy.IsServerRemote);
            proxy.Invoke(action, moduleName);
        }

        /// <summary>
        /// 调用方法
        /// </summary>
        /// <typeparam name="TChannel">WCF形式声明的接口类型</typeparam>
        /// <typeparam name="TReturn">返回类型</typeparam>
        /// <param name="func">接口方法委托</param>
        /// <returns>返回TReturn类型</returns>
        public static TReturn Invoke<TChannel, TReturn>(System.Func<TChannel, TReturn> func)
        {
            return Invoke(func, null);
        }

        /// <summary>
        /// 调用方法
        /// </summary>
        /// <typeparam name="TChannel">WCF形式声明的接口类型</typeparam>
        /// <typeparam name="TReturn">返回类型</typeparam>
        /// <param name="func">接口方法委托</param>
        /// <param name="moduleName">配置模块名</param>
        /// <returns>返回TReturn类型</returns>
        public static TReturn Invoke<TChannel, TReturn>(System.Func<TChannel, TReturn> func, string moduleName)
        {
            DataPortalClient.IDataPortalProxy proxy;
            proxy = GetDataPortalProxy(null, false);

            var dpContext =
               new Server.DataPortalContext(GetPrincipal(),
               proxy.IsServerRemote);
            return proxy.Invoke(func, moduleName);
        }

        /// <summary>
        /// 在业务类中调用工厂方法创建新对象,从数据库中初始化对象        
        /// </summary>
        /// <typeparam name="T">指定的业务对象类型.</typeparam>
        /// <param name="criteria">指定的criteria对象.</param>
        /// <returns>初始化的业务对象.</returns>
        public static T Create<T>(object criteria)
        {
            return (T)Create(typeof(T), criteria);
        }

        /// <summary>
        /// 门户请求创建对象工厂方法.   .
        /// </summary>
        /// <typeparam name="T">指定的业务对象类型.</typeparam>
        /// <returns>初始化的业务对象.</returns>
        public static T Create<T>()
        {
            return (T)Create(typeof(T), null);
        }

        internal static object Create(Type type)
        {
            return Create(type, null);
        }

        /// <summary>
        /// 门户请求创建对象工厂方法.        
        /// </summary>
        /// <param name="objectType">要创建的业务对象类型.</param>
        /// <param name="criteria">指定的criteria对象.</param>
        /// <returns>初始化的业务对象.</returns>
        public static object Create(Type objectType, object criteria)
        {
            Server.DataPortalResult result = null;
            Server.DataPortalContext dpContext = null;
            try
            {
                OnDataPortalInitInvoke(null);

                //if (!Wcdf.Rules.BusinessRules.HasPermission(Rules.AuthorizationActions.CreateObject, objectType))
                //  throw new System.Security.SecurityException(string.Format(
                //    Resources.UserNotAuthorizedException,
                //    "create",
                //    objectType.Name));                
                var method = MethodCaller.GetMethod(objectType, "DataPortal_Create", null);
                bool isLocal = Attribute.IsDefined(method, typeof(RunLocalAttribute), false);
                DataPortalClient.IDataPortalProxy proxy;
                proxy = GetDataPortalProxy(objectType, isLocal);

                dpContext =
                  new Wcdf.Server.DataPortalContext(GetPrincipal(), proxy.IsServerRemote);

                OnDataPortalInvoke(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Create));

                try
                {
                    result = proxy.Create(objectType, criteria, dpContext);
                }
                catch (Server.DataPortalException ex)
                {
                    result = ex.Result;                    
                    throw new DataPortalException(
                      string.Format("DataPortal.Create {0} ({1})", Resources.Failed, ex.InnerException.InnerException),
                      ex.InnerException, result);
                }

                if (proxy.IsServerRemote)
                    ApplicationContext.ContextManager.SetGlobalContext(result.GlobalContext);

                OnDataPortalInvokeComplete(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Create));
            }
            catch (Exception ex)
            {
                OnDataPortalInvokeComplete(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Create, ex));
                throw;
            }
            return result.ReturnObject;
        }

        /// <summary>
        /// 门户请求获取业务对象工厂方法        
        /// </summary>
        /// <typeparam name="T">指定的业务对象类型.</typeparam>
        /// <param name="criteria">指定的criteria对象.</param>
        /// <returns>包含数据库数据的业务对象实例.</returns>
        public static T Fetch<T>(object criteria)
        {
            return (T)Fetch(typeof(T), criteria);
        }

        /// <summary>
        /// 门户请求获取业务对象工厂方法  
        /// </summary>
        /// <typeparam name="T">指定的业务对象类型.</typeparam>
        /// <returns>包含数据库数据的业务对象实例.</returns>
        public static T Fetch<T>()
        {
            return (T)Fetch(typeof(T), null);
        }

        internal static object Fetch(Type objectType)
        {
            return Fetch(objectType, null);
        }

        /// <summary>
        /// 门户请求获取业务对象工厂方法
        /// </summary>
        /// <param name="objectType">指定的业务对象类型</param>
        /// <param name="criteria">指定的criteria对象</param>
        /// <returns>包含数据库数据的业务对象实例</returns>
        public static object Fetch(Type objectType, object criteria)
        {
            Server.DataPortalResult result = null;
            Server.DataPortalContext dpContext = null;
            try
            {
                OnDataPortalInitInvoke(null);

                //if (!Csla.Rules.BusinessRules.HasPermission(Rules.AuthorizationActions.GetObject, objectType))
                //  throw new System.Security.SecurityException(string.Format(Resources.UserNotAuthorizedException,
                //    "get",
                //    objectType.Name));

                //var method = Server.DataPortalMethodCache.GetFetchMethod(objectType, criteria);

                DataPortalClient.IDataPortalProxy proxy;
                proxy = GetDataPortalProxy(objectType, false);

                dpContext =
                  new Server.DataPortalContext(GetPrincipal(),
                  proxy.IsServerRemote);

                OnDataPortalInvoke(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Fetch));

                try
                {
                    result = proxy.Fetch(objectType, criteria, dpContext);
                }
                catch (Server.DataPortalException ex)
                {
                    result = ex.Result;                    
                    string innerMessage = string.Empty;
                    if (ex.InnerException is Wcdf.Reflection.CallMethodException)
                    {
                        if (ex.InnerException.InnerException != null)
                            innerMessage = ex.InnerException.InnerException.Message;
                    }
                    else
                    {
                        innerMessage = ex.InnerException.Message;
                    }
                    throw new DataPortalException(
                      String.Format("DataPortal.Fetch {0} ({1})", Resources.Failed, innerMessage),
                      ex.InnerException, result);
                }

                if (proxy.IsServerRemote)
                    ApplicationContext.ContextManager.SetGlobalContext(result.GlobalContext);

                OnDataPortalInvokeComplete(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Fetch));
            }
            catch (Exception ex)
            {
                OnDataPortalInvokeComplete(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Fetch, ex));
                throw;
            }
            return result.ReturnObject;
        }

        /// <summary>
        /// 本方法由业务对象的Save()方法中调用,向门户提交业务实例,更新数据.
        /// 包括了插入,删除,更新等 数据库操作.        
        /// </summary>
        /// <remarks>
        /// 注意本方法返回更新后的对象引用.如果门户为远程调用,则将会生成一个原实例的副本
        /// 所有的对原对象引用都要指向新生成的对象副本.        
        /// </remarks>
        /// <typeparam name="T">指定的业务对象类型.</typeparam>
        /// <param name="obj">需要更新的业务对象实例.</param>
        /// <returns>返回更新后的业务对象.</returns>
        public static T Update<T>(T obj)
        {
            Server.DataPortalResult result = null;
            Server.DataPortalContext dpContext = null;
            DataPortalOperations operation = DataPortalOperations.Update;
            Type objectType = obj.GetType();
            try
            {
                DataPortalClient.IDataPortalProxy proxy;
                proxy = GetDataPortalProxy(objectType, false);

                dpContext =
                  new Server.DataPortalContext(GetPrincipal(), proxy.IsServerRemote);

                OnDataPortalInvoke(new DataPortalEventArgs(dpContext, objectType, operation));

                try
                {
                    if (!proxy.IsServerRemote && ApplicationContext.AutoCloneOnUpdate)
                    {
                        // when using local data portal, automatically
                        // clone original object before saving
                        ICloneable cloneable = obj as ICloneable;
                        if (cloneable != null)
                            obj = (T)cloneable.Clone();
                    }
                    result = proxy.Update(obj, dpContext);
                }
                catch (Server.DataPortalException ex)
                {
                    result = ex.Result;                    
                    throw new DataPortalException(
                      String.Format("DataPortal.Update {0} ({1})", Resources.Failed, ex.InnerException.InnerException),
                      ex.InnerException, result);
                }

                if (proxy.IsServerRemote)
                    ApplicationContext.ContextManager.SetGlobalContext(result.GlobalContext);

                OnDataPortalInvokeComplete(new DataPortalEventArgs(dpContext, objectType, operation));
            }
            catch (Exception ex)
            {
                OnDataPortalInvokeComplete(new DataPortalEventArgs(dpContext, objectType, operation, ex));
                throw;
            }
            return (T)result.ReturnObject;
        }

        /// <summary>
        /// 看数据库发送删除请求.        
        /// </summary>
        /// <param name="criteria">Object-specific criteria.</param>
        /// <remarks>请求与Update中的对象中的延迟删除不同.并不传对象实例.而是直接删除请求.
        /// </remarks>
        public static void Delete<T>(object criteria)
        {
            Delete(typeof(T), criteria);
        }

        internal static void Delete(Type objectType, object criteria)
        {
            Server.DataPortalResult result = null;
            Server.DataPortalContext dpContext = null;
            try
            {
                OnDataPortalInitInvoke(null);

                //if (!Wcdf.Rules.BusinessRules.HasPermission(Rules.AuthorizationActions.DeleteObject, objectType))
                //    throw new System.Security.SecurityException(string.Format(Resources.UserNotAuthorizedException,
                //      "delete",
                //      objectType.Name));

                //var method = Server.DataPortalMethodCache.GetMethodInfo(
                //  objectType, "DataPortal_Delete", criteria);

                DataPortalClient.IDataPortalProxy proxy;
                proxy = GetDataPortalProxy(objectType, false);

                dpContext = new Server.DataPortalContext(GetPrincipal(), proxy.IsServerRemote);

                OnDataPortalInvoke(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Delete));

                try
                {
                    result = proxy.Delete(objectType, criteria, dpContext);
                }
                catch (Server.DataPortalException ex)
                {
                    result = ex.Result;
                    if (proxy.IsServerRemote)
                        ApplicationContext.ContextManager.SetGlobalContext(result.GlobalContext);
                    throw new DataPortalException(
                      String.Format("DataPortal.Delete {0} ({1})", Resources.Failed, ex.InnerException.InnerException),
                      ex.InnerException, result);
                }

                if (proxy.IsServerRemote)
                    ApplicationContext.ContextManager.SetGlobalContext(result.GlobalContext);

                OnDataPortalInvokeComplete(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Delete));
            }
            catch (Exception ex)
            {
                OnDataPortalInvokeComplete(new DataPortalEventArgs(dpContext, objectType, DataPortalOperations.Delete, ex));
                throw;
            }
        }

        #endregion

        #region Async Data Access Methods

        #region Begin Create

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// create a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to create.
        /// </typeparam>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        public static void BeginCreate<T>(EventHandler<DataPortalResult<T>> callback)
          where T : IMobileObject
        {
            BeginCreate<T>(null, callback, null);
        }

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// create a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to create.
        /// </typeparam>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        /// <param name="userState">User state object.</param>
        public static void BeginCreate<T>(EventHandler<DataPortalResult<T>> callback, object userState)
          where T : IMobileObject
        {
            BeginCreate<T>(null, callback, userState);
        }

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// create a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to create.
        /// </typeparam>
        /// <param name="criteria">
        /// Criteria describing the object to create.
        /// </param>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        public static void BeginCreate<T>(object criteria, EventHandler<DataPortalResult<T>> callback)
          where T : IMobileObject
        {
            BeginCreate<T>(criteria, callback, null);
        }

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// create a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to create.
        /// </typeparam>
        /// <param name="criteria">
        /// Criteria describing the object to create.
        /// </param>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        /// <param name="userState">User state object.</param>
        public static void BeginCreate<T>(object criteria, EventHandler<DataPortalResult<T>> callback, object userState)
          where T : IMobileObject
        {
            DataPortal<T> dp = new DataPortal<T>();
            dp.CreateCompleted += callback;
            dp.BeginCreate(criteria, userState);
        }

        #endregion

        #region Begin Fetch

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// fetch a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to fetch.
        /// </typeparam>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        public static void BeginFetch<T>(EventHandler<DataPortalResult<T>> callback)
          where T : IMobileObject
        {
            BeginFetch<T>(null, callback, null);
        }

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// fetch a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to fetch.
        /// </typeparam>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        /// <param name="userState">User state object.</param>
        public static void BeginFetch<T>(EventHandler<DataPortalResult<T>> callback, object userState)
          where T : IMobileObject
        {
            BeginFetch<T>(null, callback, userState);
        }

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// fetch a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to fetch.
        /// </typeparam>
        /// <param name="criteria">
        /// Criteria describing the object to fetch.
        /// </param>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        public static void BeginFetch<T>(object criteria, EventHandler<DataPortalResult<T>> callback)
          where T : IMobileObject
        {
            BeginFetch<T>(criteria, callback, null);
        }

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// fetch a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to fetch.
        /// </typeparam>
        /// <param name="criteria">
        /// Criteria describing the object to fetch.
        /// </param>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        /// <param name="userState">User state object.</param>
        public static void BeginFetch<T>(object criteria, EventHandler<DataPortalResult<T>> callback, object userState)
          where T : IMobileObject
        {
            var dp = new DataPortal<T>();
            dp.FetchCompleted += callback;
            dp.BeginFetch(criteria, userState);
        }

        #endregion

        #region Begin Update

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// update a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to update.
        /// </typeparam>
        /// <param name="obj">
        /// Business object to update.
        /// </param>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        public static void BeginUpdate<T>(object obj, EventHandler<DataPortalResult<T>> callback)
          where T : IMobileObject
        {
            BeginUpdate<T>(obj, callback, null);
        }

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// update a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to update.
        /// </typeparam>
        /// <param name="obj">
        /// Business object to update.
        /// </param>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        /// <param name="userState">User state object.</param>
        public static void BeginUpdate<T>(object obj, EventHandler<DataPortalResult<T>> callback, object userState)
          where T : IMobileObject
        {
            DataPortal<T> dp = new DataPortal<T>();
            dp.UpdateCompleted += callback;
            dp.BeginUpdate(obj, userState);
        }

        #endregion

        #region Begin Delete

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// delete a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to delete.
        /// </typeparam>
        /// <param name="criteria">
        /// Criteria describing the object to delete.
        /// </param>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        public static void BeginDelete<T>(object criteria, EventHandler<DataPortalResult<T>> callback)
          where T : IMobileObject
        {
            BeginDelete<T>(criteria, callback, null);
        }

        /// <summary>
        /// Starts an asynchronous data portal operation to
        /// delete a business object.
        /// </summary>
        /// <typeparam name="T">
        /// Type of business object to delete.
        /// </typeparam>
        /// <param name="criteria">
        /// Criteria describing the object to delete.
        /// </param>
        /// <param name="callback">
        /// Reference to method that will handle the 
        /// asynchronous callback when the operation
        /// is complete.
        /// </param>
        /// <param name="userState">User state object.</param>
        public static void BeginDelete<T>(object criteria, EventHandler<DataPortalResult<T>> callback, object userState)
          where T : IMobileObject
        {
            var dp = new DataPortal<T>();
            dp.DeleteCompleted += callback;
            dp.BeginDelete(criteria, userState);
        }

        #endregion
        
        #endregion

        #region Portal Proxy

        private static IDataPortalProxyFactory _ProxyFactory;

        private static IDataPortalProxy GetDataPortalProxy(Type objectType, bool forceLocal)
        {
            //if(string.IsNullOrEmpty(ApplicationContext.PortalUrlString

            if (forceLocal)
            {
                return new DataPortalClient.LocalProxy();
            }
            else
            {
                // load dataportal factory if loaded 
                if (_ProxyFactory == null)
                    LoadPortalProxyFactory();

                return _ProxyFactory.Create(objectType);
            }
        }

        /// <summary>
        /// Loads the data portal factory.
        /// </summary>
        private static void LoadPortalProxyFactory()
        {
            if (_ProxyFactory == null)
            {
                _ProxyFactory = new DataPortalClient.DefaultPortalProxyFactory();
            }
        }

        /// <summary>
        /// Resets the data portal proxy type, so the
        /// next data portal call will reload the proxy
        /// type based on current configuration values.
        /// </summary>
        public static void ResetProxyFactory()
        {
            _ProxyFactory = null;
        }

        /// <summary>
        /// Resets the data portal proxy type, so the
        /// next data portal call will reload the proxy
        /// type based on current configuration values.
        /// </summary>
        public static void ResetProxyType()
        {
            if (_ProxyFactory != null)
            {
                _ProxyFactory.ResetProxyType();
            }
        }

        /// <summary>
        /// Releases any remote data portal proxy object, so
        /// the next data portal call will create a new
        /// proxy instance.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never)]
        [Obsolete("Proxies no longer cached")]
        public static void ReleaseProxy()
        { }

        #endregion

        #region Security

        private static System.Security.Principal.IPrincipal GetPrincipal()
        {
            if (ApplicationContext.AuthenticationType == "Windows")
            {
                // 集成身份验证时，不传递身份凭据
                return null;
            }
            else
            {
                // we assume using the CSLA framework security
                return ApplicationContext.User;
            }
        }

        #endregion
    }
}
