﻿using System.Collections.Generic;
using System;

namespace Wcdf.Rules
{
    /// <summary>
    /// Flags enum to define when rule is allowed or denied to run 
    /// </summary>
    [Flags]
    public enum RunModes
    {
        /// <summary>
        /// Default value, rule can run in any context
        /// </summary>
        Default = 0,
        /// <summary>
        /// Deny rule from running in CheckRules 
        /// </summary>
        DenyCheckRules = 1,
        /// <summary>
        /// Deny rule from running as AffectedProperties from another rule.
        /// </summary>
        DenyAsAffectedProperty = 2,
        /// <summary>
        /// Deny rule from running on serverside portal
        /// </summary>
        DenyOnServerSidePortal = 4
    }

    /// <summary>
    /// 定义业务规则检验需要实现的接口  
    /// </summary>
    public interface IBusinessRule
    {
        /// <summary>
        /// 规则实现
        /// </summary>
        /// <param name="context">Rule context object.</param>
        void Execute(RuleContext context);
        /// <summary>
        /// 获取辅助属性以提供给执行的规则.
        /// Gets a list of secondary property values to be supplied to the
        /// rule when it is executed.
        /// </summary>
        List<Wcdf.Core.IPropertyInfo> InputProperties { get; }
        /// <summary>
        /// 规则所影响的属性
        /// </summary>
        List<Wcdf.Core.IPropertyInfo> AffectedProperties { get; }
        /// <summary>
        /// 规则所影响的主要属性
        /// </summary>
        Wcdf.Core.IPropertyInfo PrimaryProperty { get; }
        /// <summary>
        /// 取得特定实例的上下文业务规则URI.
        /// Gets a unique rule:// URI for the specific instance
        /// of the rule within the context of the business object
        /// where the rule is used.
        /// </summary>
        string RuleName { get; }
        /// <summary>
        /// 规则的重要程序
        /// Gets the rule priority.
        /// </summary>
        int Priority { get; }
        /// <summary>
        /// 规则执行方法是否是异步代码.    
        /// </summary>
        bool IsAsync { get; }

        /// <summary>
        /// Gets a value indicating that the Target property should
        /// be set even for an async rule (note that using Target
        /// from a background thread will cause major problems).
        /// </summary>
        bool ProvideTargetWhenAsync { get; }

        /// <summary>
        /// Gets the context in which the rule is allowed to execute.
        /// </summary>
        /// <value>The run in context.</value>
        RunModes RunMode { get; }
    }
}
