﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.IO;
using System.Runtime.Serialization;
using Wcdf.Properties;
using Wcdf.Reflection;
using System.Globalization;

namespace Wcdf.Serialization.Mobile
{
    /// <summary>
    /// 在字段级的序列化和反序列化类.
    /// 一种兼容SilverLight的二进制序列化器或者Wcf专用序列化器    
    /// </summary>
    public sealed class MobileFormatter : ISerializationFormatter
    {
        #region Serialize

        /// <summary>
        /// 将对象序列化到XML中
        /// </summary>
        /// <param name="serializationStream">
        /// 包含序列化后的数据流对象.
        /// </param>
        /// <param name="graph">
        /// 
        /// Root object of the object graph
        /// to serialize.
        /// </param>
        public void Serialize(Stream serializationStream, object graph)
        {
            XmlWriter writer = GetXmlWriter(serializationStream);
            Serialize(writer, graph);
            writer.Flush();
        }

        /// <summary>
        /// 将对象序列化到XML中.
        /// </summary>
        /// <param name="textWriter">
        /// 包含序列化后的TextWriter对象
        /// </param>
        /// <param name="graph">
        /// Root object of the object graph
        /// to serialize.
        /// </param>
        public void Serialize(TextWriter textWriter, object graph)
        {
            XmlWriter writer = XmlWriter.Create(textWriter);
            Serialize(writer, graph);
            writer.Flush();
        }

        /// <summary>
        /// 将对象序列化到XML中.
        /// </summary>
        /// <param name="writer">
        /// 包含序列化后的XmlWriter对象.
        /// </param>
        /// <param name="graph">
        /// Root object of the object graph
        /// to serialize.
        /// </param>
        public void Serialize(XmlWriter writer, object graph)
        {
            List<SerializationInfo> serialized = SerializeAsDTO(graph);

            DataContractSerializer dc = GetDataContractSerializer();

            dc.WriteObject(writer, serialized);
        }

        /// <summary>
        /// 序列对象图到DTO(数据传输对象)
        /// </summary>
        /// <param name="graph">
        /// Root object of the object graph
        /// to serialize.
        /// </param>
        public List<SerializationInfo> SerializeAsDTO(object graph)
        {
            _serializationReferences.Clear();

            SerializeObject(graph);

            ConvertEnumsToIntegers();
            List<SerializationInfo> serialized = _serializationReferences.Values.ToList();

            return serialized;
        }

        /// <summary>
        /// <para>
        /// 将<see cref="SerializationInfo" />中的枚举值转换为整型声明.
        /// 通常,<see cref="DataContractSerializer" />需要所有非标准原始类型在序列化和反序列化过程中提供自身的构造方法
        /// 因为并没有办法可知枚举值被序列化为哪种值.没有办法在反序列化时提供类型构造器.        
        /// </para>
        /// <para>
        /// 做为替代方式,我们将枚举值在序列化前转换为整数,在反序列化之后再转回枚举对象.        
        /// </para>
        /// </summary>
        /// <seealso cref="ConvertEnumsFromIntegers" />
        private void ConvertEnumsToIntegers()
        {
            foreach (SerializationInfo serializationInfo in _serializationReferences.Values)
            {
                foreach (SerializationInfo.FieldData fieldData in serializationInfo.Values.Values)
                {
                    if (fieldData.Value != null)
                    {
                        Type fieldType = fieldData.Value.GetType();

                        if (fieldType.IsEnum)
                        {
#if SILVERLIGHT
              fieldData.Value = Convert.ChangeType(fieldData.Value, Enum.GetUnderlyingType(fieldType), CultureInfo.CurrentCulture);
#else
                            fieldData.Value = Convert.ChangeType(fieldData.Value, Enum.GetUnderlyingType(fieldType));
#endif
                            fieldData.EnumTypeName = fieldType.AssemblyQualifiedName;
                        }
                    }
                }
            }
        }

        private DataContractSerializer GetDataContractSerializer()
        {
            return new DataContractSerializer(
              typeof(List<SerializationInfo>),
              new Type[] { typeof(List<int>), typeof(byte[]), typeof(DateTimeOffset) });
        }

        /// <summary>
        /// 将对象序列化SerializationInfo类型
        /// </summary>
        /// <param name="obj">待序列化对象.</param>
        /// <returns></returns>
        public SerializationInfo SerializeObject(object obj)
        {
            SerializationInfo info;
            if (obj == null)
            {
                NullPlaceholder nullPlaceholder = new NullPlaceholder();

                info = new SerializationInfo(_serializationReferences.Count + 1);
                _serializationReferences.Add(nullPlaceholder, info);

                info.TypeName = typeof(NullPlaceholder).AssemblyQualifiedName;
            }
            else
            {
                var thisType = obj.GetType();
                if (!IsSerializable(thisType))
                    throw new InvalidOperationException(
                      string.Format(Resources.ObjectNotSerializableFormatted, thisType.FullName));
                var mobile = obj as IMobileObject;
                if (mobile == null)
                    throw new InvalidOperationException(
                      string.Format(Resources.MustImplementIMobileObject,
                      thisType.Name));

                if (!_serializationReferences.TryGetValue(mobile, out info))
                {
                    info = new SerializationInfo(_serializationReferences.Count + 1);
                    _serializationReferences.Add(mobile, info);

                    info.TypeName = thisType.AssemblyQualifiedName;

                    mobile.GetChildren(info, this);
                    mobile.GetState(info);
                }
            }
            return info;
        }

        private Dictionary<IMobileObject, SerializationInfo> _serializationReferences =
          new Dictionary<IMobileObject, SerializationInfo>(new ReferenceComparer<IMobileObject>());

        private static bool IsSerializable(Type objectType)
        {
#if SILVERLIGHT
      return objectType.IsSerializable();
#else
            return objectType.IsSerializable;
#endif
        }

        #endregion

        #region Deserialize

        private Dictionary<int, IMobileObject> _deserializationReferences =
          new Dictionary<int, IMobileObject>();

        private Dictionary<string, Type> _typeCache = new Dictionary<string, Type>();

        private Type GetTypeFromCache(string typeName)
        {
            Type result;
            if (!_typeCache.TryGetValue(typeName, out result))
            {
                result = Wcdf.Reflection.MethodCaller.GetType(typeName);

                if (result == null)
                {
                    throw new SerializationException(string.Format(
                      Resources.MobileFormatterUnableToDeserialize,
                      typeName));
                }
                _typeCache.Add(typeName, result);
            }

            return result;
        }

        /// <summary>
        /// Deserialize an object from XML.
        /// </summary>
        /// <param name="serializationStream">
        /// Stream containing the serialized XML
        /// data.
        /// </param>
        /// <returns></returns>
        public object Deserialize(Stream serializationStream)
        {
            XmlReader reader = GetXmlReader(serializationStream);
            return Deserialize(reader);
        }

        /// <summary>
        /// Deserialize an object from XML.
        /// </summary>
        /// <param name="textReader">
        /// TextReader containing the serialized XML
        /// data.
        /// </param>
        /// <returns></returns>
        public object Deserialize(TextReader textReader)
        {
            XmlReader reader = XmlReader.Create(textReader);
            return Deserialize(reader);
        }

        /// <summary>
        /// Deserialize an object from XML.
        /// </summary>
        /// <param name="reader">
        /// XmlReader containing the serialized XML
        /// data.
        /// </param>
        /// <returns></returns>
        public object Deserialize(XmlReader reader)
        {
            DataContractSerializer dc = GetDataContractSerializer();

            List<SerializationInfo> deserialized = dc.ReadObject(reader) as List<SerializationInfo>;
            return Deserialize(deserialized);

        }

        /// <summary>
        /// Deserialize an object from DTO graph.
        /// </summary>
        ///<param name="deserialized">DTO grap to deserialize</param>
        /// <returns></returns>
        public object DeserializeAsDTO(List<SerializationInfo> deserialized)
        {

            _deserializationReferences = new Dictionary<int, IMobileObject>();
            foreach (SerializationInfo info in deserialized)
            {
                //Type type = Csla.Reflection.MethodCaller.GetType(info.TypeName);
                Type type = GetTypeFromCache(info.TypeName);

                if (type == null)
                {
                    throw new SerializationException(string.Format(
                      Resources.MobileFormatterUnableToDeserialize,
                      info.TypeName));
                }
                else if (type == typeof(NullPlaceholder))
                {
                    _deserializationReferences.Add(info.ReferenceId, null);
                }
                else
                {
#if SILVERLIGHT
          IMobileObject mobile = (IMobileObject)Activator.CreateInstance(type);
#else
                    IMobileObject mobile = (IMobileObject)Activator.CreateInstance(type, true);
#endif

                    _deserializationReferences.Add(info.ReferenceId, mobile);

                    ConvertEnumsFromIntegers(info);
                    mobile.SetState(info);
                }
            }

            foreach (SerializationInfo info in deserialized)
            {
                IMobileObject mobile = _deserializationReferences[info.ReferenceId];

                if (mobile != null)
                {
                    mobile.SetChildren(info, this);
                }
            }

            foreach (SerializationInfo info in deserialized)
            {
                ISerializationNotification notifiable = _deserializationReferences[info.ReferenceId] as ISerializationNotification;
                if (notifiable != null)
                    notifiable.Deserialized();
            }

            return (_deserializationReferences.Count > 0 ? _deserializationReferences[1] : null);
        }

        /// <summary>
        /// 将以Int申明的枚举值转换回枚举类型形式．        
        /// </summary>
        /// <seealso cref="ConvertEnumsToIntegers" />
        private static void ConvertEnumsFromIntegers(SerializationInfo serializationInfo)
        {
            foreach (SerializationInfo.FieldData fieldData in serializationInfo.Values.Values)
            {
                if (!String.IsNullOrEmpty(fieldData.EnumTypeName))
                {
                    Type enumType = MethodCaller.GetType(fieldData.EnumTypeName);
                    fieldData.Value = Enum.ToObject(enumType, fieldData.Value);
                }
            }
        }

        /// <summary>
        /// 通过对象的引用ID取得反序列化流中的对象        
        /// </summary>
        /// <param name="referenceId">流中的对象ID.</param>
        /// <returns></returns>
        public IMobileObject GetObject(int referenceId)
        {
            return _deserializationReferences[referenceId];
        }

        #endregion

        #region Static Helpers

        /// <summary>
        /// 将对象序列化到字节数组
        /// </summary>
        /// <param name="obj">
        /// 继承IMobileObject的序列化对象.
        /// </param>
        /// <returns></returns>
        public static byte[] Serialize(object obj)
        {
            using (var buffer = new System.IO.MemoryStream())
            {
                var formatter = new MobileFormatter();
                formatter.Serialize(buffer, obj);
                return buffer.ToArray();
            }
        }

        /// <summary>
        /// 序列化对象为DTO
        /// </summary>
        /// <param name="obj">
        /// The object to be serialized, which must implement
        /// IMobileObject.
        /// </param>
        /// <returns></returns>
        public static List<SerializationInfo> SerializeToDTO(object obj)
        {
            var formatter = new MobileFormatter();
            return formatter.SerializeAsDTO(obj);
        }

        /// <summary>
        /// 从DTO图中反序列化回对象
        /// </summary>
        /// <param name="serialized">DTO对象图</param>
        /// <returns>反序列化后的对象</returns>
        public static object DeserializeFromDTO(List<SerializationInfo> serialized)
        {
            var formatter = new MobileFormatter();
            return formatter.DeserializeAsDTO(serialized);
        }

        /// <summary>
        /// 从字节流中反序列化为对象
        /// </summary>
        /// <param name="data">
        /// Byte array containing the object's serialized
        /// data.
        /// </param>
        /// <returns>
        /// An object containing the data from the
        /// byte stream. The object must implement
        /// IMobileObject to be deserialized.
        /// </returns>
        public static object Deserialize(byte[] data)
        {
            using (var buffer = new System.IO.MemoryStream(data))
            {
                var formatter = new MobileFormatter();
                return formatter.Deserialize(buffer);
            }
        }

        /// <summary>
        /// Deserializes a byte stream into an object.
        /// </summary>
        /// <param name="data">
        /// DTO containing the object's serialized
        /// data.
        /// </param>
        /// <returns>
        /// An object containing the data from the
        /// byte stream. The object must implement
        /// IMobileObject to be deserialized.
        /// </returns>
        public static object Deserialize(List<SerializationInfo> data)
        {
            var formatter = new MobileFormatter();
            return formatter.DeserializeAsDTO(data);
        }
        #endregion

        #region XmlReader/XmlWriter

#if SILVERLIGHT
    private static bool _useBinaryXml = true;

    /// <summary>
    /// Gets or sets a value indicating whether to use
    /// binary XML for serialization. Defaults to true.
    /// </summary>
    public static bool UseBinaryXml
    {
      get { return _useBinaryXml; }
      set { _useBinaryXml = value; }
    }
#else
        private static bool _useBinaryXml;
        private static bool _useBinaryXmlSet;

        /// <summary>
        /// 是否使用二进制XML序列化.默认为True.        
        /// </summary>
        public static bool UseBinaryXml
        {
            get
            {
                if (!_useBinaryXmlSet)
                {
                    string tmp = System.Configuration.ConfigurationManager.AppSettings["CslaUseBinaryXml"];
                    if (string.IsNullOrEmpty(tmp))
                        _useBinaryXml = true;
                    else
                        _useBinaryXml = bool.Parse(tmp);
                    _useBinaryXmlSet = true;
                }
                return _useBinaryXml;
            }
            set
            {
                _useBinaryXml = value;
                _useBinaryXmlSet = true;
            }
        }
#endif


        /// <summary>
        /// 取得XmlWriter对象
        /// </summary>
        /// <param name="stream">The stream to which you
        /// want to write.</param>
        /// <returns></returns>
        /// <remarks>
        /// 根据UseBinaryXml创建不同的写入器.
        /// 当为True,为XmlDictionaryWriter
        /// 为False,为基本的XmlWriter.
        /// </remarks>
        public static XmlWriter GetXmlWriter(Stream stream)
        {
            if (UseBinaryXml)
                return XmlDictionaryWriter.CreateBinaryWriter(stream);
            else
                return XmlWriter.Create(stream);
        }

        /// <summary>
        /// 取得新的XmlReader对象
        /// </summary>
        /// <param name="stream">The stream from which you
        /// want to read.</param>
        /// <returns></returns>
        /// <remarks>
        /// 根据UseBinaryXml创建不同的写入器.
        /// 当为True,为XmlDictionaryWriter
        /// 为False,为基本的XmlWriter.
        /// </remarks>
        public static XmlReader GetXmlReader(Stream stream)
        {
            if (UseBinaryXml)
                return XmlDictionaryReader.CreateBinaryReader(stream, XmlDictionaryReaderQuotas.Max);
            else
                return XmlReader.Create(stream);
        }

        #endregion
    }
}
