﻿using System;
using System.Collections.Generic;
using System.Runtime.Serialization;

namespace Wcdf.Serialization.Mobile
{
    /// <summary>
    /// 包含特定对象序列化数据的对象    
    /// </summary>
    ///<remarks>本类是框架内容序列化的容器,对于WCF或者SL.序列化器采用DataContractSerializer,
    ///因此该类及成员都进行了序列化标记.
    ///</remarks>
    [Serializable]
    [DataContract]
    public class SerializationInfo : IMobileObject
    {
        /// <summary>
        /// 包含对象字段信息的对象        
        /// </summary>
        [Serializable]
        [DataContract]
        public class FieldData : IMobileObject
        {
            /// <summary>
            /// 字段值.
            /// </summary>
            [DataMember]
            public object Value { get; set; }

            /// <summary>
            /// 如果非NULL值.表示该值是整值,对应一个特定的enum类型.通过反序列化,整型会被转换回
            /// 枚举类型.            
            /// </summary>
            [DataMember]
            public string EnumTypeName { get; set; }

            /// <summary>
            /// 指示字段是否为脏数据.
            /// </summary>
            [DataMember]
            public bool IsDirty { get; set; }

            #region IMobileObject Members

            /// <summary>
            /// 构造方法
            /// </summary>
            public FieldData() { }

            /// <summary>
            /// 取得序列化对象中的状态信息.
            /// </summary>
            /// <param name="info">序列化上下文.</param>
            public void GetState(SerializationInfo info)
            {
                info.AddValue("FieldData.Value", Value);
                info.AddValue("FieldData.EnumTypeName", EnumTypeName);
                info.AddValue("FieldData.IsDirty", IsDirty);
            }

            /// <summary>
            /// 取得子对象序列化信息.
            /// </summary>
            /// <param name="info">序列化上下文信息.</param>
            /// <param name="formatter">序列化器实例.</param>
            public void GetChildren(SerializationInfo info, MobileFormatter formatter)
            {
            }

            /// <summary>
            /// 设置状态信息.
            /// </summary>
            /// <param name="info">序列化上下文.</param>
            public void SetState(SerializationInfo info)
            {
                Value = info.GetValue<object>("FieldData.Value");
                EnumTypeName = info.GetValue<string>("FieldData.EnumTypeName");
                IsDirty = info.GetValue<bool>("FieldData.IsDirty");
            }

            /// <summary>
            /// 设置子对象的序列化信息
            /// </summary>
            /// <param name="info">Serialization context.</param>
            /// <param name="formatter">Serializer instance.</param>
            public void SetChildren(SerializationInfo info, MobileFormatter formatter)
            {
            }

            #endregion
        }

        /// <summary>
        /// 对象的子对象引用信息.
        /// </summary>
        [Serializable]
        [DataContract]
        public class ChildData : IMobileObject
        {
            /// <summary>
            /// 子对象的引用.
            /// </summary>
            [DataMember]
            public int ReferenceId { get; set; }
            /// <summary>
            /// 指示子对象是否是脏(改动).
            /// </summary>
            [DataMember]
            public bool IsDirty { get; set; }

            #region IMobileObject Members

            /// <summary>
            /// Creates instance of object.
            /// </summary>
            public ChildData() { }

            /// <summary>
            /// 对象序列化时需要由MobileFormatter调用该方法.
            /// 序列化数据需要在SerializationInfo信息中            
            /// </summary>
            /// <param name="info">
            /// 包含序列化数据的对象            
            /// </param>
            public void GetState(SerializationInfo info)
            {
                info.AddValue("ChildData.ReferenceId", ReferenceId);
                info.AddValue("ChildData.IsDirty", IsDirty);
            }

            /// <summary>
            /// 当对象序列化该子对象时需要由MobileFormatter调用该方法.
            /// 序列化数据需要在SerializationInfo信息中.            
            /// </summary>
            /// <param name="info">
            /// Object to contain the serialized data.
            /// </param>
            /// <param name="formatter">
            /// Reference to the formatter performing the serialization.
            /// </param>
            public void GetChildren(SerializationInfo info, MobileFormatter formatter)
            {
            }

            /// <summary>
            /// 当对象需要反序列化时,由MobileFormatter调用该方法.            
            /// 反序列化数据来源于SerializationInfo信息中.
            /// </summary>
            /// <param name="info">
            /// Object containing the serialized data.
            /// </param>
            public void SetState(SerializationInfo info)
            {
                ReferenceId = info.GetValue<int>("ChildData.ReferenceId");
                IsDirty = info.GetValue<bool>("ChildData.IsDirty");
            }

            /// <summary>
            /// 当对象需要反序列化该子对象时,由MobileFormatter调用该方法.
            /// 反序列化数据来源于SerializationInfo信息中.
            /// </summary>
            /// <param name="info">
            /// Object containing the serialized data.
            /// </param>
            /// <param name="formatter">
            /// Reference to the formatter performing the deserialization.
            /// </param>
            public void SetChildren(SerializationInfo info, MobileFormatter formatter)
            {
            }

            #endregion
        }

        private Dictionary<string, ChildData> _children = new Dictionary<string, ChildData>();
        /// <summary>
        /// Dictionary对象,包含子对象引用数据.
        /// </summary>
        [DataMember()]
        public Dictionary<string, ChildData> Children
        {
            get { return _children; }
            set { _children = value; }
        }

        private Dictionary<string, FieldData> _values = new Dictionary<string, FieldData>();
        /// <summary>
        /// Dictionary对象.包含字段数据
        /// </summary>
        [DataMember()]
        public Dictionary<string, FieldData> Values
        {
            get { return _values; }
            set { _values = value; }
        }

        internal SerializationInfo(int referenceId)
        {
            this.ReferenceId = referenceId;
        }

        /// <summary>
        /// 对象引用的数值.
        /// </summary>
        [DataMember]
        public int ReferenceId { get; set; }
        /// <summary>
        /// 类型名:Assembly-qualified type name(完整限定名)        
        /// </summary>
        [DataMember]
        public string TypeName { get; set; }

        /// <summary>
        /// 增加一个字段值到序列化流中
        /// </summary>
        /// <param name="name">
        /// 字段名称.
        /// </param>
        /// <param name="value">
        /// 字段值.
        /// </param>
        public void AddValue(string name, object value)
        {
            AddValue(name, value, false);
        }

        /// <summary>
        /// 将值加到字段列表中.
        /// </summary>
        /// <param name="name">
        /// 字段名.
        /// </param>
        /// <param name="value">
        /// 字段值.
        /// </param>
        /// <param name="isDirty">
        /// 标识是否为脏数据.
        /// </param>
        public void AddValue(string name, object value, bool isDirty)
        {
            _values.Add(name, new FieldData { Value = value, IsDirty = isDirty });
        }

        /// <summary>
        /// 从字段列表中取值
        /// </summary>
        /// <typeparam name="T">
        /// 强制转换类型
        /// </typeparam>
        /// <param name="name">
        /// 字段名称.
        /// </param>
        /// <returns></returns>
        public T GetValue<T>(string name)
        {
            var value = _values[name].Value;
            return (value != null ? (T)Utilities.CoerceValue<T>(value.GetType(), null, value) : (T)value);
        }

        /// <summary>
        /// Adds a child to the list of child references.
        /// </summary>
        /// <param name="name">
        /// Name of the field.
        /// </param>
        /// <param name="referenceId">
        /// Reference id for the child object.
        /// </param>
        public void AddChild(string name, int referenceId)
        {
            AddChild(name, referenceId, false);
        }

        /// <summary>
        /// Adds a child to the list of child references.
        /// </summary>
        /// <param name="name">
        /// Name of the field.
        /// </param>
        /// <param name="referenceId">
        /// Reference id for the child object.
        /// </param>
        /// <param name="isDirty">
        /// Flag indicating whether the child is dirty.
        /// </param>
        public void AddChild(string name, int referenceId, bool isDirty)
        {
            _children.Add(name, new ChildData { ReferenceId = referenceId, IsDirty = isDirty });
        }

        #region IMobileObject Members

        /// <summary>
        /// Creates an instance of the object.
        /// </summary>
        public SerializationInfo() { }

        /// <summary>
        /// Method called by MobileFormatter when an object
        /// should serialize its data. The data should be
        /// serialized into the SerializationInfo parameter.
        /// </summary>
        /// <param name="info">
        /// Object to contain the serialized data.
        /// </param>
        public void GetState(SerializationInfo info)
        {
            info.AddValue("SerializationInfo.ReferenceId", ReferenceId);
            info.AddValue("SerializationInfo.TypeName", TypeName);
        }

        /// <summary>
        /// Method called by MobileFormatter when an object
        /// should serialize its child references. The data should be
        /// serialized into the SerializationInfo parameter.
        /// </summary>
        /// <param name="info">
        /// Object to contain the serialized data.
        /// </param>
        /// <param name="formatter">
        /// Reference to the formatter performing the serialization.
        /// </param>
        public void GetChildren(SerializationInfo info, MobileFormatter formatter)
        {
            foreach (string key in _children.Keys)
            {
                ChildData value = _children[key];
                SerializationInfo si = formatter.SerializeObject(value);
                info.AddChild(key, si.ReferenceId);
            }
            foreach (string key in _values.Keys)
            {
                FieldData value = _values[key];
                SerializationInfo si = formatter.SerializeObject(value);
                info.AddChild(key, si.ReferenceId);
            }
        }

        /// <summary>
        /// Method called by MobileFormatter when an object
        /// should be deserialized. The data should be
        /// deserialized from the SerializationInfo parameter.
        /// </summary>
        /// <param name="info">
        /// Object containing the serialized data.
        /// </param>
        public void SetState(SerializationInfo info)
        {
            ReferenceId = info.GetValue<int>("SerializationInfo.ReferenceId");
            TypeName = info.GetValue<string>("SerializationInfo.TypeName");
        }

        /// <summary>
        /// Method called by MobileFormatter when an object
        /// should deserialize its child references. The data should be
        /// deserialized from the SerializationInfo parameter.
        /// </summary>
        /// <param name="info">
        /// Object containing the serialized data.
        /// </param>
        /// <param name="formatter">
        /// Reference to the formatter performing the deserialization.
        /// </param>
        public void SetChildren(SerializationInfo info, MobileFormatter formatter)
        {
            foreach (string key in info.Children.Keys)
            {
                int referenceId = info.Children[key].ReferenceId;
                object serialized = formatter.GetObject(referenceId);
                if (serialized is ChildData)
                {
                    _children.Add(key, (ChildData)serialized);
                }
                else
                {
                    _values.Add(key, (FieldData)serialized);
                }
            }
        }

        #endregion
    }
}
