﻿using System;
using Wcdf.Properties;
using System.Security.Principal;

namespace Wcdf.Server
{
    /// <summary>
    /// 服务端的数据门户访问入口，消息路由类
    /// </summary>
    public class DataPortal : IDataPortalServer
    {
        #region Constructors
        /// <summary>
        /// Default constructor
        /// </summary>
        public DataPortal()
        { }

        #endregion

        #region Data Access

        /// <summary>
        /// Create a new business object.
        /// </summary>
        /// <param name="objectType">Type of business object to create.</param>
        /// <param name="criteria">Criteria object describing business object.</param>
        /// <param name="context">
        /// <see cref="Server.DataPortalContext" /> object passed to the server.
        /// </param>
        public DataPortalResult Create(
          Type objectType, object criteria, DataPortalContext context)
        {
            try
            {
                SetContext(context);
                DataPortalResult result;
                context.TransactionalType = TransactionalAttribute.GetTransactionalAttribute(objectType).TransactionType;

                IDataPortalServer portal;
                switch (context.TransactionalType)
                {
                    case TransactionalTypes.EnterpriseServices:
                        portal = new ComPlusDataPortal();
                        try
                        {
                            result = portal.Create(objectType, criteria, context);
                        }
                        finally
                        {
                            ((ComPlusDataPortal)portal).Dispose();
                        }

                        break;
                    case TransactionalTypes.TransactionScope:
                        portal = new TransactionalDataPortal();
                        result = portal.Create(objectType, criteria, context);

                        break;
                    default:
                        portal = new SimplePortal();
                        result = portal.Create(objectType, criteria, context);
                        break;
                }
                return result;
            }
            catch (Wcdf.Server.DataPortalException ex)
            {
                Exception tmp = ex;
                throw;
            }
            catch (Exception ex)
            {
                throw new DataPortalException(
                  "Portal.Create " + Resources.FailedOnServer,
                  ex, new DataPortalResult());
            }
            finally
            {
                ClearContext(context);
            }
        }

        /// <summary>
        /// Get an existing business object.
        /// </summary>
        /// <param name="objectType">Type of business object to retrieve.</param>
        /// <param name="criteria">Criteria object describing business object.</param>
        /// <param name="context">
        /// <see cref="Server.DataPortalContext" /> object passed to the server.
        /// </param>
        public DataPortalResult Fetch(Type objectType, object criteria, DataPortalContext context)
        {
            try
            {
                SetContext(context);

                //Authorize(new AuthorizeRequest(objectType, criteria, DataPortalOperations.Fetch));

                DataPortalResult result;

                //DataPortalMethodInfo method = DataPortalMethodCache.GetFetchMethod(objectType, criteria);
                context.TransactionalType = TransactionalAttribute.GetTransactionalAttribute(objectType).TransactionType;

                IDataPortalServer portal;                
                //业务请求区分.
                switch (context.TransactionalType)
                {
                    case TransactionalTypes.EnterpriseServices:
                        portal = new ComPlusDataPortal();
                        try
                        {
                            result = portal.Fetch(objectType, criteria, context);
                        }
                        finally
                        {
                            ((ComPlusDataPortal)portal).Dispose();
                        }
                        break;
                    case TransactionalTypes.TransactionScope:
                        portal = new TransactionalDataPortal();
                        result = portal.Fetch(objectType, criteria, context);
                        break;
                    default:
                        portal = new SimplePortal();
                        result = portal.Fetch(objectType, criteria, context);
                        break;
                }
                return result;
            }
            catch (Wcdf.Server.DataPortalException ex)
            {
                Exception tmp = ex;
                throw;
            }
            catch (Exception ex)
            {
                throw new DataPortalException(
                  "DataPortal.Fetch " + Resources.FailedOnServer,
                  ex, new DataPortalResult());
            }
            finally
            {
                ClearContext(context);
            }
        }

        /// <summary>
        /// Update a business object.
        /// </summary>
        /// <param name="obj">Business object to update.</param>
        /// <param name="context">
        /// <see cref="Server.DataPortalContext" /> object passed to the server.
        /// </param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1062:ValidateArgumentsOfPublicMethods")]
        public DataPortalResult Update(object obj, DataPortalContext context)
        {
            try
            {
                SetContext(context);
                DataPortalResult result;
                context.TransactionalType = TransactionalAttribute.GetTransactionalAttribute(obj.GetType()).TransactionType;

                IDataPortalServer portal;
                switch (context.TransactionalType)
                {
                    case TransactionalTypes.EnterpriseServices:
                        portal = new ComPlusDataPortal();
                        try
                        {
                            result = portal.Update(obj, context);
                        }
                        finally
                        {
                            ((ComPlusDataPortal)portal).Dispose();
                        }
                        break;
                    case TransactionalTypes.TransactionScope:
                        portal = new TransactionalDataPortal();
                        result = portal.Update(obj, context);
                        break;
                    default:
                        portal = new SimplePortal();
                        result = portal.Update(obj, context);
                        break;
                }
                return result;
            }
            catch (Wcdf.Server.DataPortalException ex)
            {
                Exception tmp = ex;
                throw;
            }
            catch (Exception ex)
            {
                throw new DataPortalException(
                  "DataPortal.Update " + Resources.FailedOnServer,
                  ex, new DataPortalResult());
            }
            finally
            {
                ClearContext(context);
            }
        }

        /// <summary>
        /// Delete a business object.
        /// </summary>
        /// <param name="objectType">Type of business object to create.</param>
        /// <param name="criteria">Criteria object describing business object.</param>
        /// <param name="context">
        /// <see cref="Server.DataPortalContext" /> object passed to the server.
        /// </param>
        public DataPortalResult Delete(Type objectType, object criteria, DataPortalContext context)
        {
            try
            {
                SetContext(context);

                DataPortalResult result;
                context.TransactionalType = TransactionalAttribute.GetTransactionalAttribute(objectType).TransactionType;
                IDataPortalServer portal;
                switch (context.TransactionalType)
                {
                    case TransactionalTypes.EnterpriseServices:
                        portal = new ComPlusDataPortal();
                        try
                        {
                            result = portal.Delete(objectType, criteria, context);
                        }
                        finally
                        {
                            ((ComPlusDataPortal)portal).Dispose();
                        }
                        break;
                    case TransactionalTypes.TransactionScope:
                        portal = new TransactionalDataPortal();
                        result = portal.Delete(objectType, criteria, context);
                        break;
                    default:
                        portal = new SimplePortal();
                        result = portal.Delete(objectType, criteria, context);
                        break;
                }
                return result;
            }
            catch (Wcdf.Server.DataPortalException ex)
            {
                Exception tmp = ex;
                throw;
            }
            catch (Exception ex)
            {
                throw new DataPortalException(
                  "DataPortal.Delete " + Resources.FailedOnServer,
                  ex, new DataPortalResult());
            }
            finally
            {
                ClearContext(context);
            }
        }

        #endregion

        #region Context

        ApplicationContext.LogicalExecutionLocations _oldLocation;

        private void SetContext(DataPortalContext context)
        {
            _oldLocation = Wcdf.ApplicationContext.LogicalExecutionLocation;
            ApplicationContext.SetLogicalExecutionLocation(ApplicationContext.LogicalExecutionLocations.Server);

            // if the dataportal is not remote then
            // do nothing
            if (!context.IsRemotePortal) return;

            // set the context value so everyone knows the
            // code is running on the server
            ApplicationContext.SetExecutionLocation(ApplicationContext.ExecutionLocations.Server);

            // set the app context to the value we got from the
            // client
            ApplicationContext.SetContext(context.ClientContext, context.GlobalContext);

            // set the thread's culture to match the client
            System.Threading.Thread.CurrentThread.CurrentCulture =
              new System.Globalization.CultureInfo(context.ClientCulture);
            System.Threading.Thread.CurrentThread.CurrentUICulture =
              new System.Globalization.CultureInfo(context.ClientUICulture);

            if (ApplicationContext.AuthenticationType == "Windows")
            {
                // 当采用集成身份验证时，客户端传递来凭据的应该NULL
                if (context.Principal != null)
                {
                    System.Security.SecurityException ex =
                      new System.Security.SecurityException(Resources.NoPrincipalAllowedException);
                    //ex.Action = System.Security.Permissions.SecurityAction.Deny;
                    throw ex;
                }
                // 取得服务端本机的验证凭据
                AppDomain.CurrentDomain.SetPrincipalPolicy(PrincipalPolicy.WindowsPrincipal);
            }
            else
            {
                // 当客户端采用其他认证时，需要上传客户端凭据
                if (context.Principal == null)
                {
                    System.Security.SecurityException ex =
                      new System.Security.SecurityException(
                        Resources.BusinessPrincipalException + " Nothing");
                    //ex.Action = System.Security.Permissions.SecurityAction.Deny;
                    throw ex;
                }
                //以客户端凭据当本服务端凭据，以传递到下一个服务端
                ApplicationContext.User = context.Principal;
            }
        }

        private void ClearContext(DataPortalContext context)
        {
            ApplicationContext.SetLogicalExecutionLocation(_oldLocation);
            // if the dataportal is not remote then
            // do nothing
            if (!context.IsRemotePortal) return;
            ApplicationContext.Clear();
            if (ApplicationContext.AuthenticationType != "Windows")
                ApplicationContext.User = null;
        }

        #endregion        
    }
}
