﻿using System;
using System.Security.Principal;
using Wcdf.Core;
using System.Runtime.Serialization;

namespace Wcdf.Server
{
    /// <summary>
    /// 客户端和服务端在交互时传递的上下文信息.    
    /// </summary>
    [Serializable()]
    [KnownType(typeof(GenericIdentity))]
    [KnownType(typeof(GenericPrincipal))]
    [KnownType(typeof(Wcdf.Security.WcdfIdentity))]
    [KnownType(typeof(Wcdf.Security.WcdfPrincipal))]
    [KnownType(typeof(Wcdf.Security.UnauthenticatedIdentity))]
    [KnownType(typeof(Wcdf.Security.UnauthenticatedPrincipal))]
    public class DataPortalContext
    {
        public const string Name = "Wcdf.Server.DataPortalContext";
        public const string NameSpace = "http://wcdf.stringsun.com";
        public const string LoginInfo = "Wcdf.LoginInfo";

        private IPrincipal _principal;
        private bool _remotePortal;        
        private string _clientCulture;
        private string _clientUICulture;
        private ContextDictionary _clientContext;
        private ContextDictionary _globalContext;

        [NonSerialized]
        private TransactionalTypes _transactionalType;
        //[NonSerialized]
        //private ObjectFactoryAttribute _factoryInfo;

        /// <summary>
        /// 帐户信息.
        /// </summary>
        public IPrincipal Principal
        {
            get { return _principal; }
        }

        /// <summary>
        /// 返回 <see langword="true" /> 如果是远程与服务器交互.
        /// </summary>
        public bool IsRemotePortal
        {
            get { return _remotePortal; }
        }        

        /// <summary>
        /// 客户端语言类别
        /// </summary>
        public string ClientCulture
        {
            get { return _clientCulture; }
        }

        /// <summary>
        /// 客户端设置的语言类别        
        /// </summary>
        public string ClientUICulture
        {
            get { return _clientUICulture; }
        }
        internal ContextDictionary ClientContext
        {
            get { return _clientContext; }
        }

        internal ContextDictionary GlobalContext
        {
            get { return _globalContext; }
        }

        /// <summary>
        /// 事务传输方式.可在服务端生效.
        /// Gets the current transactional type. Only valid
        /// in the server-side data portal methods after
        /// the transactional type has been determined.
        /// </summary>
        public TransactionalTypes TransactionalType
        {
            get { return _transactionalType; }
            internal set { _transactionalType = value; }
        }

        /// <summary>
        /// Creates a new DataPortalContext object.
        /// </summary>
        /// <param name="principal">The current Principal object.</param>
        /// <param name="isRemotePortal">Indicates whether the DataPortal is remote.</param>
        public DataPortalContext(IPrincipal principal, bool isRemotePortal)
        {
            if (isRemotePortal)
            {
                _principal = principal;
                _remotePortal = isRemotePortal;
                _clientCulture =
                  System.Threading.Thread.CurrentThread.CurrentCulture.Name;
                _clientUICulture =
                  System.Threading.Thread.CurrentThread.CurrentUICulture.Name;
                _clientContext = Wcdf.ApplicationContext.ContextManager.GetClientContext();
                _globalContext = Wcdf.ApplicationContext.ContextManager.GetGlobalContext();
            }
        }

        /// <summary>
        /// Creates a new DataPortalContext object.
        /// </summary>
        /// <param name="principal">The current Principal object.</param>
        /// <param name="isRemotePortal">Indicates whether the DataPortal is remote.</param>
        /// <param name="clientContext">Client context.</param>
        /// <param name="clientCulture">Client culture.</param>
        /// <param name="clientUICulture">Client UI culture.</param>
        /// <param name="globalContext">Global context.</param>
        public DataPortalContext(IPrincipal principal, bool isRemotePortal, string clientCulture, string clientUICulture, ContextDictionary clientContext, ContextDictionary globalContext)
        {
            _principal = principal;
            _clientContext = clientContext;
            _clientCulture = clientCulture;
            _clientUICulture = clientUICulture;
            _globalContext = globalContext;
            _remotePortal = isRemotePortal;
        }
    }
}
