﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Wcdf.Reflection;
using Wcdf.Properties;

namespace Wcdf.Server
{
    /// <summary>
    /// 服务端对象处理抽象类.扩展的服务端处理类从该继承.    
    /// </summary>
    public abstract class ObjectFactory
    {
        /// <summary>
        /// Sets the IsReadOnly property on the specified
        /// object, if possible.
        /// </summary>
        /// <param name="obj">Object on which to operate.</param>
        /// <param name="value">New value for IsReadOnly.</param>
        protected void SetIsReadOnly(object obj, bool value)
        {
            var list = obj as Core.IReadOnlyBindingList;
            if (list != null)
                list.IsReadOnly = value;
        }

        /// <summary>
        /// Calls the ValidationRules.CheckRules() method 
        /// on the specified object, if possible.
        /// </summary>
        /// <param name="obj">
        /// Object on which to call the method.
        /// </param>
        protected void CheckRules(object obj)
        {
            var target = obj as IDataPortalTarget;
            if (target != null)
                target.CheckRules();
            else
                MethodCaller.CallMethodIfImplemented(obj, "CheckRules", null);
        }

        /// <summary>
        /// 调用业务对象的MarkOld方法
        /// </summary>
        /// <param name="obj">
        /// 包含MarkOld方法的对象实例.
        /// </param>
        protected void MarkOld(object obj)
        {
            var target = obj as IDataPortalTarget;
            if (target != null)
                target.MarkOld();
            else
                MethodCaller.CallMethodIfImplemented(obj, "MarkOld", null);
        }

        /// <summary>
        /// Calls the MarkNew method on the specified
        /// object, if possible.
        /// </summary>
        /// <param name="obj">
        /// Object on which to call the method.
        /// </param>
        protected void MarkNew(object obj)
        {
            var target = obj as IDataPortalTarget;
            if (target != null)
                target.MarkNew();
            else
                MethodCaller.CallMethodIfImplemented(obj, "MarkNew", null);
        }

        /// <summary>
        /// Calls the MarkAsChild method on the specified
        /// object, if possible.
        /// </summary>
        /// <param name="obj">
        /// Object on which to call the method.
        /// </param>
        protected void MarkAsChild(object obj)
        {
            var target = obj as IDataPortalTarget;
            if (target != null)
                target.MarkAsChild();
            else
                MethodCaller.CallMethodIfImplemented(obj, "MarkAsChild", null);
        }

        /// <summary>
        /// Loads a property's managed field with the 
        /// supplied value calling PropertyHasChanged 
        /// if the value does change.
        /// </summary>
        /// <typeparam name="P">
        /// Type of the property.
        /// </typeparam>
        /// <param name="obj">
        /// Object on which to call the method. 
        /// </param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <param name="newValue">
        /// The new value for the property.</param>
        /// <remarks>
        /// No authorization checks occur when this method is called,
        /// and no PropertyChanging or PropertyChanged events are raised.
        /// Loading values does not cause validation rules to be
        /// invoked.
        /// </remarks>
        protected void LoadProperty<P>(object obj, PropertyInfo<P> propertyInfo, P newValue)
        {
            var target = obj as Core.IManageProperties;
            if (target != null)
                target.LoadProperty<P>(propertyInfo, newValue);
            else
                throw new ArgumentException(Resources.IManagePropertiesRequiredException);
        }

        /// <summary>
        /// Reads a property's managed field value.
        /// </summary>
        /// <typeparam name="P"></typeparam>
        /// <param name="obj">
        /// Object on which to call the method. 
        /// </param>
        /// <param name="propertyInfo">
        /// PropertyInfo object containing property metadata.</param>
        /// <remarks>
        /// No authorization checks occur when this method is called.
        /// </remarks>
        protected P ReadProperty<P>(object obj, PropertyInfo<P> propertyInfo)
        {
            var target = obj as Core.IManageProperties;
            if (target != null)
                return target.ReadProperty(propertyInfo);
            else
                throw new ArgumentException(Resources.IManagePropertiesRequiredException);
        }

        protected List<object> GetChildren(object obj)
        {
            var target = obj as Core.IManageProperties;
            if (target != null)
                return target.GetChildren();
            else
                throw new ArgumentException(Resources.IManagePropertiesRequiredException);
        }

        /// <summary>
        /// By wrapping this property inside Using block
        /// you can set property values on 
        /// <paramref name="businessObject">business object</paramref>
        /// without raising PropertyChanged events
        /// and checking user rights.
        /// </summary>
        /// <param name="businessObject">
        /// Object on with you would like to set property values
        /// </param>
        /// <returns>
        /// An instance of IDisposable object that allows
        /// bypassing of normal authorization checks during
        /// property setting.
        /// </returns>
        protected IDisposable BypassPropertyChecks(Core.BusinessBase businessObject)
        {
            return businessObject.BypassPropertyChecks;
        }

        /// <summary>
        /// Gets a value indicating whether a managed field
        /// exists for the specified property.
        /// </summary>
        /// <param name="obj">Business object.</param>
        /// <param name="property">Property info object.</param>
        protected bool FieldExists(object obj, Core.IPropertyInfo property)
        {
            var target = obj as Core.IManageProperties;
            if (target != null)
                return target.FieldExists(property);
            else
                throw new ArgumentException(Resources.IManagePropertiesRequiredException);
        }

        /// <summary>
        /// Gets the list of deleted items from 
        /// an editable collection.
        /// </summary>
        /// <typeparam name="C">Type of child objects in the colletion.</typeparam>
        /// <param name="obj">Editable collection object.</param>
        protected Core.MobileList<C> GetDeletedList<C>(object obj)
        {
            var target = obj as Core.IEditableCollection;
            if (target != null)
                return (Core.MobileList<C>)target.GetDeletedList();
            else
                throw new ArgumentException(Resources.IEditableCollectionRequiredException);
        }


        /// <summary>
        /// 添加子对象.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="C"></typeparam>
        /// <param name="obj"></param>
        /// <param name="list"></param>
        protected void FetchChild<T, C>(T obj, C list)
        {
            var child = obj as Core.BusinessBase;
            var parent = list as System.Collections.IList;
            if (parent != null && child != null)
            {
                this.MarkAsChild(obj);
                this.MarkOld(obj);
                parent.Add(obj);
            }
        }
    }
}
