﻿using System;
using Wcdf.Properties;

namespace Wcdf.Server
{
    /// <summary>
    /// 基本的业务门户处理类    
    /// </summary>
    public class SimplePortal : IDataPortalServer
    {
        #region Data Access

        /// <summary>
        /// Create a new business object.
        /// </summary>
        /// <param name="objectType">Type of business object to create.</param>
        /// <param name="criteria">Criteria object describing business object.</param>
        /// <param name="context">
        /// <see cref="Server.DataPortalContext" /> object passed to the server.
        /// </param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1303:DoNotPassLiteralsAsLocalizedParameters", MessageId = "Csla.Server.DataPortalException.#ctor(System.String,System.Exception,Csla.Server.DataPortalResult)")]
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        public DataPortalResult Create(
          Type objectType, object criteria, DataPortalContext context)
        {
            ServiceBoundObject obj = null;
            IDataPortalTarget target = null;
            var eventArgs = new DataPortalEventArgs(context, objectType, DataPortalOperations.Fetch);
            try
            {
                obj = new ServiceBoundObject(objectType);

                obj.OnDataPorta_Invoke(eventArgs);
                obj.CallMethod("DataPortal_Create", criteria);
                obj.OnDataPorta_InvokeComplete(eventArgs);

                target = obj.Instance as IDataPortalTarget;

                if (target != null)
                {
                    target.MarkNew();
                }
                else
                {
                    obj.CallMethodIfImplemented("MarkNew");
                }
                // return the populated business object as a result
                return new DataPortalResult(obj.Instance);
            }
            catch (Exception ex)
            {
                object outval = null;
                if (obj != null) outval = obj.Instance;
                throw new DataPortalException(
                  "Portal.Create " + Resources.FailedOnServer,
                  ex, new DataPortalResult(outval));
            }
        }

        /// <summary>
        /// 取得业务对象实例
        /// </summary>
        /// <param name="objectType">业务对象类型.</param>
        /// <param name="criteria">过滤条件.</param>
        /// <param name="context">
        /// <see cref="Server.DataPortalContext" /> 传递的上下文信息.
        /// </param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1303:DoNotPassLiteralsAsLocalizedParameters", MessageId = "Csla.Server.DataPortalException.#ctor(System.String,System.Exception,Csla.Server.DataPortalResult)")]
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        public DataPortalResult Fetch(Type objectType, object criteria, DataPortalContext context)
        {
            ServiceBoundObject obj = null;
            IDataPortalTarget target = null;
            var eventArgs = new DataPortalEventArgs(context, objectType, DataPortalOperations.Fetch);
            try
            {
                // 建立业务处理对象.根据是否数据门户来区分采用的数据取得方式.
                obj = new ServiceBoundObject(objectType);


                obj.OnDataPorta_Invoke(eventArgs);
                //obj.SetInstance(obj.CallMethod("DataPortal_Fetch", criteria));
                obj.CallMethod("DataPortal_Fetch", criteria);
                obj.OnDataPorta_InvokeComplete(eventArgs);

                target = obj.Instance as IDataPortalTarget;

                if (target != null)
                {
                    target.MarkOld();
                }
                else
                {
                    obj.CallMethodIfImplemented("MarkOld");
                }
                // return the populated business object as a result
                return new DataPortalResult(obj.Instance);
            }
            catch (Exception ex)
            {
                object outval = null;
                if (obj != null) outval = obj.Instance;
                throw new DataPortalException(
                  "Portal.Fetch " + Resources.FailedOnServer,
                  ex, new DataPortalResult(outval));
            }
        }

        /// <summary>
        /// Update a business object.
        /// </summary>
        /// <param name="obj">Business object to update.</param>
        /// <param name="context">
        /// <see cref="Server.DataPortalContext" /> object passed to the server.
        /// </param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1800:DoNotCastUnnecessarily")]
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1303:DoNotPassLiteralsAsLocalizedParameters", MessageId = "Csla.Server.DataPortalException.#ctor(System.String,System.Exception,Csla.Server.DataPortalResult)")]
        public DataPortalResult Update(object obj, DataPortalContext context)
        {
            Type objectType = obj.GetType();

            var eventArgs = new DataPortalEventArgs(context, objectType, DataPortalOperations.Update);
            try
            {
                ServiceBoundObject svrObj = new ServiceBoundObject(obj);

                var target = obj as IDataPortalTarget;

                svrObj.OnDataPorta_Invoke(eventArgs);
                svrObj.CallMethod("DataPortal_Update");
                svrObj.OnDataPorta_InvokeComplete(eventArgs);

                // 是否是单业务类.
                var busObj = obj as Core.BusinessBase;
                if (busObj != null)
                {
                    if (busObj.IsDeleted)
                    {
                        if (target != null)
                            target.MarkNew();
                        else
                            svrObj.CallMethodIfImplemented("MarkNew");
                    }
                    else
                    {
                        if (target != null)
                            target.MarkOld();
                        else
                            svrObj.CallMethodIfImplemented("MarkOld");
                    }
                }
                else
                {
                    // this is an updatable collection or some other
                    // non-BusinessBase type of object
                    // tell the object to update itself
                    if (target != null)
                        target.MarkOld();
                    else
                        svrObj.CallMethodIfImplemented("MarkOld");
                }

                return new DataPortalResult(obj);
            }
            catch (Exception ex)
            {
                throw new DataPortalException(
                  "DataPortal.Update " + Resources.FailedOnServer,
                  ex, new DataPortalResult(obj));
            }
        }

        /// <summary>
        /// 直接从数据库中删除请求.
        /// </summary>
        /// <param name="objectType">Type of business object to create.</param>
        /// <param name="criteria">Criteria object describing business object.</param>
        /// <param name="context">
        /// <see cref="Server.DataPortalContext" /> object passed to the server.
        /// </param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1303:DoNotPassLiteralsAsLocalizedParameters", MessageId = "Csla.Server.DataPortalException.#ctor(System.String,System.Exception,Csla.Server.DataPortalResult)")]
        public DataPortalResult Delete(Type objectType, object criteria, DataPortalContext context)
        {
            ServiceBoundObject obj = null;
            var eventArgs = new DataPortalEventArgs(context, objectType, DataPortalOperations.Delete);
            try
            {
                // 建立业务处理对象.根据是否数据门户来区分采用的数据取得方式.
                obj = new ServiceBoundObject(objectType);

                obj.OnDataPorta_Invoke(eventArgs);
                obj.CallMethod("DataPortal_Delete", criteria);
                obj.OnDataPorta_InvokeComplete(eventArgs);

                // return the populated business object as a result
                return new DataPortalResult();
            }
            catch (Exception ex)
            {
                throw new DataPortalException(
                  "Portal.Delete " + Resources.FailedOnServer,
                  ex, new DataPortalResult());
            }
        }

        #endregion
    }
}
