﻿using System;

namespace Wcdf
{
    /// <summary>
    /// 标记类采用哪种事务方式.    
    /// </summary>
    /// <remarks>
    /// <para>
    /// Each business object method may be marked with this attribute
    /// to indicate which type of transactional technology should
    /// be used by the server-side DataPortal. The possible options
    /// are listed in the
    /// <see cref="TransactionalTypes">TransactionalTypes</see> enum.
    /// </para><para>
    /// If the Transactional attribute is not applied to a 
    /// DataPortal_XYZ method then the
    /// <see cref="TransactionalTypes.Manual">Manual</see> option
    /// is assumed.
    /// </para><para>
    /// If the Transactional attribute is applied with no explicit
    /// choice for transactionType then the
    /// TransactionScope option is assumed.
    /// </para><para>
    /// Both the EnterpriseServices and TransactionScope options provide
    /// 2-phase distributed transactional support.
    /// </para>
    /// </remarks>
    [AttributeUsage(AttributeTargets.Class)]
    public sealed class TransactionalAttribute : Attribute
    {
        internal static TransactionalAttribute GetTransactionalAttribute(Type objectType)
        {
            var result = objectType.GetCustomAttributes(typeof(TransactionalAttribute), true);
            if (result != null && result.Length > 0)
                return result[0] as TransactionalAttribute;
            else
                return new TransactionalAttribute();
        }

        private TransactionalTypes _type;

        /// <summary>
        /// 构造方法,默认采用是的数据事务方式.        
        /// </summary>
        public TransactionalAttribute()
        {
            _type = TransactionalTypes.Manual;
        }

        /// <summary>
        /// 指定事务方式        
        /// </summary>
        /// <param name="transactionType">
        /// Specifies the transactional context within which the
        /// method should run.</param>
        public TransactionalAttribute(TransactionalTypes transactionType)
        {
            _type = transactionType;
        }

        /// <summary>
        /// 事务类型,<see cref="Wcdf.TransactionalTypes"/>        
        /// </summary>
        public TransactionalTypes TransactionType
        {
            get { return _type; }
        }
    }
}
