﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.ComponentModel;
using Wcdf.Reflection;

namespace Wcdf
{
    /// <summary>
    /// 包含着一些工具方法    
    /// </summary>
    public static class Utilities
    {
        #region Replacements for VB runtime functionality

        /// <summary>
        /// 判断指定值是可转为数值类型.        
        /// </summary>
        public static bool IsNumeric(object value)
        {
            double dbl;
            return double.TryParse(value.ToString(), System.Globalization.NumberStyles.Any,
              System.Globalization.NumberFormatInfo.InvariantInfo, out dbl);
        }

        /// <summary>
        /// 延迟调用方法和属性.        
        /// </summary>
        /// <param name="target">实现方法和属性的对象.</param>
        /// <param name="methodName">属性或方法名.</param>
        /// <param name="callType">Specifies how to invoke the property or method.</param>
        /// <param name="args">List of arguments to pass to the method.</param>
        /// <returns>The result of the property or method invocation.</returns>
        public static object CallByName(
          object target, string methodName, CallType callType,
          params object[] args)
        {
            switch (callType)
            {
                case CallType.Get:
                    {
                        return MethodCaller.CallPropertyGetter(target, methodName);
                    }
                case CallType.Let:
                case CallType.Set:
                    {
                        MethodCaller.CallPropertySetter(target, methodName, args[0]);
                        return null;
                    }
                case CallType.Method:
                    {
                        return MethodCaller.CallMethod(target, methodName, args);
                    }
            }
            return null;
        }

        #endregion

        /// <summary>
        /// 返回属性类型.包括泛型处理        
        /// </summary>
        /// <param name="propertyType">Type of the
        /// property as returned by reflection.</param>
        public static Type GetPropertyType(Type propertyType)
        {
            Type type = propertyType;
            if (type.IsGenericType &&
              (type.GetGenericTypeDefinition() == typeof(Nullable<>)))
                return Nullable.GetUnderlyingType(type);
            return type;
        }

        /// <summary>
        /// 返回子对象的集合类型.        
        /// </summary>
        /// <param name="listType">Type of the list.</param>
        public static Type GetChildItemType(Type listType)
        {
            Type result = null;
            if (listType.IsArray)
                result = listType.GetElementType();
            else
            {
                DefaultMemberAttribute indexer =
                  (DefaultMemberAttribute)Attribute.GetCustomAttribute(
                  listType, typeof(DefaultMemberAttribute));
                if (indexer != null)
                    foreach (PropertyInfo prop in listType.GetProperties(
                      BindingFlags.Public |
                      BindingFlags.Instance |
                      BindingFlags.FlattenHierarchy))
                    {
                        if (prop.Name == indexer.MemberName)
                            result = Utilities.GetPropertyType(prop.PropertyType);
                    }
            }
            return result;
        }

        #region  CoerceValue

        /// <summary>
        /// 试图强制转换为不同类型        
        /// </summary>
        /// <param name="desiredType">
        /// 目标类型.
        /// </param>
        /// <param name="valueType">
        /// 值的原类型
        /// </param>
        /// <param name="oldValue">
        /// 先前值,如果有的话将会被转换后的值替代.
        /// </param>
        /// <param name="value">
        /// 需要转换的值.
        /// </param>
        /// <remarks>
        /// <para>
        /// If the desired type is a primitive type or Decimal, 
        /// empty string and null values will result in a 0 
        /// or equivalent.
        /// </para>
        /// <para>
        /// If the desired type is a Nullable type, empty string
        /// and null values will result in a null result.
        /// </para>
        /// <para>
        /// If the desired type is an enum the value's ToString()
        /// result is parsed to convert into the enum value.
        /// </para>
        /// </remarks>
        public static object CoerceValue(Type desiredType, Type valueType, object oldValue, object value)
        {
            if (desiredType.IsAssignableFrom(valueType))
            {
                // types match, just return value
                return value;
            }
            else
            {
                if (desiredType.IsGenericType)
                {
                    if (desiredType.GetGenericTypeDefinition() == typeof(Nullable<>))
                        if (value == null || value is DBNull)
                            return null;
                        else if (valueType.Equals(typeof(string)) && System.Convert.ToString(value) == string.Empty)
                            return null;
                }
                desiredType = Utilities.GetPropertyType(desiredType);
            }
            //DBNULL返回null值.对于后续的如DateTime类型处理会引起不便.
            if (desiredType.Equals(typeof(DateTime)) && value is DBNull)
                return null;

            if (desiredType.IsEnum && (valueType.Equals(typeof(string)) || Enum.GetUnderlyingType(desiredType).Equals(valueType)))
                return System.Enum.Parse(desiredType, value.ToString(),true);

            //if (desiredType.Equals(typeof(SmartDate)) && oldValue != null)
            //{
            //    if (value == null)
            //        value = string.Empty;
            //    var tmp = (SmartDate)oldValue;
            //    if (valueType.Equals(typeof(DateTime)))
            //        tmp.Date = (DateTime)value;
            //    else
            //        tmp.Text = value.ToString();
            //    return tmp;
            //}

            if ((desiredType.IsPrimitive || desiredType.Equals(typeof(decimal))) &&
                valueType.Equals(typeof(string)) && string.IsNullOrEmpty((string)value))
                value = 0;

            try
            {
                if (desiredType.Equals(typeof(string)) && value != null)
                {
                    return value.ToString();
                }
                else
                    return Convert.ChangeType(value, desiredType,null);
            }
            catch
            {
                TypeConverter cnv = TypeDescriptor.GetConverter(desiredType);
                if (cnv != null && cnv.CanConvertFrom(valueType))
                    return cnv.ConvertFrom(value);
                else
                    throw;
            }
        }


        /// <summary>
        /// Attempts to coerce a value of one type into
        /// a value of a different type.
        /// </summary>
        /// <typeparam name="D">
        /// Type to which the value should be coerced.
        /// </typeparam>
        /// <param name="valueType">
        /// Original type of the value.
        /// </param>
        /// <param name="oldValue">
        /// The previous value (if any) being replaced by
        /// the new coerced value.
        /// </param>
        /// <param name="value">
        /// The value to coerce.
        /// </param>
        /// <remarks>
        /// <para>
        /// If the desired type is a primitive type or Decimal, 
        /// empty string and null values will result in a 0 
        /// or equivalent.
        /// </para>
        /// <para>
        /// If the desired type is a Nullable type, empty string
        /// and null values will result in a null result.
        /// </para>
        /// <para>
        /// If the desired type is an enum the value's ToString()
        /// result is parsed to convert into the enum value.
        /// </para>
        /// </remarks>
        public static D CoerceValue<D>(Type valueType, object oldValue, object value)
        {
            return (D)(CoerceValue(typeof(D), valueType, oldValue, value));
        }

        #endregion
    }

    /// <summary>
    /// 调用类型,参见方法<see cref="Csla.Utilities.CallByName"/>.
    /// </summary>
    public enum CallType
    {
        /// <summary>
        /// Gets a value from a property.
        /// </summary>
        Get,
        /// <summary>
        /// Sets a value into a property.
        /// </summary>
        Let,
        /// <summary>
        /// Invokes a method.
        /// </summary>
        Method,
        /// <summary>
        /// Sets a value into a property.
        /// </summary>
        Set
    }
}
